/*
	Synogate Library, a support library for Synogate IP-core generators.
	Copyright (C) 2021 Synogate UG (haftungsbeschraenkt), www.synogate.com 
	
	This program is commercial software: you must not copy, modify or
	distribute it without explicit written license by Synogate to do so. 
	
	The full license text for non-commercial use and evaluation can be found in the distribution package of this core.
	
	This program is distributed WITHOUT ANY WARRANTY; without even the
	implied warranty of MERCHANTABILITY.
*/


#include "BlockDiagram.h"

#include <boost/math/constants/constants.hpp>
#include <gatery/utils.h>

#include <set>

namespace gtry::reporting {

Shape::Shape(BlockDiagram &blockDiagram) : m_blockDiagram(blockDiagram)
{
}

void Shape::addShape(std::unique_ptr<Shape> shape)
{
    shape->m_parent = this;
    m_subShapes.push_back(std::move(shape));
}

void Shape::layout()
{
    for (auto &s : m_subShapes) {
        s->layout();
        s->m_x = std::max(s->m_x, m_margin[WEST] + s->m_padding[WEST]);
        s->m_y = std::max(s->m_y, m_margin[NORTH] + s->m_padding[NORTH]);
    }


    m_width = m_minWidth;
    m_height = m_minHeight;

    for (auto &s : m_subShapes) {
        m_width = std::max(m_width, m_margin[EAST] + s->m_x + s->m_width + s->m_padding[EAST]);
        m_height = std::max(m_height, m_margin[SOUTH] + s->m_y + s->m_height + s->m_padding[SOUTH]);
    }
}

void Shape::computeGlobalPositions()
{
    if (m_parent == nullptr) {
        m_globalX = m_x;
        m_globalY = m_y;
    } else {
        m_globalX = m_x + m_parent->m_globalX;
        m_globalY = m_y + m_parent->m_globalY;
    }

    for (auto &s : m_subShapes)
        s->computeGlobalPositions();
}

void Shape::renderSubShapes(std::ostream &stream)
{
    for (auto &s : m_subShapes)
        s->render(stream);
}

Shape *Shape::findCommonParent(const Shape *other) const
{
    std::set<const Shape*> listOfParents;
    Shape *p = m_parent;
    while (p != nullptr) {
        listOfParents.insert(p);
        p = p->m_parent;
    }

    p = other->m_parent;
    while (p != nullptr) {
        if (listOfParents.contains(p))
            return p;
        p = p->m_parent;
    }
    return nullptr;
}




void Shape::belowOf(Shape *other, float minDistance)
{
    m_blockDiagram.canvas().addConstraint<DistanceConstraint>(other, this, SOUTH, minDistance);
}

void Shape::aboveOf(Shape *other, float minDistance)
{
    m_blockDiagram.canvas().addConstraint<DistanceConstraint>(other, this, NORTH, minDistance);
}

void Shape::leftOf(Shape *other, float minDistance)
{
    m_blockDiagram.canvas().addConstraint<DistanceConstraint>(other, this, WEST, minDistance);
}

void Shape::rightOf(Shape *other, float minDistance)
{
    m_blockDiagram.canvas().addConstraint<DistanceConstraint>(other, this, EAST, minDistance);
}



void Shape::vCenterOn(Shape *other)
{
    m_blockDiagram.canvas().addConstraint<AnchoringConstraint>(this,  boost::rational<unsigned>(1,2), other, boost::rational<unsigned>(1,2), Shape::SOUTH);
}

void Shape::hCenterOn(Shape *other)
{
    m_blockDiagram.canvas().addConstraint<AnchoringConstraint>(this, boost::rational<unsigned>(1,2), other, boost::rational<unsigned>(1,2), Shape::EAST);
}


std::tuple<float, float> Shape::getSideLocation(Side side, boost::rational<unsigned> loc)
{
    float floatLoc = loc.numerator() / (float) loc.denominator();
    switch (side) {
        case NORTH:
            return { m_globalX + (1.0f - floatLoc) * m_width, m_globalY};
        case SOUTH:
            return { m_globalX + floatLoc * m_width, m_globalY + m_height};
        case WEST:
            return { m_globalX, m_globalY + floatLoc * m_height};
        case EAST:
            return { m_globalX + m_width, m_globalY + (1.0f - floatLoc) * m_height};
    }
    return {};
}


DistanceConstraint::DistanceConstraint(Shape *shape1, Shape *shape2, Shape::Side direction, float minDistance)
{
    m_shapes[0] = shape1;
    m_shapes[1] = shape2;
    m_commonParent = m_shapes[0]->findCommonParent(m_shapes[1]);
    m_direction = direction;
    m_minDistance = minDistance;

    if (m_direction == Shape::NORTH) {
        m_direction = Shape::SOUTH;
        std::swap(m_shapes[0], m_shapes[1]);
    }

    if (m_direction == Shape::WEST) {
        m_direction = Shape::EAST;
        std::swap(m_shapes[0], m_shapes[1]);
    }
}

bool DistanceConstraint::apply()
{
    float relX[2];
    float relY[2];
    Shape *par[2];
    for (auto i : utils::Range(2)) {
        relX[i] = m_shapes[i]->m_x;
        relY[i] = m_shapes[i]->m_y;
        auto p = m_shapes[i]->m_parent;
        par[i] = m_shapes[i];
        while (p != m_commonParent) {
            relX[i] += p->m_x;
            relY[i] += p->m_y;
            par[i] = p;
            p = p->m_parent;
        }
    }

    switch (m_direction) {
        case Shape::SOUTH: {
            float d = (relY[1] - m_shapes[1]->m_padding[Shape::NORTH]) - (relY[0] + m_shapes[0]->m_padding[Shape::SOUTH] + m_shapes[0]->m_height);
            if (d < m_minDistance) {
                par[1]->m_y += m_minDistance - d;
                return true;
            }
        } break;
        case Shape::EAST: {
            float d = (relX[1] - m_shapes[1]->m_padding[Shape::WEST]) - (relX[0] + m_shapes[0]->m_padding[Shape::EAST] + m_shapes[0]->m_width);
            if (d < m_minDistance) {
                par[1]->m_x += m_minDistance - d;
                return true;
            }
        } break;
        default:
            HCL_ASSERT_HINT(false, "Invalid direction!");
    }
    return false;
}


AnchoringConstraint::AnchoringConstraint(Shape *movingShape, boost::rational<unsigned> movingShapeLoc, Shape *anchorShape, boost::rational<unsigned> anchoringShapeLoc, Shape::Side direction) :
        m_movingShape(movingShape),
        m_movingShapeLoc(movingShapeLoc),
        m_anchorShape(anchorShape),
        m_anchoringShapeLoc(anchoringShapeLoc),
        m_direction(direction)
{
    if (m_anchorShape == m_movingShape->m_parent)
        m_refFrame = m_anchorShape;
    else
        m_refFrame = m_movingShape->findCommonParent(m_anchorShape);

    if (m_direction == Shape::SOUTH)
        m_direction = Shape::NORTH;

    if (m_direction == Shape::WEST)
        m_direction = Shape::EAST;

}

bool AnchoringConstraint::apply()
{
    float target;
    if (m_direction == Shape::EAST)
        target = m_anchorShape->m_globalX + m_anchoringShapeLoc.numerator() / (float) m_anchoringShapeLoc.denominator() * m_anchorShape->m_width;
    else
        target = m_anchorShape->m_globalY + m_anchoringShapeLoc.numerator() / (float) m_anchoringShapeLoc.denominator() * m_anchorShape->m_height;

    float currentLoc;
    if (m_direction == Shape::EAST)
        currentLoc = m_movingShape->m_globalX + m_movingShapeLoc.numerator() / (float) m_movingShapeLoc.denominator() * m_movingShape->m_width;
    else
        currentLoc = m_movingShape->m_globalY + m_movingShapeLoc.numerator() / (float) m_movingShapeLoc.denominator() * m_movingShape->m_height;

    Shape *p = m_movingShape;
    while (p->getParent() != m_refFrame)
        p = p->getParent();

    float shift = target - currentLoc;

    if (shift < 1e-8f) return false;

    if (m_direction == Shape::EAST)
        p->m_x += shift;
    else
        p->m_y += shift;

    return true;
}



Connection::Connection(Port source, Port destination) : m_sourcePort(source), m_destinationPort(destination)
{

}

void Connection::render(std::ostream &stream)
{
    auto [srcX, srcY] = m_sourcePort.shape->getSideLocation(m_sourcePort.side, m_sourcePort.loc);
    auto [dstX, dstY] = m_destinationPort.shape->getSideLocation(m_destinationPort.side, m_destinationPort.loc);

    unsigned sideDiff = (4 + (unsigned)m_sourcePort.side - m_destinationPort.side) % 4;

    unsigned numNodes = 1;
    if (sideDiff == 0 || sideDiff == 2)
        numNodes = 2;


    float nodeX[2];
    float nodeY[2];

    float nameLocX;
    float nameLocY;

    float widthLocX;
    float widthRangeX;
    float widthLocY;

    if (numNodes == 1) {
        if (m_sourcePort.side == Shape::EAST || m_sourcePort.side == Shape::WEST) {
            nodeX[0] = dstX;
            nodeY[0] = srcY;
        } else {
            nodeX[0] = srcX;
            nodeY[0] = dstY;
        }
        nameLocX = (srcX+dstX)/2.0f;
        nameLocY = nodeY[0];
        widthLocX = std::min(srcX, dstX);
        widthRangeX = std::max(srcX, dstX) - widthLocX;
        widthLocY = nodeY[0];
        stream << R"D(
            <path d="M )D"<<srcX<<' '<<srcY<<" L "<<nodeX[0]<<' '<<nodeY[0]<<" L "<<dstX<<' '<<dstY<<R"D(" class=")D"<<m_style<<R"D(">
        )D";
    } else {
        float srcShift = m_sourcePort.loc.numerator() / (float)m_sourcePort.loc.denominator();
        float dstShift = m_destinationPort.loc.numerator() / (float)m_destinationPort.loc.denominator();


        if (m_sourcePort.side == Shape::EAST || m_sourcePort.side == Shape::WEST) {

            float sharedX = (srcX+dstX*2.0f)/3.0f;

            if ((srcY > dstY) == (m_sourcePort.side == Shape::EAST))
                sharedX -= (srcShift - dstShift) * 20.0f;
            else
                sharedX += (srcShift - dstShift) * 20.0f;

            nodeX[0] = sharedX;
            nodeY[0] = srcY;
            nodeX[1] = sharedX;
            nodeY[1] = dstY;


            nameLocX = (srcX+nodeX[0])/2.0f;
            nameLocY = nodeY[0];
            widthLocX = std::min(srcX, nodeX[0]);
            widthRangeX = std::max(srcX, nodeX[0]) - widthLocX;
            widthLocY = nodeY[0];


        } else {
            float sharedY = (srcY+dstY)/2.0f;

            if ((srcX > dstX) == (m_sourcePort.side == Shape::NORTH))
                sharedY -= (srcShift - dstShift) * 20.0f;
            else
                sharedY += (srcShift - dstShift) * 20.0f;


            nodeX[0] = srcX;
            nodeY[0] = sharedY;
            nodeX[1] = dstX;
            nodeY[1] = sharedY;

            nameLocX = (nodeX[0] + nodeX[1])/2.0f;
            nameLocY = nodeY[0];
            widthLocX = std::min(nodeX[0], nodeX[1]);
            widthRangeX = std::max(nodeX[0], nodeX[1]) - widthLocX;
            widthLocY = nodeY[0];
        }

        stream << R"D(
            <path d="M )D"<<srcX<<' '<<srcY<<" L "<<nodeX[0]<<' '<<nodeY[0]<<" L "<<nodeX[1]<<' '<<nodeY[1]<<" L "<<dstX<<' '<<dstY<<R"D(" class=")D"<<m_style<<R"D(">
        )D";

    }

    std::string tooltip = m_tooltip;
    if (tooltip.empty()) {
        if (!m_name.empty()) {
            tooltip = m_name;
            if (!m_width.empty())
                tooltip += '\n'+m_width;
        } else
            tooltip = m_width;
    }


    if (!tooltip.empty())
        stream << "<title>" << tooltip << "</title>\n";

    stream << "</path>\n";

    if (!m_name.empty() || !m_width.empty()) {
        stream << "<g>\n";
        if (!m_name.empty())
            stream << R"D(
            <text text-anchor="middle" x=")D" << nameLocX << R"D(" y=")D" << nameLocY-2.0f << R"D(" class="bd_connection_name">)D" << m_name << R"D(</text>
            )D";

        if (!m_width.empty()) {

            float widthOffset = std::min(widthRangeX, 9.0f);

            stream << R"D(
            <line x1=")D" << widthLocX + widthOffset-2.0f << R"D(" y1=")D" << widthLocY-4.0f << R"D(" x2=")D" << widthLocX + widthOffset+2.0f << R"D(" y2=")D" << widthLocY+4.0f << R"D(" class="bd_connection_width_line"/>
            <text x=")D" << widthLocX + widthOffset + 5.0f << R"D(" y=")D" << widthLocY+8.0f << R"D(" class="bd_connection_width">)D" << m_width << R"D(</text>
            )D";
        }
        if (!tooltip.empty())
            stream << "<title>" << tooltip << "</title>\n";
        stream << "</g>\n";
    }
}

void Connection::addConstraints()
{

}




Group::Group(BlockDiagram &blockDiagram) : StyledShapeBase<Group>(blockDiagram)
{
    m_style = "bd_group";
}

void Group::render(std::ostream &stream)
{
   stream << R"D(
    <rect x=")D" << m_globalX << R"D(" y=")D" << m_globalY << R"D(" width=")D" << m_width << R"D(" height=")D" << m_height << R"D("
          class=")D" << m_style << R"D(">
    )D";
    if (!m_tooltip.empty())
        stream << "<title>" << m_tooltip << "</title>\n";
    stream << "</rect>\n";
    renderSubShapes(stream);
}

ConnectionRouting::ConnectionRouting(BlockDiagram &blockDiagram) : ShapeBase<ConnectionRouting>(blockDiagram)
{
    m_minWidth = 0.0f;
    m_minHeight = 0.0f;
}

void ConnectionRouting::render(std::ostream &stream)
{
}




Block::Block(BlockDiagram &blockDiagram) : NamedShapeBase<Block>(blockDiagram)
{
    for (auto i : utils::Range(4))
        m_padding[i] = 10.0f;

    for (auto i : utils::Range(4))
        m_margin[i] = 10.0f;

    m_style = "bd_operation";
}


void Block::doSetTitle(std::string title)
{
    NamedShapeBase<Block>::doSetTitle(std::move(title));
    m_margin[NORTH] = 25.0f;
}


void Block::render(std::ostream &stream)
{

    stream << R"D(
    <g>
    <rect x=")D" << m_globalX << R"D(" y=")D" << m_globalY << R"D(" width=")D" << m_width << R"D(" height=")D" << m_height << R"D("
          class=")D" << m_style << R"D("/>
    )D";

    if (m_subShapes.empty()) {
        stream << R"D(
        <text dominant-baseline="middle" text-anchor="middle" x=")D" << m_globalX+m_width/2.0f << R"D(" y=")D" << m_globalY+m_height/2.0f << R"D(" class="bd_title">)D" << m_title << R"D(</text>
        )D";
    } else {
        stream << R"D(
        <text x=")D" << m_globalX+5.0f << R"D(" y=")D" << m_globalY+16.0f << R"D(" class="bd_title">)D" << m_title << R"D(</text>
        <line x1=")D" << m_globalX << R"D(" y1=")D" << m_globalY+20.0f << R"D(" x2=")D" << m_globalX+m_width << R"D(" y2=")D" << m_globalY+20.0f << R"D("
            class=")D" << m_style << R"D("/>
        )D";

    }
    if (!m_tooltip.empty())
        stream << "<title>" << m_tooltip << "</title>\n";
    stream << "</g>";
    renderSubShapes(stream);
}




Multiplexer::Multiplexer(BlockDiagram &blockDiagram, Side outputSide) : StyledShapeBase<Multiplexer>(blockDiagram), m_outputSide(outputSide)
{
    for (auto i : utils::Range(4))
        m_padding[i] = 10.0f;

    m_minWidth = 20.0f;
    m_minHeight = 20.0f;
    m_style = "bd_operation";
}

void Multiplexer::render(std::ostream &stream)
{

    float x[4];
    float y[4];

    x[0] = m_globalX;
    y[0] = m_globalY;

    x[1] = m_globalX + m_width;
    y[1] = m_globalY;

    x[2] = m_globalX + m_width;
    y[2] = m_globalY + m_height;

    x[3] = m_globalX;
    y[3] = m_globalY + m_height;

    switch (m_outputSide) {
        case NORTH:
            x[0] += m_width * 0.2f;
            x[1] -= m_width * 0.2f;
        break;
        case SOUTH:
            x[3] += m_width * 0.2f;
            x[2] -= m_width * 0.2f;
        break;
        case EAST:
            y[1] += m_height * 0.2f;
            y[2] -= m_height * 0.2f;
        break;
        case WEST:
            y[0] += m_height * 0.2f;
            y[3] -= m_height * 0.2f;
        break;
    }

    stream << R"D(
    <polygon points=")D";

    stream
        << x[0] << ' ' << y[0] << ' '
        << x[1] << ' ' << y[1] << ' '
        << x[2] << ' ' << y[2] << ' '
        << x[3] << ' ' << y[3];
    stream << R"D(" class=")D" << m_style << R"D(">
    )D";
    if (!m_tooltip.empty())
        stream << "<title>" << m_tooltip << "</title>\n";
    stream << "</polygon>\n";
    renderSubShapes(stream);
}



Register::Register(BlockDiagram &blockDiagram) : NamedShapeBase<Register>(blockDiagram)
{
    m_minHeight = 30.0f;
    m_style = "bd_operation";
}

void Register::render(std::ostream &stream)
{

    float clockHeightHalf = std::min(m_height/4.0f, 3.0f);
    float clockYPos = m_globalY + m_height - clockHeightHalf - 2.0f;
    float clockWidth = std::min(m_width, clockHeightHalf*1.5f);

    stream << R"D(
    <g>
    <rect x=")D" << m_globalX << R"D(" y=")D" << m_globalY << R"D(" width=")D" << m_width << R"D(" height=")D" << m_height << R"D("
          class=")D" << m_style << R"D("/>
    <path d="M )D"<<m_globalX<<' '<<clockYPos-clockHeightHalf<<" L "<<m_globalX+clockWidth<<' '<<clockYPos<<" L "<<m_globalX<<' '<<clockYPos+clockHeightHalf<<R"D(" class=")D"<<m_style<<R"D("/>
    <text dominant-baseline="middle" text-anchor="middle" x=")D" << m_globalX+m_width/2.0f << R"D(" y=")D" << m_globalY+m_height/2.0f << R"D(" class="bd_title">)D" << m_title << R"D(</text>
    )D";
    if (!m_tooltip.empty())
        stream << "<title>" << m_tooltip << "</title>\n";
    stream << "</g>\n";
}




Pin::Pin(BlockDiagram &blockDiagram) : StyledShapeBase<Pin>(blockDiagram)
{
    m_minWidth = 20.0f;
    m_minHeight = 20.0f;
    m_style = "bd_pin";
}


void Pin::render(std::ostream &stream)
{
    float r = m_width/2.0f;
    stream << R"D(
    <circle cx=")D" << m_globalX+r << R"D(" cy=")D" << m_globalY+r << R"D(" r=")D" << r << R"D(" class=")D" << m_style << R"D(">
    )D";
    if (!m_tooltip.empty())
        stream << "<title>" << m_tooltip << "</title>\n";
    stream << "</circle>\n";
}

std::tuple<float, float> Pin::getSideLocation(Side side, boost::rational<unsigned> loc)
{
    float alpha = ((size_t)side + loc.numerator() / (float) loc.denominator() - 0.5f) * boost::math::float_constants::pi / 2.0f;
    float r = m_width/2.0f;
    return {m_globalX+r+r*std::cos(alpha), m_globalY+r-r * std::sin(alpha)};
}



Canvas::Canvas(BlockDiagram &blockDiagram) : Shape(blockDiagram)
{
    for (auto i : utils::Range(4))
        m_margin[i] = 3.0f;
}

Connection &Canvas::connect(Shape *source, Shape::Side sourceSide, Shape *destination, Shape::Side destinationSide)
{
    m_connections.push_back(Connection(
        {
            .shape = source,
            .side = sourceSide,
            .loc = {1,2}
        },
        {
            .shape = destination,
            .side = destinationSide,
            .loc = {1,2}
        }
    ));
    return m_connections.back();
}

void Canvas::layout()
{
    bool done;
    unsigned iters = 0;
    do {
        HCL_ASSERT_HINT(++iters < 1000, "Tried really hard, but there seems to be a cyclic constraint that can't be satisfied!");
        done = true;

        Shape::layout();
        computeGlobalPositions();

        for (const auto &c : m_constraints)
            if (c->apply()) done = false;

    } while (!done);

    Shape::layout();
    computeGlobalPositions();
}

void Canvas::render(std::ostream &stream)
{
//    <svg width=")D" << m_width << R"D(" height=")D" << m_height << R"D(" xmlns="http://www.w3.org/2000/svg">
    stream << R"D(
    <svg viewbox="0 0 )D" << m_width << ' ' << m_height << R"D(" xmlns="http://www.w3.org/2000/svg">
    <style>
        .bd_background {
            fill:none;
            stroke:none;
        }
        .bd_operation {
            fill:#fff;
            stroke:black;
            stroke-width:1px;
        }
        .bd_entity {
            fill:#ddd;
            stroke:black;
            stroke-width:1px;
        }
        .bd_entity2 {
            fill:#eaeaea;
            stroke:black;
            stroke-width:1px;
            stroke-dasharray:3,3;
        }
        .bd_bram {
            fill:#bbe;
            stroke:black;
            stroke-width:1px;
        }

        .bd_connection {
            fill:none;
            stroke:black;
            stroke-width:1px;
            marker-end:url(#connection_head);
        }

        .bd_connection_no_head {
            fill:none;
            stroke:black;
            stroke-width:1px;
        }

        .bd_connection.important {
            stroke:red;
        }
        .bd_connection_no_head.important {
            stroke:red;
        }


        .bd_connection_width_line{
            stroke:gray;
            stroke-width:1px;
        }

        .bd_connection_width {
            font-size:8px;
            fill:#333;
        }

        .bd_connection_name {
            font-size:10px;
        }

        .bd_pin {
            fill:#fdd;
            stroke:black;
            stroke-width:1px;
        }

        .bd_group {
            fill:none;
            stroke:none;
        }

        .bd_title {
            font-size:15px;
        }
    </style>

    <defs>
        <marker id="connection_head" markerWidth="10" markerHeight="8" refX="10" refY="4" orient="auto">
            <polygon points="0 0, 10 4, 0 8, 1 4" fill="black"/>
        </marker>
    </defs>

    <rect x="0" y="0" width=")D" << m_width << R"D(" height=")D" << m_height << R"D("
          class="bd_background"/>
    )D";

    renderSubShapes(stream);

    for (auto &c : m_connections)
        c.render(stream);

    stream << R"D(
    </svg>
    )D";
}

void Canvas::addConnectionConstraints()
{
    for (auto &c : m_connections)
        c.addConstraints();
}


void BlockDiagram::write(std::ostream &stream)
{
    m_canvas.addConnectionConstraints();
    m_canvas.layout();
    m_canvas.computeGlobalPositions();
    m_canvas.render(stream);
}


}