/*
	Synogate Library, a support library for Synogate IP-core generators.
	Copyright (C) 2021 Synogate UG (haftungsbeschraenkt), www.synogate.com 
	
	This program is commercial software: you must not copy, modify or
	distribute it without explicit written license by Synogate to do so. 
	
	The full license text for non-commercial use and evaluation can be found in the distribution package of this core.
	
	This program is distributed WITHOUT ANY WARRANTY; without even the
	implied warranty of MERCHANTABILITY.
*/


#pragma once

#include <boost/rational.hpp>

#include <array>
#include <vector>
#include <list>
#include <memory>
#include <string>
#include <ostream>

namespace gtry::reporting {

class BlockDiagram;

class Shape {
    public:
        enum Side {
            EAST,
            NORTH,
            WEST,
            SOUTH,
        };


        Shape(BlockDiagram &blockDiagram);
        virtual ~Shape() = default;

        template<class ShapeType, typename... Args>
        ShapeType &add(Args&&... args) {
            ShapeType *res;
            addShape(std::unique_ptr<Shape>(res = new ShapeType(m_blockDiagram, std::forward<Args>(args)...)));
            return *res;
        }

        void addShape(std::unique_ptr<Shape> shape);

        virtual void layout();
        void computeGlobalPositions();

        virtual void render(std::ostream &stream) = 0;
        virtual void renderSubShapes(std::ostream &stream);

        Shape *findCommonParent(const Shape *other) const;

        virtual std::tuple<float, float> getSideLocation(Side side, boost::rational<unsigned> loc);

        Shape *getParent() { return m_parent; }

        void belowOf(Shape *other, float minDistance = 0.0f);
        void aboveOf(Shape *other, float minDistance = 0.0f);
        void leftOf(Shape *other, float minDistance = 0.0f);
        void rightOf(Shape *other, float minDistance = 0.0f);
        void vCenterOn(Shape *other);
        void hCenterOn(Shape *other);
        virtual void increaseMinWidthHeight(float width, float height) { m_minWidth = std::max(m_minWidth, width); m_minHeight = std::max(m_minHeight, height); }
        void setToolTip(std::string tooltip) { m_tooltip = std::move(tooltip); }

    protected:
        BlockDiagram &m_blockDiagram;

        float m_globalX = 0.0f;
        float m_globalY = 0.0f;
        float m_x = 0.0f;
        float m_y = 0.0f;
        float m_width = 1.0f;
        float m_height = 1.0f;
        float m_minWidth = 1.0f;
        float m_minHeight = 1.0f;
        std::array<float, 4> m_padding = {};
        std::array<float, 4> m_margin = {};
        Shape *m_parent = nullptr;

        std::vector<std::unique_ptr<Shape>> m_subShapes;

        std::string m_tooltip;

        friend class Canvas;
        friend class DistanceConstraint;
        friend class AnchoringConstraint;

};

template<class FinalType>
class ShapeBase : public Shape
{
    public:
        ShapeBase(BlockDiagram &blockDiagram) : Shape(blockDiagram) { }

        FinalType &setPadding(std::array<float, 4> padding) { m_padding = padding; return (FinalType&)*this; }
        FinalType &setMargin(std::array<float, 4> margin) { m_margin = margin; return (FinalType&)*this; }
        FinalType &belowOf(Shape *other, float minDistance = 0.0f) { Shape::belowOf(other, minDistance); return (FinalType&)*this; }
        FinalType &aboveOf(Shape *other, float minDistance = 0.0f) { Shape::aboveOf(other, minDistance); return (FinalType&)*this; }
        FinalType &leftOf(Shape *other, float minDistance = 0.0f) { Shape::leftOf(other, minDistance); return (FinalType&)*this; }
        FinalType &rightOf(Shape *other, float minDistance = 0.0f) { Shape::rightOf(other, minDistance); return (FinalType&)*this; }
        FinalType &vCenterOn(Shape *other) { Shape::vCenterOn(other); return (FinalType&)*this; }
        FinalType &hCenterOn(Shape *other) { Shape::hCenterOn(other); return (FinalType&)*this; }

        FinalType &setPos(float x, float y) { m_x = x; m_y = y; return (FinalType&)*this; }

        FinalType &incMinWidthHeight(float width, float height) { increaseMinWidthHeight(width, height); return (FinalType&)*this; }

        FinalType &setToolTip(std::string tooltip) { Shape::setToolTip(std::move(tooltip)); return (FinalType&)*this; }
};

template<class FinalType>
class StyledShapeBase : public ShapeBase<FinalType>
{
    public:
        StyledShapeBase(BlockDiagram &blockDiagram) : ShapeBase<FinalType>(blockDiagram) { }

        FinalType &setStyle(std::string style) { m_style = std::move(style); return (FinalType&)*this; }
    protected:
        std::string m_style;
};

template<class FinalType>
class NamedShapeBase : public StyledShapeBase<FinalType>
{
    public:
        using ancestor = StyledShapeBase<FinalType>;
        NamedShapeBase(BlockDiagram &blockDiagram) : StyledShapeBase<FinalType>(blockDiagram) { }

        virtual void doSetTitle(std::string title) {
            m_title = std::move(title);
            ancestor::m_minWidth = std::max(ancestor::m_minWidth, m_title.length() * 10.0f);
            ancestor::m_minHeight = std::max(ancestor::m_minHeight, 20.0f);
        }

        FinalType &setTitle(std::string title) { doSetTitle(std::move(title)); return (FinalType&)*this; }
    protected:
        std::string m_title;
};



class LayoutConstraint {
    public:
        virtual ~LayoutConstraint() = default;
        virtual bool apply() = 0;
};

class DistanceConstraint : public LayoutConstraint {
    public:
        DistanceConstraint(Shape *shape1, Shape *shape2, Shape::Side direction, float minDistance);
        virtual bool apply() override;
    protected:
        Shape *m_shapes[2];
        Shape *m_commonParent;
        Shape::Side m_direction;
        float m_minDistance;
};

class AnchoringConstraint : public LayoutConstraint {
    public:
        AnchoringConstraint(Shape *movingShape, boost::rational<unsigned> movingShapeLoc, Shape *anchorShape, boost::rational<unsigned> anchoringShapeLoc, Shape::Side direction);
        virtual bool apply() override;
    protected:
        Shape *m_movingShape;
        boost::rational<unsigned> m_movingShapeLoc;
        Shape *m_anchorShape;
        boost::rational<unsigned> m_anchoringShapeLoc;
        Shape::Side m_direction;

        Shape *m_refFrame;
};


class Connection {
    public:
        struct Port {
            Shape *shape;
            Shape::Side side;
            boost::rational<unsigned> loc;
        };
        Connection(Port source, Port destination);

        Connection &setSrcLoc(boost::rational<unsigned> loc) { m_sourcePort.loc = loc; return *this; }
        Connection &setDstLoc(boost::rational<unsigned> loc) { m_destinationPort.loc = loc; return *this; }

        void render(std::ostream &stream);
        void addConstraints();

        Connection &setName(std::string name) { m_name = std::move(name); return *this; }
        Connection &setStyle(std::string style) { m_style = std::move(style); return *this; }
        Connection &setWidth(std::string width) { m_width = std::move(width); return *this; }
        Connection &setToolTip(std::string tooltip) { m_tooltip = std::move(tooltip); return *this; }
    protected:
        Port m_sourcePort;
        Port m_destinationPort;
        std::string m_name;
        std::string m_style = "bd_connection";
        std::string m_width;
        std::string m_tooltip;
};

/*
class ConnectionGroup {
    public:
    protected:
        std::vector<Connection*> m_connections;
};
*/

class Group : public StyledShapeBase<Group> {
    public:
        Group(BlockDiagram &blockDiagram);

        virtual void render(std::ostream &stream) override;
};

class ConnectionRouting : public ShapeBase<ConnectionRouting> {
    public:
        ConnectionRouting(BlockDiagram &blockDiagram);

        ConnectionRouting &setHorizontal(float v) { m_minWidth = v; m_minHeight = 0; return *this; }
        ConnectionRouting &setVertical(float v) { m_minWidth = 0; m_minHeight = v; return *this; }

        virtual void render(std::ostream &stream) override;
    protected:
};


class Block : public NamedShapeBase<Block> {
    public:
        Block(BlockDiagram &blockDiagram);

        virtual void doSetTitle(std::string title) override;
        virtual void render(std::ostream &stream) override;
};


class Multiplexer : public StyledShapeBase<Multiplexer> {
    public:
        Multiplexer(BlockDiagram &blockDiagram, Side outputSide);

        virtual void render(std::ostream &stream) override;
    protected:
        Side m_outputSide;
};


class Register : public NamedShapeBase<Register> {
    public:
        Register(BlockDiagram &blockDiagram);

        virtual void render(std::ostream &stream) override;
};


class Pin : public StyledShapeBase<Pin> {
    public:
        Pin(BlockDiagram &blockDiagram);

        virtual void increaseMinWidthHeight(float width, float height) override {
            m_minWidth = m_minHeight = std::max(m_minWidth, std::max(width, height));
        }

        virtual void render(std::ostream &stream) override;
        virtual std::tuple<float, float> getSideLocation(Side side, boost::rational<unsigned> loc) override;
};



class Canvas : public Shape
{
    public:
        Canvas(BlockDiagram &blockDiagram);
        virtual void layout() override;
        virtual void render(std::ostream &stream) override;

        Connection &connect(Shape *source, Shape::Side sourceSide, Shape *destination, Shape::Side destinationSide);

        template<class ConstraintType, typename... Args>
        ConstraintType &addConstraint(Args&&... args) {
            ConstraintType *res;
            m_constraints.push_back(std::unique_ptr<LayoutConstraint>(res = new ConstraintType(std::forward<Args>(args)...)));
            return *res;
        }

        inline std::list<Connection> &getConnections() { return m_connections; }
    protected:
        std::vector<std::unique_ptr<LayoutConstraint>> m_constraints;
        std::list<Connection> m_connections;

        void addConnectionConstraints();
        friend class BlockDiagram;

};

class BlockDiagram
{
    public:
        BlockDiagram() : m_canvas(*this) { }

        Canvas &canvas() { return m_canvas; }

        void write(std::ostream &stream);
    protected:
        Canvas m_canvas;
};

}