/*
	Synogate Library, a support library for Synogate IP-core generators.
	Copyright (C) 2021 Synogate UG (haftungsbeschraenkt), www.synogate.com 
	
	This program is commercial software: you must not copy, modify or
	distribute it without explicit written license by Synogate to do so. 
	
	The full license text for non-commercial use and evaluation can be found in the distribution package of this core.
	
	This program is distributed WITHOUT ANY WARRANTY; without even the
	implied warranty of MERCHANTABILITY.
*/

#include "InterfaceDocumentation.h"

#include "Page.h"
#include "Waveform.h"

#include <gatery/hlim/Node.h>
#include <gatery/simulation/waveformFormats/MemoryTrace.h>
#include <gatery/hlim/ClockRational.h>

#include <regex>


namespace gtry::reporting {


Section &InterfaceDocumentation::generateDefaultMainPage(const std::string ipCoreName)
{
    auto &page = createPage(ipCoreName + " - IP Interface Description", "index.html");
    {
        auto &sec = page.addSection("Introduction", "main_intro");
        sec << "This is the generated interface documentation of the " << ipCoreName << '.';
        sec << R"Delim(
            For a detailed explanation of this core's functionality, its generation and what parameters are available, see the accompanying IP User Guide.
            This document only describes characteristics of the core specific to the chosen configuration such as registers, timings, etc.
            For reference, the chosen configuration options can be reviewed in <<<Section main_options>>>.
        )Delim";
    }
    m_tocSection = &page.addSection("Table of Contents", "main_toc");
    m_LOFSection[0] = &page.addSection("List of Tables", "main_lot");
    m_LOFSection[1] = &page.addSection("List of Waveforms", "main_low");
    m_LOFSection[2] = &page.addSection("List of Figures", "main_lof");

    return page.addSection("Configuration Options", "main_options");
}

void InterfaceDocumentation::generate()
{
	requireResource("images/header.png");
	requireResource("images/logo_white.svg");
	requireResource("images/exist/BMAS.svg");
	requireResource("images/exist/ESF.svg");
	requireResource("images/exist/EU_Logo.svg");
	requireResource("images/exist/exist.svg");
	requireResource("images/exist/zzg.svg");
	requireResource("css/interfaceDocu.css");
	requireResource("extern/assets/css/fontawesome-all.min.css");
	requireResource("extern/assets/css/main.css");
	requireResource("extern/assets/css/noscript.css");
	requireResource("extern/assets/js/breakpoints.min.js");
	requireResource("extern/assets/js/browser.min.js");
	requireResource("extern/assets/js/jquery.dropotron.min.js");
	requireResource("extern/assets/js/jquery.min.js");
	requireResource("extern/assets/js/jquery.scrollex.min.js");
	requireResource("extern/assets/js/jquery.scrolly.min.js");
	requireResource("extern/assets/js/main.js");
	requireResource("extern/assets/js/util.js");
	requireResource("extern/assets/images/bg02.png");
	requireResource("extern/assets/webfonts/fa-solid-900.woff");
	requireResource("extern/assets/webfonts/fa-solid-900.woff2");
	requireResource("extern/assets/webfonts/fa-solid-900.ttf");

    Report::generate();
}

void InterfaceDocumentation::generatePageHeader(std::ostream &stream, const Page *page)
{
    stream << R"Delim(<!DOCTYPE HTML>
<!--
	Helios by HTML5 UP
	html5up.net | @ajlkn
	Free for personal and commercial use under the CCA 3.0 license (html5up.net/license)
-->
<html>
	<head>
		<title>)Delim" << page->getName() << R"Delim(</title>
		<meta charset="utf-8" />
		<meta name="viewport" content="width=device-width, initial-scale=1, user-scalable=no" />
		<link rel="stylesheet" href="extern/assets/css/main.css" />
		<link rel="stylesheet" href="css/interfaceDocu.css" />
		<noscript><link rel="stylesheet" href="extern/assets/css/noscript.css" /></noscript>
	</head>
	<body class="left-sidebar is-preload">
		<div id="page-wrapper">

			<!-- Header -->
				<div id="header">

					<!-- Inner -->
						<div class="inner">
							<header>
                                <h1><a href="index.html" id="logo"><img style="width:300px" src="images/logo_white.svg" alt="logo" /></a></h1>
							</header>
						</div>

					<!-- Nav -->
						<nav id="nav">
							<ul>
)Delim";

        for (auto &page : m_pages)
			buildTableOfContents(stream, &page, 3);

    stream << R"Delim(
							</ul>
						</nav>
				</div>
)Delim";
}

void InterfaceDocumentation::generatePageFooter(std::ostream &stream, const Page *page)
{
    stream << R"Delim(

			<!-- Footer -->
				<div id="footer">
					<div class="container">
						<div class="row">
							<div class="col-3 col-12-narrow">

								<!-- Links -->
									<section>
										<h2>Navigation</h2>
										<ul class="divided">
)Delim";

        for (auto &page : m_pages)
			buildTableOfContents(stream, &page, 0);

    stream << R"Delim(
										</ul>
									</section>
							</div>
							<div class="col-9 col-12-narrow imp-medium">

								<!-- Contact -->
									<section>
										<h2>Get in touch</h2>
										<div>
											<div class="row">
												<div class="col-6 col-12-mobile">
													<dl class="contact">
														<dt>Github</dt>
														<dd><a href="https://github.com/synogate">https://github.com/synogate</a></dd>
														<dt>WWW</dt>
														<dd><a href="https://synogate.com">https://synogate.com</a></dd>
														<dt>Email</dt>
														<dd><a href="mailto:mail@synogate.com">mail@synogate.com</a></dd>
													</dl>
												</div>
												<div class="col-6 col-12-mobile">
													<dl class="contact">
														<dt>Address</dt>
														<dd>
															Wegedornstr. 32<br />
															12524 Berlin<br />
															Germany
														</dd>
														<dt>Phone</dt>
														<dd>+49 30 209346601</dd>
													</dl>
												</div>
											</div>
										</div>
									</section>
							</div>
							<div class="col-12">
								<div class="row exist_white_background">
									<div class="col-2 col-4-narrow col-6-mobile">
										<img src="images/exist/BMAS.svg" width="100%" alt="exist-logo" />
									</div>
									<div class="col-2 col-4-narrow col-6-mobile">
										<img src="images/exist/exist.svg" width="100%" alt="exist-logo" />
									</div>
									<div class="col-2 col-4-narrow col-6-mobile">
										<img src="images/exist/ESF.svg" width="100%" alt="exist-logo" />
									</div>
									<div class="col-2 col-4-narrow col-6-mobile">
										<img src="images/exist/EU_Logo.svg" width="80%" alt="exist-logo" />
									</div>
									<div class="col-4 col-6-narrow col-10-mobile">
										<img src="images/exist/zzg.svg" width="100%" alt="exist-logo" />
									</div>
								</div>
								Synogate wurde im Rahmen des EXIST-Programms in 2021 durch das Bundesministerium für Wirtschaft und Energie und den Europäischen Sozialfonds gefördert.
								Ziel der Europäischen Union ist es, dass alle Menschen eine berufliche Perspektive
								erhalten. Der Europäische Sozialfonds (ESF) verbessert die Beschäftigungschancen,
								unterstützt die Menschen durch Ausbildung und Qualifizierung und trägt zum Abbau
								von Benachteiligungen auf dem Arbeitsmarkt bei. Mehr zum ESF unter: <a href="https://www.esf.de">www.esf.de</a>
							</div>
							<div class="col-12">
								<div id="copyright">
									<ul class="menu">
										<li>Design: <a href="http://html5up.net">HTML5 UP</a></li>
									</ul>
								</div>
							</div>
						</div>
					</div>
				</div>

		</div>
		<!-- Scripts -->
			<script src="extern/assets/js/jquery.min.js"></script>
			<script src="extern/assets/js/jquery.dropotron.min.js"></script>
			<script src="extern/assets/js/jquery.scrolly.min.js"></script>
			<script src="extern/assets/js/jquery.scrollex.min.js"></script>
			<script src="extern/assets/js/browser.min.js"></script>
			<script src="extern/assets/js/breakpoints.min.js"></script>
			<script src="extern/assets/js/util.js"></script>
			<script src="extern/assets/js/main.js"></script>


	</body>
</html>
)Delim";
}

void InterfaceDocumentation::buildSidebarPage(std::ostream &stream, const Page *page)
{
    stream << R"Delim(
			<!-- Main -->
				<div class="wrapper style1">

					<div class="container">
						<div class="row gtr-200">
							<div class="col-3 col-12-mobile" id="sidebar">
								<hr class="first" />
								<section>
									<header>
										<h3>TOC</h3>
									</header>
									<div class="sidebar_toc_container">
									)Delim";
	buildSidebarTableOfContents(stream, page);
    stream << R"Delim(
									</div>
								</section>
								<hr />
								<section>
									<header>
										<h3>List of figures</h3>
									</header>
									<div class="sidebar_low_container">
									)Delim";
	buildSidebarListOfFloats(stream, page, [](const Float *ptr)->bool{ return dynamic_cast<const Figure*>(ptr);});
    stream << R"Delim(
									</div>
								</section>
								<hr />
								<section>
									<header>
										<h3>List of waveforms</h3>
									</header>
									<div class="sidebar_low_container">
									)Delim";
	buildSidebarListOfFloats(stream, page, [](const Float *ptr)->bool{ return dynamic_cast<const Waveform*>(ptr);});
    stream << R"Delim(
									</div>
								</section>
								<hr />
								<section>
									<header>
										<h3>List of tables</h3>
									</header>
									<div class="sidebar_lot_container">
									)Delim";
	buildSidebarListOfFloats(stream, page, [](const Float *ptr)->bool{ return dynamic_cast<const Table*>(ptr);});
    stream << R"Delim(
									</div>
								</section>
							</div>
							<div class="col-9 col-12-mobile imp-mobile" id="content">
								<article id="main">
									<header>
										<h2><a href="#">)Delim" << page->getName() << R"Delim(</a></h2>
									</header>
)Delim";

    Report::generateSection(stream, page, 0);

    stream << R"Delim(
								</article>
							</div>
						</div>
					</div>
				</div>
)Delim";
}

void InterfaceDocumentation::buildTableOfContents(std::ostream &stream, const Page *page, unsigned maxDepth)
{
	std::function<void(const Section*, unsigned)> reccurPrintSections;
	reccurPrintSections = [&](const Section *section, unsigned depth) {
		if (dynamic_cast<const Page*>(section))
			stream << "<li class=\"toc_page\"><a href=\""<<page->getLabel()<<"\">" << page->getName() << "</a>\n";
		else
			stream << "<li class=\"toc_section\"><a href=\"" << m_refs[section].href << "\">" << m_refs[section].label << ": " << section->getName() << "</a>";
		if (!section->getSubSections().empty() && depth < maxDepth) {
			stream << "<ul>\n";
			for (auto &subsec : section->getSubSections())
				reccurPrintSections(&subsec, depth+1);
			stream << "</ul>\n";
		}
		stream << "</li>\n";
	};
	reccurPrintSections(page, 0);
}

void InterfaceDocumentation::buildSidebarTableOfContents(std::ostream &stream, const Page *page)
{
	stream << "<ul>";

	if (page == nullptr) {
		for (auto &page : m_pages)
			buildTableOfContents(stream, &page, 3);
	} else
		buildTableOfContents(stream, page, 3);

	stream << "</ul>";
}

void InterfaceDocumentation::buildListOfFloats(std::ostream &stream, const Page *page, std::function<bool(const Float*)> floatSelector)
{
    std::function<void(const Section*)> reccurPrint;
    reccurPrint = [&](const Section *section) {
        for (auto &subsection : section->getSubSections())
            reccurPrint(&subsection);

        for (const auto &float_ : section->getFloats())
			if (floatSelector(float_.get())) {
				auto it = m_refs.find(float_.get());
				HCL_ASSERT(it != m_refs.end());
				stream << "<li class=\"lof_float\"><a href=\"" << it->second.href << "\">" << it->second.label << ": " << float_->getName() << "</a></li>\n";
			}
    };
	reccurPrint(page);
}


void InterfaceDocumentation::buildSidebarListOfFloats(std::ostream &stream, const Page *page, std::function<bool(const Float*)> floatSelector)
{
	stream << "<ul>";
    buildListOfFloats(stream, page, floatSelector);
	stream << "</ul>";
}



void InterfaceDocumentation::generatePage(std::ostream &stream, const Page *page)
{
    generatePageHeader(stream, page);
    buildSidebarPage(stream, page);
    generatePageFooter(stream, page);
}

void InterfaceDocumentation::generateSection(std::ostream &stream, const Section *section, unsigned depth)
{
    stream << "<section>\n";
    Report::generateSection(stream, section, depth);
    stream << "</section>\n";

    if (section == m_tocSection) {
        stream << "<div class=\"toc_container\"><ul>";

        for (auto &page : m_pages)
			buildTableOfContents(stream, &page, 10);

        stream << "</ul></div>";
    } else
    if (section == m_LOFSection[0]) {
        stream << "<div class=\"lof_container\"><ul>";
        for (auto &page : m_pages) {
			auto it = m_refs.find(&page);
			HCL_ASSERT(it != m_refs.end());
			stream << "<li class=\"lof_page\">" << it->second.label << ": " << page.getName() << "<ul>\n";
			buildListOfFloats(stream, &page, [](const Float *ptr)->bool{ return dynamic_cast<const Table*>(ptr);});
			stream << "</ul></li>";
		}
        stream << "</ul></div>";
    } else
    if (section == m_LOFSection[1]) {
        stream << "<div class=\"lof_container\"><ul>";
        for (auto &page : m_pages) {
			auto it = m_refs.find(&page);
			HCL_ASSERT(it != m_refs.end());
			stream << "<li class=\"lof_page\">" << it->second.label << ": " << page.getName() << "<ul>\n";
			buildListOfFloats(stream, &page, [](const Float *ptr)->bool{ return dynamic_cast<const Waveform*>(ptr);});
			stream << "</ul></li>";
		}
        stream << "</ul></div>";
    } else
    if (section == m_LOFSection[2]) {
        stream << "<div class=\"lof_container\"><ul>";
        for (auto &page : m_pages) {
			auto it = m_refs.find(&page);
			HCL_ASSERT(it != m_refs.end());
			stream << "<li class=\"lof_page\">" << it->second.label << ": " << page.getName() << "<ul>\n";
			buildListOfFloats(stream, &page, [](const Float *ptr)->bool{ return dynamic_cast<const Figure*>(ptr);});
			stream << "</ul></li>";
		}
        stream << "</ul></div>";
    }
}

void InterfaceDocumentation::generateWaveformContent(std::ostream &stream, const Waveform *waveform)
{
	if (waveform->getTrace() == nullptr) return;

	auto startTime = waveform->getTrace()->events.front().timestamp;
	auto endTime = waveform->getTrace()->events.back().timestamp;

	if (!waveform->getOptions().selectionAnnotationId.empty()) {

		auto it = waveform->getTrace()->annotations.find(waveform->getOptions().selectionAnnotationId);
		HCL_DESIGNCHECK_HINT(it != waveform->getTrace()->annotations.end(), "The given annotation ID for the waveform is not present in the trace");

		startTime = it->second.ranges.front().start;
		endTime = it->second.ranges.back().end;
	}

	HCL_ASSERT_HINT(!waveform->getOptions().collapse, "Not implemented yet!");

	struct RenderedSignalChange {
		hlim::ClockRational time;
		size_t state;
	};

	struct RenderedSignal {
		size_t id;
		bool isClock;
		Waveform::Style style;
		size_t width;
		std::vector<RenderedSignalChange> changes;
	};

	std::vector<RenderedSignal> renderedSignals;

	for (const auto &s : waveform->getTraceSignals()) {
		std::regex regex(s.n);
		for (auto i : utils::Range(waveform->getTrace()->signals.size())) {
			const auto &sig = waveform->getTrace()->signals[i];
			if (std::regex_match(sig.name, regex)) {
				renderedSignals.push_back({.id=i, .isClock=(sig.clock!=nullptr), .style=s.style, .width = sig.width});
			}
		}
	}

	std::vector<hlim::ClockRational> verticalRulers;

	for (auto &sig : renderedSignals) {
		auto id = sig.id;

		for (const auto &event : waveform->getTrace()->events) {
			if (event.timestamp > endTime) break;
			for (const auto &change : event.changes)
				if (change.sigIdx == id) {
					if (event.timestamp < startTime && sig.changes.size() > 0)
						sig.changes.back() = {.time = event.timestamp, .state=change.dataOffset};
					else
						sig.changes.push_back({.time = event.timestamp, .state=change.dataOffset});

					if (event.timestamp > startTime && (sig.style.verticalRulerFallingEdge || sig.style.verticalRulerRisingEdge)) {
						bool makeRuler = false;
						if (sig.width != 1)
							makeRuler = true;
						else {
							bool defined = waveform->getTrace()->data.get(sim::DefaultConfig::DEFINED, sig.changes.back().state);
							if (!defined)
								makeRuler = true;
							else {
								bool value = waveform->getTrace()->data.get(sim::DefaultConfig::VALUE, sig.changes.back().state);
								if ((value && sig.style.verticalRulerRisingEdge) || (!value && sig.style.verticalRulerFallingEdge))
									makeRuler = true;
							}
						}

						if (makeRuler)
							verticalRulers.push_back(event.timestamp);
					}
				}
		}
	}





	size_t numSignals = renderedSignals.size();
	size_t signalHeight = 35;

	size_t timelineHeight = 40;
	size_t legendWidth = 175;

	size_t vbH = numSignals * signalHeight + timelineHeight + 5;

	auto time2x = [&](hlim::ClockRational t)-> float {
		if (t < startTime)
			return float(legendWidth);
		if (t > endTime)
			return 1000;
		auto deltaT = t - startTime;
		auto t_frac = deltaT / (endTime - startTime);
		return t_frac.numerator() / (float) t_frac.denominator() * (1000 - legendWidth) + legendWidth;
	};




	stream << R"D(
		<svg id=")D" << waveform->getLabel() << R"D(_svg" viewbox="0 0 1000 )D" << vbH <<R"D(" xmlns="http://www.w3.org/1500/svg">
			<rect class="wf_background" width="100%" height="100%"/>
			<clipPath id=")D"<<waveform->getLabel()<<R"D(_nameBox">
				<rect x="0" y="0" width=")D"<<legendWidth<<R"D(" height="100%"/>
			</clipPath>
			<text clip-path="url(#)D"<<waveform->getLabel()<<R"D(_nameBox)" class="wf_signalName" x="0" y="20">Time:</text>
			<text clip-path="url(#)D"<<waveform->getLabel()<<R"D(_nameBox)" class="wf_signalValue" x="20" y="37" id="time_label">N/A</text>
	)D" << std::endl;



	for (auto i : utils::Range(renderedSignals.size())) {
		size_t y = timelineHeight + i * signalHeight;
		stream << R"D(
			<rect id=")D" << waveform->getLabel() << R"D(_bg_)D" << i << R"D(" class="wf_signalBackground" x="0" y=")D" << y << R"D(" width="100%" height=")D" << signalHeight << R"D("/>
		)D" << std::endl;
	}


	for (const auto &v : verticalRulers) {
		float x = time2x(v);
		stream << R"D(<line x1=")D"<<x<<R"D(" y1="0" x2=")D"<<x<<R"D(" y2="100%" class="wf_verticalRuler"/>
		)D";
	}

	size_t clipPathId = 0;


	auto formatValue = [&stream, waveform](size_t state, size_t width, const Waveform::Style &style) {
		switch (style.base) {
			case 16:
				stream << "0x";
			break;
			case 10:
			break;
			case 2:
				stream << "0b";
			break;
			default:
				HCL_ASSERT_HINT(false, "Not yet implemeted!");
		}
		if (state != ~0ull)
			sim::formatRange(stream, waveform->getTrace()->data, style.base, state, width);
		else {
			size_t logBase = utils::Log2C(style.base);
			for ([[maybe_unused]] auto i : utils::Range((width+logBase-1)/logBase))
				stream << 'X';
		}
	};


	auto drawSegment = [&formatValue, &stream, waveform, signalHeight, &clipPathId](float y, float x1, float x2, size_t state, size_t width, bool isClock, const Waveform::Style &style, bool drawLeftNode, bool drawRightNode) {
		const float transitionHalfWidth = 2.5f;
		float y_high = y - signalHeight/2 * 0.7f;
		float y_low = y + signalHeight/2 * 0.7f;

		size_t logBase = utils::Log2C(style.base);
		size_t approxLabelChars = 2 + (width + logBase - 1)/logBase;
		bool drawLabel = style.showLabel;
		if ((x2 - x1) < 7.0f * approxLabelChars)
			drawLabel = false;
		bool undefined = (state == ~0ull ) || !sim::allDefined(waveform->getTrace()->data, state, width);
		bool drawBlock = undefined || (width > 1);

		bool drawAngledNodes = style.angledTransitions;
		if ((x2 - x1) < 2*transitionHalfWidth)
			drawAngledNodes = false;
		if (!drawLeftNode && !drawRightNode)
			drawAngledNodes = false;

		std::string lineStyle, areaStyle;
		if (undefined) {
			lineStyle = "wf_undefined_line";
			areaStyle = "wf_undefined_area";
		} else {
			lineStyle = style.cssPrefix+"_line";
			areaStyle = style.cssPrefix+"_area";
		}

		if (drawBlock) {
			if (!drawAngledNodes) {
				stream << R"D(
					<polygon points=")D"<<x1<<R"D(,)D"<<y_high<<R"D(
									)D"<<x2<<R"D(,)D"<<y_high<<R"D(
									)D"<<x2<<R"D(,)D"<<y_low<<R"D(
									)D"<<x1<<R"D(,)D"<<y_low<<R"D("
									class=")D"<<areaStyle<<R"D(" />
					<line x1=")D"<<x1<<R"D(" y1=")D"<<y_high<<R"D(" x2=")D"<<x2<<R"D(" y2=")D"<<y_high<<R"D(" class=")D"<<lineStyle<<R"D(" />
					<line x1=")D"<<x1<<R"D(" y1=")D"<<y_low<<R"D(" x2=")D"<<x2<<R"D(" y2=")D"<<y_low<<R"D(" class=")D"<<lineStyle<<R"D(" />
					)D";
				if (drawLeftNode)
					stream << R"D(
						<line x1=")D"<<x1<<R"D(" y1=")D"<<y_low<<R"D(" x2=")D"<<x1<<R"D(" y2=")D"<<y_high<<R"D(" class=")D"<<lineStyle<<R"D(" />
					)D";

				if (drawRightNode)
					stream << R"D(
						<line x1=")D"<<x2<<R"D(" y1=")D"<<y_low<<R"D(" x2=")D"<<x2<<R"D(" y2=")D"<<y_high<<R"D(" class=")D"<<lineStyle<<R"D(" />
					)D";

			} else {
				if (!drawLeftNode) {
					stream << R"D(
						<polygon points=")D"<<x1<<R"D(,)D"<<y_high<<R"D(
										)D"<<x2-transitionHalfWidth<<R"D(,)D"<<y_high<<R"D(
										)D"<<x2<<R"D(,)D"<<y<<R"D(
										)D"<<x2-transitionHalfWidth<<R"D(,)D"<<y_low<<R"D(
										)D"<<x1<<R"D(,)D"<<y_low<<R"D("
										class=")D"<<areaStyle<<R"D(" />
						<line x1=")D"<<x2<<R"D(" y1=")D"<<y<<R"D(" x2=")D"<<x2-transitionHalfWidth<<R"D(" y2=")D"<<y_high<<R"D(" class=")D"<<lineStyle<<R"D(" />
						<line x1=")D"<<x2<<R"D(" y1=")D"<<y<<R"D(" x2=")D"<<x2-transitionHalfWidth<<R"D(" y2=")D"<<y_low<<R"D(" class=")D"<<lineStyle<<R"D(" />
						<line x1=")D"<<x1<<R"D(" y1=")D"<<y_high<<R"D(" x2=")D"<<x2-transitionHalfWidth<<R"D(" y2=")D"<<y_high<<R"D(" class=")D"<<lineStyle<<R"D(" />
						<line x1=")D"<<x1<<R"D(" y1=")D"<<y_low<<R"D(" x2=")D"<<x2-transitionHalfWidth<<R"D(" y2=")D"<<y_low<<R"D(" class=")D"<<lineStyle<<R"D(" />
						)D";
				} else if (!drawRightNode) {
					stream << R"D(
						<polygon points=")D"<<x1<<R"D(,)D"<<y<<R"D(
										)D"<<x1+transitionHalfWidth<<R"D(,)D"<<y_high<<R"D(
										)D"<<x2<<R"D(,)D"<<y_high<<R"D(
										)D"<<x2<<R"D(,)D"<<y_low<<R"D(
										)D"<<x1+transitionHalfWidth<<R"D(,)D"<<y_low<<R"D("
										class=")D"<<areaStyle<<R"D(" />
						<line x1=")D"<<x1<<R"D(" y1=")D"<<y<<R"D(" x2=")D"<<x1+transitionHalfWidth<<R"D(" y2=")D"<<y_high<<R"D(" class=")D"<<lineStyle<<R"D(" />
						<line x1=")D"<<x1<<R"D(" y1=")D"<<y<<R"D(" x2=")D"<<x1+transitionHalfWidth<<R"D(" y2=")D"<<y_low<<R"D(" class=")D"<<lineStyle<<R"D(" />
						<line x1=")D"<<x1+transitionHalfWidth<<R"D(" y1=")D"<<y_high<<R"D(" x2=")D"<<x2<<R"D(" y2=")D"<<y_high<<R"D(" class=")D"<<lineStyle<<R"D(" />
						<line x1=")D"<<x1+transitionHalfWidth<<R"D(" y1=")D"<<y_low<<R"D(" x2=")D"<<x2<<R"D(" y2=")D"<<y_low<<R"D(" class=")D"<<lineStyle<<R"D(" />
						)D";
				} else {
					stream << R"D(
						<polygon points=")D"<<x1<<R"D(,)D"<<y<<R"D(
										)D"<<x1+transitionHalfWidth<<R"D(,)D"<<y_high<<R"D(
										)D"<<x2-transitionHalfWidth<<R"D(,)D"<<y_high<<R"D(
										)D"<<x2<<R"D(,)D"<<y<<R"D(
										)D"<<x2-transitionHalfWidth<<R"D(,)D"<<y_low<<R"D(
										)D"<<x1+transitionHalfWidth<<R"D(,)D"<<y_low<<R"D("
										class=")D"<<areaStyle<<R"D(" />
						<line x1=")D"<<x1<<R"D(" y1=")D"<<y<<R"D(" x2=")D"<<x1+transitionHalfWidth<<R"D(" y2=")D"<<y_high<<R"D(" class=")D"<<lineStyle<<R"D(" />
						<line x1=")D"<<x1<<R"D(" y1=")D"<<y<<R"D(" x2=")D"<<x1+transitionHalfWidth<<R"D(" y2=")D"<<y_low<<R"D(" class=")D"<<lineStyle<<R"D(" />
						<line x1=")D"<<x2<<R"D(" y1=")D"<<y<<R"D(" x2=")D"<<x2-transitionHalfWidth<<R"D(" y2=")D"<<y_high<<R"D(" class=")D"<<lineStyle<<R"D(" />
						<line x1=")D"<<x2<<R"D(" y1=")D"<<y<<R"D(" x2=")D"<<x2-transitionHalfWidth<<R"D(" y2=")D"<<y_low<<R"D(" class=")D"<<lineStyle<<R"D(" />
						<line x1=")D"<<x1+transitionHalfWidth<<R"D(" y1=")D"<<y_high<<R"D(" x2=")D"<<x2-transitionHalfWidth<<R"D(" y2=")D"<<y_high<<R"D(" class=")D"<<lineStyle<<R"D(" />
						<line x1=")D"<<x1+transitionHalfWidth<<R"D(" y1=")D"<<y_low<<R"D(" x2=")D"<<x2-transitionHalfWidth<<R"D(" y2=")D"<<y_low<<R"D(" class=")D"<<lineStyle<<R"D(" />
						)D";
				}
			}
		} else {
			bool high = waveform->getTrace()->data.get(sim::DefaultConfig::VALUE, state);
			float y_v = high?y_high:y_low;
			if (!drawAngledNodes) {
				stream << R"D(
					<line x1=")D"<<x1<<R"D(" y1=")D"<<y_v<<R"D(" x2=")D"<<x2<<R"D(" y2=")D"<<y_v<<R"D(" class=")D"<<lineStyle<<R"D(" />
				)D";

				if (drawLeftNode)
					stream << R"D(
						<line x1=")D"<<x1<<R"D(" y1=")D"<<y<<R"D(" x2=")D"<<x1<<R"D(" y2=")D"<<y_v<<R"D(" class=")D"<<lineStyle<<R"D(" />
					)D";

				if (drawRightNode)
					stream << R"D(
						<line x1=")D"<<x2<<R"D(" y1=")D"<<y<<R"D(" x2=")D"<<x2<<R"D(" y2=")D"<<y_v<<R"D(" class=")D"<<lineStyle<<R"D(" />
					)D";
			} else {
				if (!drawLeftNode)
					stream << R"D(
						<line x1=")D"<<x2<<R"D(" y1=")D"<<y<<R"D(" x2=")D"<<x2-transitionHalfWidth<<R"D(" y2=")D"<<y_v<<R"D(" class=")D"<<lineStyle<<R"D(" />
						<line x1=")D"<<x1<<R"D(" y1=")D"<<y_v<<R"D(" x2=")D"<<x2-transitionHalfWidth<<R"D(" y2=")D"<<y_v<<R"D(" class=")D"<<lineStyle<<R"D(" />
					)D";
				else if (!drawRightNode)
					stream << R"D(
						<line x1=")D"<<x1<<R"D(" y1=")D"<<y<<R"D(" x2=")D"<<x1+transitionHalfWidth<<R"D(" y2=")D"<<y_v<<R"D(" class=")D"<<lineStyle<<R"D(" />
						<line x1=")D"<<x1+transitionHalfWidth<<R"D(" y1=")D"<<y_v<<R"D(" x2=")D"<<x2<<R"D(" y2=")D"<<y_v<<R"D(" class=")D"<<lineStyle<<R"D(" />
					)D";
				else
					stream << R"D(
						<line x1=")D"<<x1<<R"D(" y1=")D"<<y<<R"D(" x2=")D"<<x1+transitionHalfWidth<<R"D(" y2=")D"<<y_v<<R"D(" class=")D"<<lineStyle<<R"D(" />
						<line x1=")D"<<x2<<R"D(" y1=")D"<<y<<R"D(" x2=")D"<<x2-transitionHalfWidth<<R"D(" y2=")D"<<y_v<<R"D(" class=")D"<<lineStyle<<R"D(" />
						<line x1=")D"<<x1+transitionHalfWidth<<R"D(" y1=")D"<<y_v<<R"D(" x2=")D"<<x2-transitionHalfWidth<<R"D(" y2=")D"<<y_v<<R"D(" class=")D"<<lineStyle<<R"D(" />
					)D";
			}
		}
		if (drawLabel) {
			clipPathId++;
			stream << R"D(
			<clipPath id=")D"<<waveform->getLabel()<<R"D(_signal_label_clip_box_)D"<<clipPathId<<R"D(">
				<rect x=")D"<<x1<<R"D(" y=")D"<<y_high<<R"D(" width=")D"<<x2-x1<<R"D(" height=")D"<<y_low-y_high<<R"D("/>
			</clipPath>

			<text clip-path="url(#)D"<<waveform->getLabel()<<R"D(_signal_label_clip_box_)D"<<clipPathId<<R"D()" x=")D"<<(x1+x2)/2<<R"D(" y=")D"<<y<<R"D("
						class="wf_signalValue" dominant-baseline="middle" text-anchor="middle">)D";


			formatValue(state, width, style);
			stream << "</text>\n";
		}
	};



	for (auto i : utils::Range(renderedSignals.size())) {
		auto id = renderedSignals[i].id;
		size_t y = timelineHeight + i * signalHeight;
		stream << R"D(
			<text clip-path="url(#)D"<<waveform->getLabel()<<R"D(_nameBox)" class="wf_signalName" x="3" y=")D" << y+signalHeight/2 << "\">" << waveform->getTrace()->signals[id].name << R"D(:</text>
			<text clip-path="url(#)D"<<waveform->getLabel()<<R"D(_nameBox)" class="wf_signalValue" x="20" y=")D" << y+signalHeight-3 << R"D(" id="signal_label_)D" << i << R"D(">N/A</text>
		)D" << std::endl;



		size_t lastState = ~0ull;
		float lastX = float(legendWidth);

		float base_Y = y+signalHeight/2.f;

		if (renderedSignals[i].changes.empty()) {
			drawSegment(base_Y, lastX, 1000, lastState, renderedSignals[i].width, renderedSignals[i].isClock, renderedSignals[i].style, false, false);
		} else {
			bool first = true;
			for (const auto &change : renderedSignals[i].changes) {
				float x_position = time2x(change.time);

				drawSegment(base_Y, lastX, x_position, lastState, renderedSignals[i].width, renderedSignals[i].isClock, renderedSignals[i].style, !first, true);
				first = false;

				lastX = x_position;
				lastState = change.state;
			}
			drawSegment(base_Y, lastX, 1000, lastState, renderedSignals[i].width, renderedSignals[i].isClock, renderedSignals[i].style, !first, false);
		}
	}

	stream << R"D(
			<line class="wf_majorGridLine" x1=")D"<<legendWidth<<R"D(" y1="0" x2=")D"<<legendWidth<<R"D(" y2="100%"/>
			<line class="wf_majorGridLine" x1="0" y1=")D" << timelineHeight <<R"D(" x2="1000" y2=")D" << timelineHeight <<R"D("/>
			<line id="cursor" class="wf_cursor" x1=")D"<<legendWidth<<R"D(" y1="0" x2=")D"<<legendWidth<<R"D(" y2="100%"/>
		</svg>
		<script>
			document.getElementById(')D" << waveform->getLabel() << R"D(_svg').addEventListener('mousemove', function(e) {


				var valueLookupTables = [
	)D";

	for (auto i : utils::Range(renderedSignals.size())) {
		if (i > 0) stream << ", \n";
		stream << R"D(
			[
		)D";
		for (auto j : utils::Range(renderedSignals[i].changes.size())) {
			const auto &change = renderedSignals[i].changes[j];
			if (j > 0)
				stream << ", \n";
			else {
				if (change.time > startTime) {
					stream << R"D(
						{
							x: )D" << time2x(startTime) << R"D(,
							v: ")D";
					formatValue(~0ull, renderedSignals[i].width, renderedSignals[i].style);
					stream << R"D("
						},
					)D";
				}
			}
			stream << R"D(
				{
					x: )D" << time2x(change.time) << R"D(,
					v: ")D";
			formatValue(change.state, renderedSignals[i].width, renderedSignals[i].style);
			stream << R"D("
				}
			)D";
		}
		stream << R"D(
			]
		)D";
	}

	stream << R"D(
				];

				function lookupValue(table, x) {
					if (table.length == 0) return "";
					var min = 0;
					var max = table.length;
					while (min+1 < max) {
						var center = ((min+max)/2) >> 0;
						if (table[center].x > x)
							max = center;
						else
							min = center;
					}
					return table[min].v;
				}


				function lookupTime(x) {
					var lambda = (x - )D"<<legendWidth<<R"D() / (1000 - )D"<<legendWidth<<R"D();
					var startTime_ns = )D" << hlim::floor(startTime * 1'000'000'000ull)<<R"D(;
					var endTime_ns = )D" << hlim::floor(endTime * 1'000'000'000ull)<<R"D(;
					return (startTime_ns + lambda * (endTime_ns - startTime_ns)) + "ns";
				}


				var svg = document.getElementById(')D" << waveform->getLabel() << R"D(_svg');

				var cursor = svg.getElementById('cursor');

				var svgLoc = svg.createSVGPoint();
	      		svgLoc.x = e.clientX;
				svgLoc.y = e.clientY;
			    svgLoc = svgLoc.matrixTransform(svg.getScreenCTM().inverse());

				for (i = 0; i < )D" << renderedSignals.size() << R"D(; i++) {
					rect = svg.getElementById(")D" << waveform->getLabel() << R"D(_bg_"+i);
					y0 = i * )D" << signalHeight << " + " << timelineHeight << R"D(;
					y1 = (i+1) * )D" << signalHeight << " + " << timelineHeight << R"D(;
					if (svgLoc.y >= y0 && svgLoc.y < y1)
						rect.setAttribute("class", "wf_signalBackground_hover");
					else
						rect.setAttribute("class", "wf_signalBackground");
				}
				if (svgLoc.x > )D"<<legendWidth<<R"D() {
					cursor.setAttribute("x1", svgLoc.x);
					cursor.setAttribute("x2", svgLoc.x);


					value = svg.getElementById("time_label");
					value.textContent = lookupTime(svgLoc.x);
					for (i = 0; i < )D" << renderedSignals.size() << R"D(; i++) {
						value = svg.getElementById("signal_label_"+i);
						value.textContent = lookupValue(valueLookupTables[i], svgLoc.x);
					}
				}
			});
		</script>
	)D";
}

}
