/*
	Synogate Library, a support library for Synogate IP-core generators.
	Copyright (C) 2021 Synogate UG (haftungsbeschraenkt), www.synogate.com 
	
	This program is commercial software: you must not copy, modify or
	distribute it without explicit written license by Synogate to do so. 
	
	The full license text for non-commercial use and evaluation can be found in the distribution package of this core.
	
	This program is distributed WITHOUT ANY WARRANTY; without even the
	implied warranty of MERCHANTABILITY.
*/

#include "Report.h"

#include "Page.h"
#include "Table.h"
#include "Waveform.h"
#include "Figure.h"

#include "../../synogate_data.h"


#include <gatery/utils.h>

#include <boost/format.hpp>

#include <fstream>

namespace gtry::reporting {

Report::Report(std::filesystem::path destination) : m_destination(destination)
{
}

void Report::requireResource(std::string path)
{
    m_requiredResources.emplace_back(std::move(path));
}


Page &Report::createPage(std::string name, std::string label)
{
    m_pages.emplace_back(*this, std::move(name), std::move(label));
    return m_pages.back();
}

void Report::generate()
{
    if (!m_destination.empty())
        std::filesystem::create_directories(m_destination);

    {
        unsigned i = 1;
        for (auto &page : m_pages) {
            collectRefs(&page, boost::lexical_cast<std::string>(i));
            i++;
        }
    }

    for (auto &page : m_pages) {
        std::fstream file((m_destination / page.getLabel()).string(), std::fstream::out);
        HCL_DESIGNCHECK(file);

        generatePage(file, &page);
    }

    copyRequiredResources();
}


void Report::collectRefs(const Page *page, const std::string &pagePrefix)
{

    m_refs[page] = {.label = page->getName(), .href = page->getLabel()};
    m_refLabelToRef[page->getLabel()] = m_refs[page];

    unsigned tableCounter = 1;
    unsigned wfCounter = 1;
    unsigned figCounter = 1;

    std::function<void(const Section*, std::string)> collectReccursive;
    collectReccursive = [&](const Section *section, std::string secPrefix) {
        unsigned subSectIdx = 1;
        for (auto &subsection : section->getSubSections()) {
            std::string label;
            if (secPrefix.empty())
                label = boost::lexical_cast<std::string>(subSectIdx);
            else
                label = (boost::format("%s.%d") % secPrefix % subSectIdx).str();
            std::string href = (page->getLabel() + "#" + subsection.getLabel());
            m_refs[&subsection] = {.label = label, .href = href};
            m_refLabelToRef[subsection.getLabel()] = m_refs[&subsection];
            collectReccursive(&subsection, label);
            ++subSectIdx;
        }
        for (auto &table : section->getTables()) {
            std::string label = (boost::format("%s.%d") % pagePrefix % tableCounter).str();
            std::string href = (page->getLabel() + "#" + table.getLabel());
            m_refs[&table] = {.label = label, .href = href};
            m_refLabelToRef[table.getLabel()] = m_refs[&table];
            ++tableCounter;
        }
        for (auto &wf : section->getWaveforms()) {
            std::string label = (boost::format("%s.%d") % pagePrefix % wfCounter).str();
            std::string href = (page->getLabel() + "#" + wf.getLabel());
            m_refs[&wf] = {.label = label, .href = href};
            m_refLabelToRef[wf.getLabel()] = m_refs[&wf];
            ++wfCounter;
        }
        for (auto &fig : section->getFigures()) {
            std::string label = (boost::format("%s.%d") % pagePrefix % figCounter).str();
            std::string href = (page->getLabel() + "#" + fig.getLabel());
            m_refs[&fig] = {.label = label, .href = href};
            m_refLabelToRef[fig.getLabel()] = m_refs[&fig];
            ++figCounter;
        }
    };

    collectReccursive(page, pagePrefix);
}

void Report::generatePageHeader(std::ostream &stream, const Page *page)
{
    stream << R"Delim(<!DOCTYPE html>
<html>
<head>
  <title>)Delim" << page->getName() << R"Delim(</title>
</head>
<body>
<h1>)Delim" << page->getName() << R"Delim(</h1>
)Delim";
}

void Report::generatePageFooter(std::ostream &stream, const Page *page)
{
    stream << R"Delim(
</body>
</html>
)Delim";
}

void Report::generatePage(std::ostream &stream, const Page *page)
{
    generatePageHeader(stream, page);
    generateSection(stream, page, 0);
    generatePageFooter(stream, page);
}

void Report::generateSection(std::ostream &stream, const Section *section, unsigned depth)
{
    if (depth > 0)
        stream << "<header><h"<<depth+1<<" id=\""<< section->getLabel() <<"\">Section " << m_refs[section].label << ": " <<section->getName() << "</h"<< depth+1<<"></header>\n";

    for (auto &content : section->getContent()) {
        switch (content.type) {
            case Section::ContentSection::TEXT:
                generateText(stream, content.text.str());
            break;
            case Section::ContentSection::FLOAT:
                generateFloat(stream, content.float_);
            break;
        }
    }
    for (auto &subsection : section->getSubSections())
        generateSection(stream, &subsection, depth+1);
}


void Report::generateText(std::ostream &stream, const std::string_view &text)
{
    size_t curr = 0;

    while (curr != text.size()) {
        auto markerPos = text.find("<<<", curr);
        if (markerPos == std::string::npos) {
            stream << text.substr(curr);
            curr = text.size();
            return;
        }

        stream << text.substr(curr, markerPos-curr);
        curr = markerPos+3;

        auto markerEndPos = text.find(">>>", curr);
        HCL_DESIGNCHECK_HINT(markerEndPos != std::string::npos, "Error in report text, missing closing format >>>");

        auto spacePos = text.rfind(" ", markerEndPos);
        HCL_DESIGNCHECK_HINT(spacePos != std::string::npos, "Error in report text, missing closing format >>>");

        auto label = std::string(text.substr(spacePos+1, markerEndPos-spacePos-1));
        auto hrefIt = m_refLabelToRef.find(label);
        HCL_DESIGNCHECK_HINT(hrefIt != m_refLabelToRef.end(), std::string("Error in documentation, can't find label ")+label);

        stream << "<a href=\"" <<hrefIt->second.href << "\">" << text.substr(curr, spacePos-curr+1) << hrefIt->second.label << "</a>";
        curr = markerEndPos+3;
    }
}

void Report::generateFloat(std::ostream &stream, const Float *float_)
{
    if (auto *table = dynamic_cast<const Table*>(float_))
        generateTable(stream, table);
    else if (auto *waveform = dynamic_cast<const Waveform*>(float_))
        generateWaveform(stream, waveform);
    else if (auto *figure = dynamic_cast<const Figure*>(float_))
        generateFigure(stream, figure);
    else
        HCL_ASSERT_HINT(false, "Unhandled float type!");
}


void Report::generateTableHeader(std::ostream &stream, const Table *table)
{
    stream << "<div class=\"tbl_container\"><table class=\"tbl\" id=\"" << table->getLabel() <<"\">\n";
}

void Report::generateTableFooter(std::ostream &stream, const Table *table)
{
    stream << "</table>\n";
    stream << "<div class=\"caption_container\">Table " << m_refs[table].label << ": ";
    generateText(stream, table->getCaption().str());
    stream << "</div></div>";
}

void Report::generateTable(std::ostream &stream, const Table *table)
{
    generateTableHeader(stream, table);
    for (auto row : utils::Range(table->rows())) {
        stream << "<tr>\n";
        for (auto col : utils::Range(table->cols())) {
            if (!table->cellEmpty(row, col)) {
                const auto &cell = (*table)(row, col);
                if (cell.header)
                    stream << "<th class=\"tbl\">\n";
                else
                    stream << "<td class=\"tbl\">\n";

                generateText(stream, cell.content.str());
                if (cell.header)
                    stream << "</th>\n";
                else
                    stream << "</td>\n";
            } else {
                stream << "<td class=\"tbl\"></td>\n";
            }
        }
        stream << "</tr>\n";
    }
    generateTableFooter(stream, table);
}

void Report::generateWaveformContent(std::ostream &stream, const Waveform *waveform)
{
    stream << "No waveformviewer implemented!\n";
}

void Report::generateWaveform(std::ostream &stream, const Waveform *waveform)
{
    stream << "<div class=\"wf_container\" id=\"" << waveform->getLabel() <<"\">\n";
    generateWaveformContent(stream, waveform);
    stream << "<div class=\"caption_container\">Waveform " << m_refs[waveform].label << ": ";
    generateText(stream, waveform->getCaption().str());
    stream << "</div></div>";
}

void Report::generateFigure(std::ostream &stream, const Figure *figure)
{
    stream << "<div class=\"fig_container\" id=\"" << figure->getLabel() <<"\">\n";
    stream << figure->getInlineData().str();
    stream << "<div class=\"caption_container\">Figure " << m_refs[figure].label << ": ";
    generateText(stream, figure->getCaption().str());
    stream << "</div></div>";
}

static bool ExtractResourceToFile(std::string name, std::filesystem::path dst)
{
    if (!dst.parent_path().empty())
        std::filesystem::create_directories(dst.parent_path());

    for (auto& resource : res_synogate_data)
        if (!strcmp(resource.name, name.c_str()))
        {
            std::ofstream of{ dst.c_str(), std::ofstream::binary };
            of.write((const char*)resource.data, resource.size);
            return true;
        }
    return false;
}

void Report::copyRequiredResources()
{
    for (const auto &res : m_requiredResources) {
        bool ok = ExtractResourceToFile(res, m_destination / res);
        HCL_DESIGNCHECK_HINT(ok, "resource not found: " + res);
    }
}




}