/*
	Synogate Library, a support library for Synogate IP-core generators.
	Copyright (C) 2021 Synogate UG (haftungsbeschraenkt), www.synogate.com 
	
	This program is commercial software: you must not copy, modify or
	distribute it without explicit written license by Synogate to do so. 
	
	The full license text for non-commercial use and evaluation can be found in the distribution package of this core.
	
	This program is distributed WITHOUT ANY WARRANTY; without even the
	implied warranty of MERCHANTABILITY.
*/

#pragma once

#include "Table.h"
#include "Waveform.h"
#include "Figure.h"

#include <gatery/utils.h>

#include <string>
#include <sstream>
#include <list>
#include <ranges>

namespace gtry::reporting {

class Report;
class Float;

class Section {
    public:
        Section(Report &report, std::string name, std::string label);
        virtual ~Section();

        struct ContentSection {
            enum Type {
                TEXT,
                FLOAT
            } type;
            std::stringstream text;
            Float *float_;

            ContentSection(Type t) { type = t; }
        };

        template<typename type>
        Section &operator<<(const type &v) {
            HCL_ASSERT(m_subSections.empty());
            if (m_contentSections.empty() || m_contentSections.back().type != ContentSection::TEXT)
                m_contentSections.emplace_back(ContentSection::TEXT);
            m_contentSections.back().text << v;
            return *this;
        }

        Section &addSection(std::string name, std::string label);

        template<class FloatType>
        FloatType &addFloat(std::string name, std::string label) {
            m_floats.push_back(std::unique_ptr<Float>(new FloatType(m_report, std::move(name), std::move(label))));
            m_contentSections.emplace_back(ContentSection::FLOAT);
            m_contentSections.back().float_ = m_floats.back().get();
            return (FloatType&)*m_floats.back();
        }

        const std::list<Section> &getSubSections() const { return m_subSections; }
        const std::vector<std::unique_ptr<Float>> &getFloats() const { return m_floats; }

        struct floatIsTable { inline bool operator()(const std::unique_ptr<Float> &elem) const { return dynamic_cast<const Table*>(elem.get()); }};
        struct floatToTable { inline auto &operator()(const std::unique_ptr<Float> &elem) const { return *(const Table*)elem.get(); }};
        auto getTables() const { return m_floats | std::ranges::views::filter(floatIsTable{}) | std::views::transform(floatToTable{}); }
        Table &addTable(std::string name, std::string label) { return addFloat<Table>(std::move(name), std::move(label)); }

        struct floatIsWaveform { inline bool operator()(const std::unique_ptr<Float> &elem) const { return dynamic_cast<const Waveform*>(elem.get()); }};
        struct floatToWaveform { inline auto &operator()(const std::unique_ptr<Float> &elem) const { return *(const Waveform*)elem.get(); }};
        auto getWaveforms() const { return m_floats | std::ranges::views::filter(floatIsWaveform{}) | std::views::transform(floatToWaveform{}); }
        Waveform &addWaveform(std::string name, std::string label) { return addFloat<Waveform>(std::move(name), std::move(label)); }

        struct floatIsFigure { inline bool operator()(const std::unique_ptr<Float> &elem) const { return dynamic_cast<const Figure*>(elem.get()); }};
        struct floatToFigure { inline auto &operator()(const std::unique_ptr<Float> &elem) const { return *(const Figure*)elem.get(); }};
        auto getFigures() const { return m_floats | std::ranges::views::filter(floatIsFigure{}) | std::views::transform(floatToFigure{}); }
        Figure &addFigure(std::string name, std::string label) { return addFloat<Figure>(std::move(name), std::move(label)); }

        const std::list<ContentSection> &getContent() const { return m_contentSections; }
        const std::string &getName() const { return m_name; }
        const std::string &getLabel() const { return m_label; }
    protected:
        Report &m_report;
        std::string m_name;
        std::string m_label;

        std::list<ContentSection> m_contentSections;

        std::list<Section> m_subSections;
        std::vector<std::unique_ptr<Float>> m_floats;
};

}