/*
	Synogate Anti Replay Core, an IP core implementing an algorithm
	designed to detect packet duplication based on packet sequence numbers.
	Copyright (C) 2021 Synogate UG (haftungsbeschraenkt), www.synogate.com 
	
	This program is commercial software: you must not copy, modify or
	distribute it without explicit written license by Synogate to do so. 
	
	This program is distributed WITHOUT ANY WARRANTY; without even the
	implied warranty of MERCHANTABILITY.
*/

#include "pch.h"
#include "ARCGenerator.h"

#include "AntiReplayDocu.h"

#include <gatery/simulation/waveformFormats/VCDSink.h>
#include <gatery/export/vhdl/VHDLExport.h>

#include <gatery/utils/Range.h>
#include <gatery/hlim/Subnet.h>
#include <gatery/hlim/SignalDelay.h>
#include <gatery/hlim/RegisterRetiming.h>
#include <gatery/frontend/GraphTools.h>

#include <gatery/export/DotExport.h>
#include <gatery/hlim/Subnet.h>
#include <gatery/scl/synthesisTools/IntelQuartus.h>
#include <gatery/scl/synthesisTools/XilinxVivado.h>

#include <gatery/hlim/supportNodes/Node_Attributes.h>

#include "tests/AntiReplayFuzzer.h"

#include <external/magic_enum.hpp>

void Settings::load(gtry::utils::ConfigTree tree)
{ 
    //interfaceType = tree["interface_type"].as(interfaceType);
    interfaceType = Interface::packet;
}

ARCGenerator::ARCGenerator(gtry::utils::ConfigTree &configTree) : Generator(configTree, "Anti-Replay Core", "AntiReplayCore")
{
    m_settings.load(configTree);
}


void ARCGenerator::prepDocu(gtry::reporting::InterfaceDocumentation &doc)
{
    Generator::prepDocu(doc);


    m_cmdStreamDesc << "Command stream:";
    m_resultStreamDesc << "Result stream:";
    m_inputPacketStreamDesc << "Input packet stream:";
    m_outputPacketStreamDesc << "Output packet stream:";

#if 0
    {
        auto &table = *m_docuHooks.tables["tbl:configOpts"];

        table.addRow();
        table(-1, 0) << "Clock";
        
        std::ostringstream clockProperties;
        clockProperties << m_basicSettings.clocks.front();
        table(-1, 1) << clockProperties.str();

        table.addRow();
        table(-1, 0) << "Interface type";
        table(-1, 1) << magic_enum::enum_name(m_settings.interfaceType);
			
        table.addRow();
        table(-1, 0) << "Connection index bits ( log2(Maximum number of connections) )";
        table(-1, 1) << "TODO";// m_settings.arcSettings.connectionAddrWidth;

        table.addRow();
        table(-1, 0) << "Number of bits in window address";
        table(-1, 1) << "TODO";//m_settings.arcSettings.m_windowAddrWidth << " (physical window size of " << m_settings.arcSettings.windowSize().bits() << ')';

        table.addRow();
        table(-1, 0) << "Number of bits in sequence number";
        table(-1, 1) << "TODO";//m_settings.arcSettings.sequenceNumberWidth;

        table.addRow();
        table(-1, 0) << "Virtual window width";
        table(-1, 1) << "TODO";//m_settings.arcSettings.virtualWindowWidth;

        table.addRow();
        table(-1, 0) << "Second stage data bus width";
        table(-1, 1) << "TODO";//m_settings.arcSettings.m_secondWordWidth;

        table.addRow();
        table(-1, 0) << "Pipelining amount";
        table(-1, 1) << "TODO";//magic_enum::enum_name(m_settings.arcSettings.pipelining);
    }
#endif


    {
        auto &page = doc.createPage("Interface Description", "overview.html");

        {
            auto &sec = page.addSection("Block Description", "sec:block");

            sec << R"D(
                An overview of the configured IP-Core with the specified or inferred signal widths
                can be seen in <<<Figure fig:blockDiagram>>>.
                All important constants and configured or inferred parameters are also exposed in a vhdl interface package.
            )D";

            auto &fig = sec.addFigure("Block Diagram of IP-Core", "fig:blockDiagram");
            fig << "Block diagram of the Anti-Replay Core. Hover over labels for details.";
            m_docuHooks.figures["ipcore_blockdiagram"] = &fig;

/*
            if (m_settings.interfaceType == Settings::Interface::packet)
                sec << R"D(
                    The &quot;Packet IF&quot; is a wrapper around the &quot;ESN/SA only IF&quot;.
                )D";
*/
                sec << R"D(
                    The &quot;Packet IF&quot; is a wrapper around the inner, actual Anti-Replay Core.
                )D";

            sec << R"D(
                The input stream for commands and the output stream for results is described in detail in <<<Section sec:stream_io>>>.
            )D";
            if (m_settings.interfaceType == Settings::Interface::packet)
                sec << R"D(
                    The packet based input and output streams are described in <<<Section sec:packet_io>>>.
                )D";

            sec << R"D(
                Potential signals for external memories as well as clock and reset signals are handled on separate pages.
                For example waveforms, see <<<Page waveforms.html>>>.
            )D";

        }
        
        {
            auto &sec = page.addSection("Common Signals", "sec:common_io");
            if (m_settings.interfaceType == Settings::Interface::packet) {
                sec << R"D(
                    All streams are Avalon-Streams with one particularity: They share a common ready signal, that is, both input streams share a ready signal and both output streams share one.
                    This is necessary because results are associated with packets and having a common output ready signal ensures that result and packet streams can not be forced out of sync.
                    The IP-Core in turn gives the same guarantee, that input command and packet streams can only be ready simultaneously, an thus has only one input ready signal.
                    Signals common to all ports are listed in <<<Table tbl:input_cmd_stream>>>.
                )D";
                /*
                An asserted valid signal indicates to the core that a request is being issued.
                Once asserted, request and valid signals must remain stable until transaction occures (a clock cycle in which both valid and ready are asserted high).
                */

                {
                    auto &table = sec.addTable("Input cmd stream", "tbl:common_signals");
                    table.addHeader(4);
                    table(-1, 0) << "Field Name";
                    table(-1, 1) << "Width";
                    table(-1, 2) << "In/Out/Meta";
                    table(-1, 3) << "Description";
                    table << "Signals common to command/result and packet streams.";

                    m_docuHooks.tables["tbl:common_signals"] = &table;
                }

            } 
        }

        {
            auto &sec = page.addSection("Command and Result Streams", "sec:stream_io");

            sec << R"D(
                The core is controlled by issuing commands on the command stream. 
                All commands are processed in order and their effects are causal: A command only affects subsequent operations/commands, not preceding ones.
            )D";
            if (m_settings.interfaceType == Settings::Interface::packet)
                sec << R"D(
                    Commands can be issued during a packet frame or outside of it.
                    Their interaction with packets/frames are described in <<<Section sec:packet_io>>>.
                )D";

            sec << "<br/>The signals of the command interface can be seen in <<<Table tbl:input_cmd_stream>>>.";

            {
                auto &table = sec.addTable("Input cmd stream", "tbl:input_cmd_stream");
                table.addHeader(4);
                table(-1, 0) << "Field Name";
                table(-1, 1) << "Width";
                table(-1, 2) << "In/Out/Meta";
                table(-1, 3) << "Description";
                table << "Input command stream signals and parameters.";

                m_docuHooks.tables["tbl:input_cmd_stream"] = &table;
            }

            {
                auto &table = sec.addTable("Command opcodes", "tbl:cmd_opcodes");
                table.addHeader(3);
                table(-1, 0) << "Opcode";
                table(-1, 1) << "Binary Representation";
                table(-1, 2) << "Meaning";
                table << "Opcodes of the input command stream and their meaning. All opcodes can also be found in the vhdl interface package as vhdl constants.";

                m_docuHooks.tables["tbl:cmd_opcodes"] = &table;
            }


            sec << R"D(
                <br>The request signal is composed of a code, that indicates the type of request, and arguments to that request.
                The possible codes and their binary representation are listed in <<<Table tbl:cmd_opcodes>>>.
                Most requests configure the core and do not trigger a response. 
                Only the actual replay check OpUpdate prompts a result on the output result stream, the composition of which can be seen in <<<Table tbl:output_out_stream>>>.
                The binary representations of the result codes were chosen such that the least significant bit indicates whether the packet sequence number passed or failed the replay check.
            )D";

            {
                auto &table = sec.addTable("Output result stream", "tbl:output_out_stream");
                table.addHeader(4);
                table(-1, 0) << "Field Name";
                table(-1, 1) << "Width";
                table(-1, 2) << "In/Out/Meta";
                table(-1, 3) << "Description";
                table << "Input stream signals and parameters.";

                m_docuHooks.tables["tbl:output_out_stream"] = &table;
            }

            {
                auto &table = sec.addTable("Result opcodes", "tbl:result_opcodes");
                table.addHeader(3);
                table(-1, 0) << "Opcode";
                table(-1, 1) << "Binary Representation";
                table(-1, 2) << "Meaning";
                table << "Result codes of the result output stream and their meaning. Note that the binary outcome of the test can be checked from the least significant bit alone. All result codes can also be found in the vhdl interface package as vhdl constants.";

                m_docuHooks.tables["tbl:result_opcodes"] = &table;
            }
        }

        if (m_settings.interfaceType == Settings::Interface::packet) {

            auto &sec = page.addSection("Packet Streams", "sec:packet_io");

            sec << R"D(
                The packet input (see <<<Table tbl:input_packet_stream>>>) and output (see <<<Table tbl:output_packet_stream>>>) streams follow the usual semantics of packet based streams.
                The content is forwarded through an internal fifo to the output and largely ignored.
            )D";

            {
                auto &table = sec.addTable("Input packet stream", "tbl:input_packet_stream");
                table.addHeader(4);
                table(-1, 0) << "Field Name";
                table(-1, 1) << "Width";
                table(-1, 2) << "In/Out/Meta";
                table(-1, 3) << "Description";
                table << "Input packet stream signals and parameters.";

                m_docuHooks.tables["tbl:input_packet_stream"] = &table;
            }

            {
                auto &table = sec.addTable("Output packet stream", "tbl:output_packet_stream");
                table.addHeader(4);
                table(-1, 0) << "Field Name";
                table(-1, 1) << "Width";
                table(-1, 2) << "In/Out/Meta";
                table(-1, 3) << "Description";
                table << "Input packet stream signals and parameters.";

                m_docuHooks.tables["tbl:output_packet_stream"] = &table;
            }            

        }
    }
    {
        auto &page = doc.createPage("Memory Interfaces", "memory_interfaces.html");

        m_docuHooks.sections["memory_interfaces"] = &page;
    }
    {
        auto &page = doc.createPage("Clocks and Resets", "clks.html");
        m_docuHooks.sections["clocks_and_resets"] = &page;
    }    
    {
        auto &page = doc.createPage("Example Waveforms", "waveforms.html");


        page << R"D(
            These waveforms demonstrate the core's operation in typical use cases. The waveforms build on each other chronologically, i.e., inherit the state from the previous waveforms.
        )D";

        if (m_basicSettings.writeWaveforms) {
            page << "They are excerpts from a simulation that can be seen in the accompanying waveform .vcd file (" << (m_basicSettings.waveformPath / "PacketIFCoreDemonstration.vcd").string() << ") or";
        } else {
            page << "They are excerpts from a simulation reproduced";
        }

        page << " by simulating the demonstration testbench (" << (m_basicSettings.testbenchPath / "AntiReplay_TB_demonstration.vhdl").string() << ").";
       

        {
            auto &wf = page.addWaveform("Example initialization", "wf:init");
            m_docuHooks.waveforms["wf:init"] = &wf;
        }

        {
            auto &wf = page.addWaveform("Lookups and updates", "wf:update");
            m_docuHooks.waveforms["wf:update"] = &wf;
        }

        {
            auto &wf = page.addWaveform("Lookups and updates with intermixed configuration changes", "wf:update_width_config_change");
            m_docuHooks.waveforms["wf:update_width_config_change"] = &wf;
        }

    }
}


void ARCGenerator::buildCore()
{
    Generator::buildCore();

    switch (m_settings.interfaceType)
    {
    case Settings::Interface::packet:
        buildPacketIFCore();
        break;
    case Settings::Interface::esn_sa_only:
        buildAntiReplayCore();
        break;
    }
}

void ARCGenerator::prepExport()
{
    Generator::prepExport();

    if (m_basicSettings.fuzzingTest.numBatches > 0)
        m_vhdlExport->addTestbenchRecorder(m_fuzzingSimulation, std::string("TB_")+m_basicSettings.library+"_fuzzing", false);
    m_vhdlExport->addTestbenchRecorder(m_demonstrationSimulation, std::string("TB_")+m_basicSettings.library+"_demonstration", true);

    auto &interfacePackage = m_vhdlExport->getInterfacePackage();

    interfacePackage.addNatural("MINIMUM_LATENCY", *m_coreLatency, "The number of pipeline stages of the anti replay core, which is also the minimum latency (can be more depending on external memory type)");
    if (*m_clockNeedsReset)
        interfacePackage.addNatural("MINIMUM_RESET_CYCLES", *m_resetCycles, "The minimum number of cycles that the reset needs to be asserted for the reset to complete.");
//    interfacePackage.addNatural("WINDOW_ADDR_WIDTH", m_settings.arcSettings.m_windowAddrWidth.bits(), ""); // todo: Needed?

    gtry::hlim::NodeGroup *top = gtry::DesignScope::get()->getCircuit().getRootNodeGroup();
    gtry::hlim::NodeGroup *frameAdapter = top->findChild("FrameAdapter0");
    gtry::hlim::NodeGroup *arc = frameAdapter->findChild("AntiReplay0");

    interfacePackage.addNatural("SECOND_WORD_WIDTH", arc->properties()["memory_word_width"].as<size_t>(), "Width (number of bits) of the data bus to the \"main storage\"");
    interfacePackage.addNatural("CONNECTION_ADDR_WIDTH", top->properties()["connection_addr_width"].as<size_t>(), "Number of bits of the connection indices (the maximal number of connections is a power of 2)");
    interfacePackage.addNatural("VIRTUAL_WINDOW_WIDTH", top->properties()["virtual_window_width"].as<size_t>(), "Number of bits for representing the virtual window size");
    interfacePackage.addNatural("SEQUENCE_NUMBER_WIDTH", top->properties()["sequence_number_width"].as<size_t>(), "Number of bits of the sequence numbers");
    interfacePackage.addNatural("PHYSICAL_WINDOW_SIZE", arc->properties()["window_size"].as<size_t>(), "The total physical window size");
    if (m_packetDataWidth)
        interfacePackage.addNatural("PACKET_DATA_WIDTH", *m_packetDataWidth, "Number of data bits of the packet stream");
    if (m_packetEmptyWidth)
        interfacePackage.addNatural("PACKET_EMPTY_WIDTH", *m_packetEmptyWidth, "Number of bits in the 'empty' signal of the packet stream");

    interfacePackage.addNatural("OPCODE_WIDTH", ((gtry::BVec)arc::OpUpdate()).size(), "Number of bits of the opcode field");
    interfacePackage.addNatural("RESULT_WIDTH", ((gtry::BVec)arc::FailReplay()).size(), "Number of bits of the result field");

    interfacePackage.addBVecConstant(arc::OpUpdate());
    interfacePackage.addBVecConstant(arc::OpActivate());
    interfacePackage.addBVecConstant(arc::OpDeactivate());
    interfacePackage.addBVecConstant(arc::OpSetSeqNr());
    interfacePackage.addBVecConstant(arc::OpSetWindowSize());

    interfacePackage.addBVecConstant(arc::PassAboveWindow());
    interfacePackage.addBVecConstant(arc::PassCheckDisabled());
    interfacePackage.addBVecConstant(arc::PassInVirtualWindow());
    interfacePackage.addBVecConstant(arc::PassInPhysicalWindow());
    interfacePackage.addBVecConstant(arc::FailBelowWindow());
    interfacePackage.addBVecConstant(arc::FailReplay());
}


void ARCGenerator::runSimulations()
{
    Generator::runSimulations();

/*
gtry::sim::SimulatorConsoleOutput consoleOutput;
m_fuzzingSimulation.addCallbacks(&consoleOutput);
*/
    m_demonstrationSimulation.powerOn();
    m_demonstrationSimulation.advance(gtry::hlim::ClockRational(*m_resetCycles + 1024, 1) / *m_basicSettings.clocks.front().absoluteFrequency());

    if (m_basicSettings.fuzzingTest.numBatches > 0) {
        m_fuzzingSimulation.powerOn();
        auto testDuration = gtry::hlim::ClockRational(
            *m_resetCycles + m_basicSettings.fuzzingTest.numBatches * m_basicSettings.fuzzingTest.numCyclesPerBatch,
            1
        ) / *m_basicSettings.clocks.front().absoluteFrequency();

        if (m_basicSettings.fuzzingTest.reportProgress) {
            class FuzzingTestProgress : public gtry::sim::SimulatorCallbacks
            {
                public:
                    size_t i = 0;
                    gtry::hlim::ClockRational testDuration;

                    FuzzingTestProgress(gtry::hlim::ClockRational testDuration) : testDuration(testDuration) { }
                    virtual void onNewTick(const gtry::hlim::ClockRational &simulationTime) { 
                        if (i == 0) {
                            std::cout << "Running extensive test:" << std::endl;
                        }
                        if (i++ % 64 == 0) {
                            auto fractionDone = simulationTime / testDuration;

                            std::cout << "\r    Progress: " << fractionDone.numerator() * 100 / (float) fractionDone.denominator() << " %        " << std::flush;
                        }
                    }
            };

            FuzzingTestProgress testOutput(testDuration);        
            m_fuzzingSimulation.addCallbacks(&testOutput); // a bit dangerous in case sim gets continued somewhere

            m_fuzzingSimulation.advance(testDuration);
            std::cout << std::endl;
        } else
            m_fuzzingSimulation.advance(testDuration);
    }
}

void ARCGenerator::finishDocu(gtry::reporting::InterfaceDocumentation &doc)
{
    Generator::finishDocu(doc);
    
    documentClocks();
    documentMemoryIntefaces();
    
    {
        gtry::hlim::NodeGroup *rootNodeGroup = gtry::DesignScope::get()->getCircuit().getRootNodeGroup();
        gtry::hlim::NodeGroup *arcNodeGroup = rootNodeGroup->findChild("FrameAdapter0")->findChild("AntiReplay0");

        auto &table = *m_docuHooks.tables["tbl:configOpts"];

        table.addRow();
        table(-1, 0) << "Maximum number of connections";
        table(-1, 1) << (1 << rootNodeGroup->properties()["connection_addr_width"].as<size_t>());

        table.addRow();
        table(-1, 0) << "Number of bits in sequence number";
        table(-1, 1) << rootNodeGroup->properties()["sequence_number_width"].as<size_t>();

        table.addRow();
        table(-1, 0) << "Virtual window size range";
        table(-1, 1) << "1 - " << (1 << rootNodeGroup->properties()["virtual_window_width"].as<size_t>())-1 << " (inclusive)";

        table.addRow();
        table(-1, 0) << "Physical window size";
        table(-1, 1) << (1 << arcNodeGroup->properties()["window_addr_width"].as<size_t>()) - arcNodeGroup->properties()["memory_word_width"].as<size_t>() << " (highest seen sequence number is part of window)";

        table.addRow();
        table(-1, 0) << "Pipelining amount (core latency)";
        table(-1, 1) << *m_coreLatency << " cycles";

        table.addRow();
        table(-1, 0) << "Minimum reset cycles";
        if (*m_clockNeedsReset) {
            table(-1, 1) << *m_resetCycles << " cycles";
        } else {
            table(-1, 1) << "Explicit reset disabled";
        }
    }

    {
        m_docuHooks.waveforms["wf:init"]->fromMemoryTrace(
            m_memoryTrace, m_waveformSignalsInitialization,
            {
                .selectionAnnotationId = "initialization",
                //.collapse=true
            });

        m_docuHooks.waveforms["wf:update"]->fromMemoryTrace(
            m_memoryTrace, m_waveformSignalsAll,
            {
                .selectionAnnotationId = "update",
                //.collapse=true
            });

        m_docuHooks.waveforms["wf:update_width_config_change"] ->fromMemoryTrace(
            m_memoryTrace, m_waveformSignalsAll,
            {
                .selectionAnnotationId = "update_with_config_change",
                //.collapse=true
            });
    }

    {
        gtry::reporting::BlockDiagram diagram;

        using namespace gtry::reporting;


        auto &cmdStreamIn = diagram.canvas().add<Pin>();
        auto &resultStreamOut = diagram.canvas().add<Pin>();


        auto &mainBlock = diagram.canvas().add<Block>()
            .setStyle("bd_entity2")
            .setMargin({20.0f, 40.0f, 20.0f, 20.0f});

        cmdStreamIn.leftOf(&mainBlock, 20.0f);
        cmdStreamIn.aboveOf(&mainBlock, 20.0f);
        resultStreamOut.rightOf(&mainBlock, 20.0f);
        resultStreamOut.aboveOf(&mainBlock, 20.0f);


        if (m_settings.interfaceType == Settings::Interface::packet) {
            mainBlock.setTitle("Anti-Replay Core (Packet IF)");
            drawPacketIFCore(diagram, mainBlock, cmdStreamIn, resultStreamOut);
        } else {
            gtry::reporting::Shape *inputRouting = nullptr;
            gtry::reporting::Shape *outputRouting = nullptr;

            mainBlock.setTitle("Anti-Replay Core (ESN/SA only IF)");
            //drawAntiReplayCore(diagram, mainBlock, inputRouting, outputRouting);

            diagram.canvas().connect(&cmdStreamIn, Shape::SOUTH, inputRouting, Shape::EAST)
                .setName("cmd")
                .setToolTip(m_cmdStreamDesc.str())
                .setStyle("bd_connection_no_head");

            diagram.canvas().connect(outputRouting, Shape::EAST, &resultStreamOut, Shape::SOUTH)
                .setName("result")
                .setToolTip(m_resultStreamDesc.str());

        }

        diagram.write(m_docuHooks.figures["ipcore_blockdiagram"]->getInlineData());
    }

}

void ARCGenerator::documentClocks()
{
    auto &page = *m_docuHooks.sections["clocks_and_resets"];

    page << R"D(
        The IP-Core is driven by a single clock and doesn't create any additional clocks internally.
    )D";

    if (*m_clockNeedsReset) {
        page << "The core does is configured to have an explicit reset. This reset affects all registers and memories necessary to bring or return the core to its initial state.";
        page << "The reset needs to be asserted for at least " << *m_resetCycles << " cycles for the reset to complete but may be asserted longer.";

        if (m_clock->getRegAttribs().initializeRegs && m_clock->getRegAttribs().initializeMemory) {
            page << "<br/>In addition, all required registers and memories are initialized on power-on through default values in VHDL so that performing the explicit reset may be skipped after power-on.";
        }

        if (m_clock->getRegAttribs().resetActive == gtry::hlim::RegisterAttributes::Active::HIGH)
            page << "<br/>The reset is high-active, meaning it is considered asserted when '1'.";
        else
            page << "<br/>The reset is low-active, meaning it is considered asserted when '0'.";
    } else {
        page << "In the current configuration, the core does not have an explicit reset. Rather, all registers are initialized on power-on through default values in VHDL.";
        /*
        if (!m_clock->getRegAttribs().initializeMemory) {
            // all initializations are to zero, which FPGAs will do anyway
        }
        */
    }
}

void ARCGenerator::documentMemoryIntefaces()
{
    auto &page = *m_docuHooks.sections["memory_interfaces"];

    page << "This page describes all memories that must be implemented outside of the core.\n";

    bool pageEmpty = true;

    std::vector<gtry::hlim::NodeGroup *> stack = {gtry::DesignScope::get()->getCircuit().getRootNodeGroup()};

    size_t sectionCounter = 0;

    while (!stack.empty()) {
        gtry::hlim::NodeGroup *ng = stack.back();
        stack.pop_back();

        if (ng->getName() == "scl_memory") { // Not sure this is a good idea

            if (ng->properties()["type"].as<std::string>() == "EXTERNAL") {


                std::string path = ng->instancePath();

                std::string name = path;
                if (path == "/FrameAdapter0/AntiReplay0/SecondLookupStage0/scl_memory0")
                    name = "Second Stage Memory";

                auto &sec = page.addSection(name, "");

                if (path == "/FrameAdapter0/AntiReplay0/SecondLookupStage0/scl_memory0")
                    sec << "This memory is the actual bitmap of the physical windows of all connections.\n";

                std::string signalTblId = (boost::format("tbl:mem_signals_%d") % sectionCounter).str();
                std::string memPropTblId = (boost::format("tbl:mem_properties_%d") % sectionCounter).str();

                sec 
                    << "An external implementation of this memory must be connected to the signals described <<<Table "<<signalTblId<<">>>.\n"
                    << "These signals are part of the portmap of this core's top entity.\n"
                    << "<br/>The properties that are expected of the external memory are listed in <<<Table "<<memPropTblId<<">>>.\n";

                auto &physMemProps = ng->findChild("physical_memory0")->properties();
                {
                    auto &table = sec.addTable("Signals of external memory", signalTblId);
                    table.addHeader(4);
                    table(-1, 0) << "Field Name";
                    table(-1, 1) << "Width";
                    table(-1, 2) << "In/Out/Meta";
                    table(-1, 3) << "Description";
                    table << "Signals of the portmap of the core's top entity to which the external memory must be connected. In/out direction from the perspective of this core.";

                    bool hasReadEnable = physMemProps["port_0_has_readEnable"].as<bool>();
                    if (hasReadEnable) {
                        table.addRow();
                        table(-1, 0) << physMemProps["port_0_pinName_readEnable"];
                        table(-1, 1) << gtry::BitWidth(1);
                        table(-1, 2) << "out";
                        table(-1, 3) << "Asserted high to signal a read. When high, the memory content adddressed by" << physMemProps["port_0_pinName_addr"] 
                                    << " is expected on " << physMemProps["port_0_pinName_readData"] << " after " << ng->properties()["readLatency"] 
                                    << " cycles (subject to read-during-write conflicts). Similarly when low, the corresponding beat after " << ng->properties()["readLatency"]  << " cycles in " 
                                    << physMemProps["port_0_pinName_readData"] << " may be undefined.";
                    }

                    table.addRow();
                    table(-1, 0) << physMemProps["port_0_pinName_addr"];
                    table(-1, 1) << physMemProps["port_0_width_addr"].as<gtry::BitWidth>();
                    table(-1, 2) << "out";
                    table(-1, 3) << "Memory address to read from. Address is in words.";

                    table.addRow();
                    table(-1, 0) << physMemProps["port_0_pinName_readData"];
                    table(-1, 1) << physMemProps["port_0_width_readData"].as<gtry::BitWidth>();
                    table(-1, 2) << "in";
                    table(-1, 3) << "Memory read from the address applied to " << physMemProps["port_0_pinName_addr"] << " " << ng->properties()["readLatency"] << " cycles earlier (subject to read-during-write conflicts).";


                    bool hasWriteEnable = physMemProps["port_1_has_writeEnable"].as<bool>();
                    if (hasWriteEnable) {
                        table.addRow();
                        table(-1, 0) << physMemProps["port_1_pinName_writeEnable"];
                        table(-1, 1) << gtry::BitWidth(1);
                        table(-1, 2) << "out";
                        table(-1, 3) << "Asserted high to signal a write.";
                    }

                    table.addRow();
                    table(-1, 0) << physMemProps["port_1_pinName_addr"];
                    table(-1, 1) << physMemProps["port_1_width_addr"].as<gtry::BitWidth>();
                    table(-1, 2) << "out";
                    table(-1, 3) << "Memory address to write to. Address is in words.";

                    table.addRow();
                    table(-1, 0) << physMemProps["port_1_pinName_writeData"];
                    table(-1, 1) << physMemProps["port_1_width_writeData"].as<gtry::BitWidth>();
                    table(-1, 2) << "in";
                    table(-1, 3) << "Data to write to the memory at the address " << physMemProps["port_1_pinName_addr"];

                    m_docuHooks.tables[signalTblId] = &table;

                    if (hasReadEnable) {
                        m_waveformSignalsAll.push_back({
                            .n = physMemProps["port_0_pinName_readEnable"].as<std::string>(), 
                            .style = gtry::reporting::Waveform::defaultBin
                        });
                    }

                    m_waveformSignalsAll.push_back({
                        .n = physMemProps["port_0_pinName_addr"].as<std::string>(), 
                        .style = gtry::reporting::Waveform::defaultHex
                    });

                    m_waveformSignalsAll.push_back({
                        .n = physMemProps["port_0_pinName_readData"].as<std::string>(), 
                        .style = gtry::reporting::Waveform::defaultInHex
                    });
                    
                    if (hasWriteEnable) {
                        m_waveformSignalsAll.push_back({
                            .n = physMemProps["port_1_pinName_writeEnable"].as<std::string>(), 
                            .style = gtry::reporting::Waveform::defaultBin
                        });
                    }

                    m_waveformSignalsAll.push_back({
                        .n = physMemProps["port_1_pinName_addr"].as<std::string>(), 
                        .style = gtry::reporting::Waveform::defaultHex
                    });

                    m_waveformSignalsAll.push_back({
                        .n = physMemProps["port_1_pinName_writeData"].as<std::string>(), 
                        .style = gtry::reporting::Waveform::defaultHex
                    });
                }

                {
                    auto &table = sec.addTable("External memory properties", memPropTblId);
                    table.addHeader(2);
                    table(-1, 0) << "Parameter Name";
                    table(-1, 1) << "Parameter Value";
                    table << "Properties that are expected of the external memory.";

                    table.addRow();
                    table(-1, 0) << "Total size";
                    table(-1, 1) << gtry::BitWidth(ng->properties()["word_width"].as<size_t>() * ng->properties()["num_words"].as<size_t>());

                    table.addRow();
                    table(-1, 0) << "Width (size of a word)";
                    table(-1, 1) << ng->properties()["word_width"];

                    table.addRow();
                    table(-1, 0) << "Depth (number of words)";
                    table(-1, 1) << ng->properties()["num_words"];

                    table.addRow();
                    table(-1, 0) << "Read latency (in cycles)";
                    table(-1, 1) << ng->properties()["readLatency"];

                    table.addRow();
                    table(-1, 0) << "Read during write policy (cross port)";
                    table(-1, 1) << physMemProps["crossPortReadDuringWrite"];

                    m_docuHooks.tables[memPropTblId] = &table;
                }

                if (ng->properties()["readLatency"].as<size_t>()) {
                    sec 
                        << "<br/> The memory is expected to have a fixed read latency of " << ng->properties()["readLatency"] << " cycles.\n"
                        << "A change of read address should result in (potentially) different data being output " << ng->properties()["readLatency"] << " cycles later.\n";
                }

                if (physMemProps["crossPortReadDuringWrite"].as<std::string>() == "DONT_CARE")
                    sec << "If reads and writes happen to the same address but in different cycles, they are expected to be consistent chronologically, i.e. a later read picking up an earlier write.\n"
                        << "If a read and write happen to the same address in the same cycle, the write is expected to succeed, but the read may return undefined data.";
            

                pageEmpty = false;
                sectionCounter++;
            }

        } else 
            for (auto &n : ng->getChildren())
                stack.push_back(n.get());
    }
    
    
    if (pageEmpty) {
        page << "However, in the current configuration, all memories are implemented internally.\n";
    }
}


void ARCGenerator::buildAntiReplayCore()
{
    using namespace gtry;

    Clock clock(m_basicSettings.clocks.front());
    ClockScope clkScp(clock);
    m_clock = clock.getClk();
    m_clockNeedsReset = clock.getClk()->getRegAttribs().resetType != hlim::RegisterAttributes::ResetType::NONE;

    HCL_DESIGNCHECK_HINT(m_clock->getRegAttribs().initializeRegs || *m_clockNeedsReset,
        "The clock was configured to not have a reset, while also disabling default value initialization of the registers. This leaves the circuit in an undefined state uppon power-on from which it can not be recovered.");

    arc::CmdStage cmd;
    cmd.loadConfig();
    cmd.pinIn("cmd_");

    arc::AntiReplayResult out;
    BitWidth windowSize;
    {
        arc::AntiReplay ar{ cmd };
        ar.generateFixedLatency();
        ar.result(out);
        windowSize = ar.windowSize();
    }
    //std::cerr << "pipeline stages: " << ar.latency() << '\n';

    pinOut(out.valid).setName("out_valid");
    pinOut(out.code).setName("out_code");

    generateCmdDocumentation(cmd);
    generateResultDocumentation(out.valid, out.code);

    arc::AntiReplayFuzzer fuzzer{ 18055 };
    fuzzer.distanceBits(10);
    fuzzer.saMask(1);

    m_fuzzingSimulation.addSimulationProcess([&]() -> SimProcess {

        for (size_t i = 0; i < 2; ++i)
        {
            simu(cmd.valid) = 1;
            simu(cmd.opcode) = 2;
            simu(cmd.connectionIndex) = i;
            simu(cmd.sequenceNumber) = 0;
            simu(cmd.windowSize) = 0;
            co_await WaitClk(clock);

            simu(cmd.opcode) = 5;
            co_await WaitClk(clock);
        }
        simu(cmd.opcode) = 3;
        simu(cmd.windowSize) = 300;
        co_await WaitClk(clock);

        simu(cmd.valid) = 0;
        simu(cmd.opcode) = 0;
        simu(cmd.windowSize) = 0;

        for (size_t i = 0; i < 4; ++i)
            co_await WaitClk(clock);

        simu(cmd.valid) = 1;
        while (true)
        {
            auto [saId, saIdx] = fuzzer.next();
            simu(cmd.connectionIndex) = saId;
            simu(cmd.sequenceNumber) = saIdx;
            co_await WaitClk(clock);
        }

        });

    m_fuzzingSimulation.addSimulationProcess(arc::AntiReplayVerifier{ clock, cmd, out, windowSize, false });

    DesignScope::get()->getCircuit().postprocess(gtry::DefaultPostprocessing{});


    m_fuzzingSimulation.compileProgram(DesignScope::get()->getCircuit());
    m_demonstrationSimulation.compileProgram(DesignScope::get()->getCircuit());
}

void ARCGenerator::buildPacketIFCore()
{
    using namespace gtry;

    Clock clock(m_basicSettings.clocks.front());
    ClockScope clkScp(clock);
    m_clock = clock.getClk();
    m_clockNeedsReset = clock.getClk()->getRegAttribs().resetType != hlim::RegisterAttributes::ResetType::NONE;

    HCL_DESIGNCHECK_HINT(m_clock->getRegAttribs().initializeRegs || *m_clockNeedsReset,
        "The clock was configured to not have a reset, while also disabling default value initialization of the registers. This leaves the circuit in an undefined state upon power-on from which it can not be recovered.");


    GroupScope& ent = DesignScope::get()->getRootGroup();
    ent.setComment("This is the top module of the Synogate Anti-Replay core in the Packet Interface configuration.");

    describeCoreClock("clock", "Clock of the IP-Core");

    arc::CmdStage cmd;
    cmd.loadConfig();
    cmd.pinIn("in_cmd_");

    arc::FrameAdapter adapter;
    auto &antiReplayCore = adapter.setupAntiReplay(cmd);
    antiReplayCore.generateFixedLatency();

    arc::FrameStream& dataIn = adapter.dataIn();
    dataIn.setup(ent.config("frame_stream"));

    arc::Stream<std::tuple<arc::FrameStream, arc::Valid<BVec>>>& out = adapter.generate();
    adapter.assignPins();

    m_coreLatency = antiReplayCore.latency();
    m_packetDataWidth = out.get<arc::FrameStream>().data.size();
    m_packetEmptyWidth = out.get<arc::FrameStream>().empty.size();

    // All inputs
    describeCoreCommonSignal(gtry::findOutputPin(adapter.readyIn())->getName().c_str(), false, true, 1_b, "Shared ready signal for command and input packet streams");
    generateCmdDocumentation(cmd);
    generateInputPacketDocumentation(dataIn);

    // All outputs
    describeCoreCommonSignal(gtry::findInputPin(out.ready)->getName().c_str(), true, true, 1_b, "Shared ready signal for result and output packet streams");
    generateResultDocumentation(out.get<arc::Valid<gtry::BVec>>().valid, out.get<arc::Valid<gtry::BVec>>().data);
    generateOutputPacketDocumentation(out.get<arc::FrameStream>());


    auto *wf_init = m_docuHooks.waveforms["wf:init"];
    auto *wf_update = m_docuHooks.waveforms["wf:update"];
    auto *wf_update_width_config_change = m_docuHooks.waveforms["wf:update_width_config_change"];

    // test objects
    auto demoStreamDriver = std::make_shared<arc::FrameStreamTestDriver>(clock, dataIn, adapter.readyIn());
    auto demoStreamVerifier = std::make_shared<arc::FrameStreamTestVerifier>(clock, out.get<arc::FrameStream>(), out.ready);
    auto demoCmdGenerator = std::make_shared<arc::AntiReplayFuzzer>(18055);
    demoCmdGenerator->cmd(clock, cmd);
    demoCmdGenerator->cmdReady(adapter.readyIn());
    demoCmdGenerator->saMask(1);
    demoCmdGenerator->distanceBits(utils::Log2C(antiReplayCore.windowSize().bits()));

    m_demonstrationSimulation.addSimulationProcess([=, this]() -> SimProcess {

        simu(dataIn.valid) = 0;
        simu(cmd.valid) = 0;
        simu(out.ready) = 1;
        co_await WaitClk(clock);

        const char *initDesc =             R"D(
        Shows a sample "initialization" (could also happen during normal operation) comprising: 
            1) Setting the virtual window size to 100 (it defaults to the physical window size);
            2) Initializing the sequence number (OpSetSeqNr) of connection 0 to 0; 
            3) Activating (OpActivate) connection 0;
            4) Initializing the sequence number (OpSetSeqNr) of connection 1 to 0; 
            5) Activating (OpActivate) connection 1;
)D";
        *wf_init << initDesc;
        simAnnotationStart("initialization", initDesc);
        co_await WaitClk(clock);

        demoCmdGenerator->virtualWindowSize(100);

		for (size_t i = 0; i < 2; ++i)
		{
            demoCmdGenerator->setseq(i, 0, true);
            demoCmdGenerator->activate(i);
		}

        while(demoCmdGenerator->busy())
            co_await WaitClk(clock);

        co_await WaitClk(clock);
        simAnnotationEnd("initialization");

        const char *updateDesc = R"D(
        Shows a normal replay check and update operation comprising:
            1) Receiving a 3-beat packet on connection 0 with sequence number 5 (a new highest sequence number).
            2) Receiving the exact same packet again: same connection index, same sequence number. The payload happens to be the same but that is irrelevant for the test.
            3) Receiving a single-beat packet on connection 16 with sequence number 32 (the connection has not been activated).
            
        The core delays the first packet because of its internal latency but then responds in the same rate as the rate with which packets were received.
        The responses coincide with the packets on the output stream and comprise:
            1) A PassAboveWindow for the first packet.
            2) A FailReplay for the duplicate of the first packet.
            3) A PassCheckDisabled for the last packet on the deactivated connection.
)D";
        *wf_update << updateDesc;
        simAnnotationStart("update", updateDesc);
        co_await WaitClk(clock);

        for (size_t i = 0; i < 2; ++i)
        {
            demoStreamDriver->pushBeats(std::array<uint64_t, 3>{ 0xCADDE, 0xBEEF, 0x1337 });
            co_await WaitClk(clock);
            co_await WaitClk(clock);
            demoCmdGenerator->update(0, 5);
            co_await WaitClk(clock);
        }

        demoStreamDriver->pushBeats(std::array<uint64_t, 1>{ 0xCADDE });
        demoCmdGenerator->update(0x10, 0x20);

        {
            size_t waitCmd = 0;
            do {
                if (simu(out.get<arc::Valid<gtry::BVec>>().valid))
                    waitCmd++;
                co_await WaitClk(clock);
            } while (waitCmd < 3);
        }

        co_await WaitClk(clock);
        simAnnotationEnd("update");

        const char *update_w_c_chg = R"D(
        Shows normal replay check operation where the configuration is changed during operation:
            1) Receiving of a single-beat packet on connection 32 with sequence number 1 (the connection has not yet been activated).
            2) A bubble in the packet stream during which the sequence number of connection 32 is set to 0 (OpSetSeqNr).
            3) An unrelated single beat packet in the packet stream for which a replay check is not required and during which the connection 32 is activated (OpActivate).
            3) The exact same packet of 1) is received again.
            4) The exact same packet of 1) is received a third time.
            
        The core delays the first packet because of its internal latency but then responds in the same rate as the rate with which packets were received.
        The responses coincide with the packets on the output stream and comprise:
            1) A PassCheckDisabled for the first packet on the deactivated connection.
            2) The unrelated packet with no corresponding result, transmitted as soon as possible but without breaking order.
            3) A bubble corresponding to the bubble on the input stream while the result of the replay check is awaited (necessary, because the unrelated packet was transmitted one beat earlier).
            4) A PassAboveWindow for the first replay (because the sequence number was set to zero).
            5) A FailReplay for the second replay.
)D";

        *wf_update_width_config_change << update_w_c_chg;
        simAnnotationStart("update_with_config_change", update_w_c_chg);
        co_await WaitClk(clock);

        demoStreamDriver->pushBeats(std::array<uint64_t, 1>{0xCAFFE});
        demoCmdGenerator->update(0x20, 1);
        co_await WaitClk(clock);

        demoCmdGenerator->setseq(0x20, 0, true);
        co_await WaitClk(clock);
        
        demoCmdGenerator->activate(0x20);
        demoStreamDriver->pushBeats(std::array<uint64_t, 1>{0x1234});

        demoCmdGenerator->update(0x20, 1);
        demoStreamDriver->pushBeats(std::array<uint64_t, 1>{0xCAFFE});
        
        demoCmdGenerator->update(0x20, 1);
        demoStreamDriver->pushBeats(std::array<uint64_t, 1>{0xCAFFE});

        {
            size_t waitCmd = 0;
            do {
                if (simu(out.get<arc::Valid<gtry::BVec>>().valid))
                    waitCmd++;
                co_await WaitClk(clock);
            } while (waitCmd < 3);
        }

        co_await WaitClk(clock);
       simAnnotationEnd("update_with_config_change");

        m_demonstrationSimulation.abort();
    });

    m_demonstrationSimulation.addSimulationProcess(std::ref(*demoStreamDriver));
    m_demonstrationSimulation.addSimulationProcess(std::ref(*demoCmdGenerator));

    const size_t m_windowAddrWidthBits = utils::Log2C(antiReplayCore.windowSize().bits());
    m_fuzzingSimulation.addSimulationProcess([=, this]() -> SimProcess {

        simu(dataIn.valid) = 0;
        simu(cmd.valid) = 0;
        simu(out.ready) = 1;
        co_await WaitClk(clock);


        for (size_t i = 0; i < 2; ++i)
        {
            demoCmdGenerator->setseq(i, 0, true);
            demoCmdGenerator->activate(i);
        }
        size_t batch = 0;
        while (true)
        {
            uint32_t seed = std::random_device{}();
            std::cerr << "Fuzzing test batch: " << batch++ << ", seed: " << seed << '\n';

            std::mt19937 rng{ seed };
            for(size_t iteration = 0; iteration < m_basicSettings.fuzzingTest.numCyclesPerBatch; ++iteration)
            {
                if (!demoStreamDriver->busy() && rng() % 3 != 0)
                {
                    auto& frame = demoStreamDriver->pushRandom(rng);
                    demoStreamVerifier->pushExpectedFrame(frame);
                }

                if (!demoCmdGenerator->busy() && rng() % 2)
                    demoCmdGenerator->updateNext();

                if (rng() % 256 == 0)
                    demoCmdGenerator->activate(rng() % 2, rng() % 2 != 0);

                if (rng() % 256 == 0)
                {
                    uint64_t sequenceNumber = uint64_t(rng()) << 32 | rng();
                    uint64_t connectionIndex = rng() % 2;
                    demoCmdGenerator->setseq(connectionIndex, sequenceNumber, true);
                }

                if (rng() % 256 == 0) {
                    demoCmdGenerator->distanceBits(rng() % (m_windowAddrWidthBits + 1) + 1);
                }

                simu(out.ready) = rng() % 2; // add some chaos
                co_await WaitClk(clock);
            }
        }

    });

    m_fuzzingSimulation.addSimulationProcess(std::ref(*demoStreamDriver));
    m_fuzzingSimulation.addSimulationProcess(std::ref(*demoStreamVerifier));
    m_fuzzingSimulation.addSimulationProcess(std::ref(*demoCmdGenerator));

    {
        arc::AntiReplayVerifier arVerifier{ clock, cmd, out.get<arc::Valid<BVec>>(), antiReplayCore.windowSize(), false };
        arVerifier.cmdReady(adapter.readyIn());
        arVerifier.outReady(out.ready);
        m_fuzzingSimulation.addSimulationProcess(std::move(arVerifier));
    }

    if (m_basicSettings.writeWaveforms)
        if (!m_basicSettings.waveformPath.empty())
            std::filesystem::create_directories(m_basicSettings.waveformPath);

    if (m_basicSettings.writeWaveforms) {
        m_vcdSinkFuzzing.emplace(DesignScope::get()->getCircuit(), m_fuzzingSimulation, (m_basicSettings.waveformPath / "PacketIFCoreFuzzing.vcd").string().c_str());
        m_vcdSinkFuzzing->addAllPins();
        //m_vcdSinkFuzzing->addAllNamedSignals();
    }

    DesignScope::get()->postprocess();

    m_memoryTraceRecorder.emplace(m_memoryTrace, DesignScope::get()->getCircuit(), m_demonstrationSimulation);
    m_memoryTraceRecorder->addAllPins();    

    if (m_basicSettings.writeWaveforms) {
        m_vcdSinkDemonstration.emplace(DesignScope::get()->getCircuit(), m_demonstrationSimulation, (m_basicSettings.waveformPath / "PacketIFCoreDemonstration.vcd").string().c_str());
        m_vcdSinkDemonstration->addAllPins();
    }

    m_fuzzingSimulation.compileProgram(DesignScope::get()->getCircuit());
    m_demonstrationSimulation.compileProgram(DesignScope::get()->getCircuit());

    m_resetCycles = clock.getClk()->getMinResetCycles();
}

void ARCGenerator::drawPacketIFCore(gtry::reporting::BlockDiagram &diagram, gtry::reporting::Shape &target, gtry::reporting::Shape &input, gtry::reporting::Shape &output)
{
    using namespace gtry::reporting;

    auto &packetStreamIn = diagram.canvas().add<Pin>();
    packetStreamIn.leftOf(&input, 20);
    packetStreamIn.belowOf(&input, 20);
    auto &packetStreamOut = diagram.canvas().add<Pin>();
    packetStreamOut.rightOf(&output, 20);
    packetStreamOut.belowOf(&output, 20);


    auto &coreBlock = target.add<Block>()
        //.setTitle("ESN/SA only IF")
        .setTitle("Inner Anti-Replay Core")
        .setStyle("bd_entity2")
        .setMargin({20.0f, 40.0f, 20.0f, 20.0f});


    gtry::reporting::Shape *coreInputRouting;
    gtry::reporting::Shape *coreOutputRouting;

    //drawAntiReplayCore(diagram, coreBlock, coreInputRouting, coreOutputRouting);
    gtry::hlim::NodeGroup *frameAdapter = gtry::DesignScope::get()->getCircuit().getRootNodeGroup()->findChild("FrameAdapter0");
    gtry::hlim::NodeGroup *arc = frameAdapter->findChild("AntiReplay0");
    arc::drawAntiReplayCore(arc, diagram, coreBlock, coreInputRouting, coreOutputRouting);

    auto &wrapperGroup = target.add<Block>()
        .setStyle("bd_group")
        .setMargin({20.0f, 40.0f, 20.0f, 20.0f})
        .hCenterOn(&coreBlock)
        .belowOf(&coreBlock);

    auto &preFifoLogic = wrapperGroup.add<Block>()
        .incMinWidthHeight(20.0f, 50.0f);



    gtry::hlim::NodeGroup *fifo1Group = frameAdapter->findChild("FrameFifo0")->findChild("scl_fifo0");
    gtry::hlim::NodeGroup *fifo1MemGroup = fifo1Group->findChild("scl_memory0");
    auto &fifo1 = wrapperGroup.add<Block>()
        .setTitle("FIFO")
        .incMinWidthHeight(300.0f, 50.0f)
        .rightOf(&preFifoLogic, 50.0f);

    auto &fifo1Memory = fifo1.add<Block>()
        .setStyle("bd_bram")
        .incMinWidthHeight(160.0f, 20.0f);
    arc::fillInMemoryInfo(fifo1MemGroup, "Fifo Memory", fifo1Memory);


    gtry::hlim::NodeGroup *fifo2Group = frameAdapter->findChild("scl_fifo0");
    gtry::hlim::NodeGroup *fifo2MemGroup = fifo2Group->findChild("scl_memory0");
    auto &fifo2 = wrapperGroup.add<Block>()
        .setTitle("FIFO")
        .incMinWidthHeight(150.0f, 50.0f)
        .hCenterOn(&fifo1)
        .aboveOf(&fifo1, 20);

    auto &fifo2Memory = fifo2.add<Block>()
        .setStyle("bd_bram")
        .incMinWidthHeight(80.0f, 20.0f);
    arc::fillInMemoryInfo(fifo2MemGroup, "Fifo Memory", fifo2Memory);


    auto &postFifoLogic = wrapperGroup.add<Block>()
        .incMinWidthHeight(20.0f, 50.0f)
        .rightOf(&fifo1, 80);

    preFifoLogic.vCenterOn(&fifo1);
    postFifoLogic.vCenterOn(&fifo1);

    diagram.canvas().connect(&input, Shape::SOUTH, &preFifoLogic, Shape::WEST)
        .setName("in_cmd")
        .setToolTip(m_cmdStreamDesc.str())
        .setDstLoc({1,3});

    diagram.canvas().connect(&packetStreamIn, Shape::SOUTH, &preFifoLogic, Shape::WEST)
        .setName("in_packet")
        .setToolTip(m_inputPacketStreamDesc.str())
        .setDstLoc({2,3});

    diagram.canvas().connect(&preFifoLogic, Shape::EAST, &fifo1, Shape::WEST);

    diagram.canvas().connect(&fifo2, Shape::EAST, &postFifoLogic, Shape::WEST)
        .setDstLoc({1,3});
    diagram.canvas().connect(&fifo1, Shape::EAST, &postFifoLogic, Shape::WEST)
        .setDstLoc({2,3});



    diagram.canvas().connect(&postFifoLogic, Shape::EAST, &output, Shape::SOUTH)
        .setName("out_result")
        .setToolTip(m_resultStreamDesc.str())
        .setSrcLoc({2,3});

    diagram.canvas().connect(&postFifoLogic, Shape::EAST, &packetStreamOut, Shape::SOUTH)
        .setName("out_packet")
        .setToolTip(m_outputPacketStreamDesc.str())
        .setSrcLoc({1,3});


    diagram.canvas().connect(&input, Shape::SOUTH, coreInputRouting, Shape::EAST)
        .setName("in_cmd")
        .setToolTip(m_cmdStreamDesc.str())
        .setStyle("bd_connection_no_head");



    auto &routingPoint1 = target.add<ConnectionRouting>()
        .belowOf(&coreBlock, 10.0f)
        .rightOf(&coreBlock, 10.0f);

    auto &routingPoint2 = target.add<ConnectionRouting>()
        .hCenterOn(&preFifoLogic)
        .vCenterOn(&routingPoint1);

    diagram.canvas().connect(coreOutputRouting, Shape::WEST, &routingPoint1, Shape::NORTH)
        .setName("out_result")
        .setToolTip(m_resultStreamDesc.str())
        .setStyle("bd_connection_no_head");


    diagram.canvas().connect(&routingPoint1, Shape::WEST, &routingPoint2, Shape::NORTH)
        .setStyle("bd_connection_no_head");
    diagram.canvas().connect(&routingPoint2, Shape::SOUTH, &fifo2, Shape::WEST);

}



void ARCGenerator::generateCmdDocumentation(arc::CmdStage &cmd)
{
    describeCoreCmdInput(gtry::findInputPin(cmd.valid)->getName().c_str(), true, cmd.valid.getWidth(), 
        "Whether the lookup request is valid.");
    describeCoreCmdInput(gtry::findInputPin(cmd.opcode)->getName().c_str(), false, cmd.opcode.getWidth(), 
        "Indicates the operation to be performed. See <<<Table tbl:cmd_opcodes>>>");
    describeCoreCmdInput(gtry::findInputPin(cmd.connectionIndex)->getName().c_str(), false, cmd.connectionIndex.getWidth(), 
        "For commands referring to a specific connection (OpUpdate, OpActivate, OpDeactivate, OpSetSeqNr), the connection index..");
    describeCoreCmdInput(gtry::findInputPin(cmd.sequenceNumber)->getName().c_str(), false, cmd.sequenceNumber.getWidth(), 
        "In case of OpSetSeqNr the sequence number to set, in case of OpUpdate the new sequence number, otherwise unused.");
    describeCoreCmdInput(gtry::findInputPin(cmd.windowSize)->getName().c_str(), false, cmd.windowSize.getWidth(), 
        "In case of OpSetWindowSize the new virtual window size to use for all connections.");

    {
        auto &table = *m_docuHooks.tables["tbl:cmd_opcodes"];

        table.addRow();
        table(-1, 0) << arc::OpUpdate{}.getName();
        table(-1, 1) << arc::OpUpdate{}.getValue();
        table(-1, 2) << arc::OpUpdate{}.getDescription();

        table.addRow();
        table(-1, 0) << arc::OpActivate{}.getName();
        table(-1, 1) << arc::OpActivate{}.getValue();
        table(-1, 2) << arc::OpActivate{}.getDescription();

        table.addRow();
        table(-1, 0) << arc::OpDeactivate{}.getName();
        table(-1, 1) << arc::OpDeactivate{}.getValue();
        table(-1, 2) << arc::OpDeactivate{}.getDescription();

        table.addRow();
        table(-1, 0) << arc::OpSetSeqNr{}.getName();
        table(-1, 1) << arc::OpSetSeqNr{}.getValue();
        table(-1, 2) << arc::OpSetSeqNr{}.getDescription();

        table.addRow();
        table(-1, 0) << arc::OpSetWindowSize{}.getName();
        table(-1, 1) << arc::OpSetWindowSize{}.getValue();
        table(-1, 2) << arc::OpSetWindowSize{}.getDescription();
    }
}

void ARCGenerator::generateResultDocumentation(gtry::Bit &valid, gtry::BVec &code)
{
    describeCoreResultOutput(gtry::findOutputPin(valid)->getName().c_str(), true, valid.getWidth(), "Whether the output is valid.");
    describeCoreResultOutput(gtry::findOutputPin(code)->getName().c_str(), false, code.getWidth(), "Result code of the operation. See <<<Table tbl:result_opcodes>>>.");

    {
        auto &table = *m_docuHooks.tables["tbl:result_opcodes"];

        table.addRow();
        table(-1, 0) << arc::PassAboveWindow{}.getName();
        table(-1, 1) << arc::PassAboveWindow{}.getValue();
        table(-1, 2) << arc::PassAboveWindow{}.getDescription();

        table.addRow();
        table(-1, 0) << arc::PassCheckDisabled{}.getName();
        table(-1, 1) << arc::PassCheckDisabled{}.getValue();
        table(-1, 2) << arc::PassCheckDisabled{}.getDescription();

        table.addRow();
        table(-1, 0) << arc::PassInVirtualWindow{}.getName();
        table(-1, 1) << arc::PassInVirtualWindow{}.getValue();
        table(-1, 2) << arc::PassInVirtualWindow{}.getDescription();

        table.addRow();
        table(-1, 0) << arc::PassInPhysicalWindow{}.getName();
        table(-1, 1) << arc::PassInPhysicalWindow{}.getValue();
        table(-1, 2) << arc::PassInPhysicalWindow{}.getDescription();

        table.addRow();
        table(-1, 0) << arc::FailBelowWindow{}.getName();
        table(-1, 1) << arc::FailBelowWindow{}.getValue();
        table(-1, 2) << arc::FailBelowWindow{}.getDescription();

        table.addRow();
        table(-1, 0) << arc::FailReplay{}.getName();
        table(-1, 1) << arc::FailReplay{}.getValue();
        table(-1, 2) << arc::FailReplay{}.getDescription();
    }
}

void ARCGenerator::generateInputPacketDocumentation(arc::FrameStream& dataIn)
{
    describeCorePacketInput(gtry::findInputPin(dataIn.valid)->getName().c_str(), true, dataIn.valid.getWidth(), 
            "Indicates that the current beat is a valid one.");

    if (dataIn.has_channel)
    {
        describeCorePacketInput(gtry::findInputPin(dataIn.channel)->getName().c_str(), false, dataIn.channel.getWidth(),
            "Identifies the channel to which the beat belongs.");
    }

    {
        describeCorePacketInput(gtry::findInputPin(dataIn.startofpacket)->getName().c_str(), true, dataIn.startofpacket.getWidth(), 
                "Indicates the start of a packet. Requests issued on the same beat or any following beat until (and including) EOP are considered to happen during the packet.");
        describeCorePacketInput(gtry::findInputPin(dataIn.endofpacket)->getName().c_str(), true, dataIn.endofpacket.getWidth(), 
                "Indicates the last beat of a packet.");
        describeCorePacketInput(gtry::findInputPin(dataIn.empty)->getName().c_str(), false, dataIn.empty.getWidth(), 
                "Can be used to indicate how many words of the payload contain valid data. This is useful for packets whose total size is not a full multiple of the packet stream payload size. The signal is only forwarded but otherwise ignored by the core.");

    }

    if (dataIn.has_error)
    {
        describeCorePacketInput(gtry::findInputPin(dataIn.error)->getName().c_str(), true, dataIn.error.getWidth(),
            "Indicates errors.");
    }

    describeCorePacketInput(gtry::findInputPin(dataIn.data)->getName().c_str(), false, dataIn.data.getWidth(), 
            "Packet payload");

    if (dataIn.has_userdata)
    {
        describeCorePacketInput(gtry::findInputPin(dataIn.userdata)->getName().c_str(), false, dataIn.userdata.getWidth(),
            "User defined data that is passed through.");
    }
}

void ARCGenerator::generateOutputPacketDocumentation(arc::FrameStream& dataOut)
{
    describeCorePacketOutput(gtry::findOutputPin(dataOut.valid)->getName().c_str(), true, dataOut.valid.getWidth(), 
            "Indicates that the current beat is a valid one.");

    if (dataOut.has_channel)
    {
        describeCorePacketOutput(gtry::findOutputPin(dataOut.channel)->getName().c_str(), false, dataOut.channel.getWidth(),
            "Identifies the channel to which the beat belongs.");
    }

    {
        describeCorePacketOutput(gtry::findOutputPin(dataOut.startofpacket)->getName().c_str(), true, dataOut.startofpacket.getWidth(), 
                "Indicates the start of a packet. Requests issued while receiving a packet will yield a result at the earliest on the corresponding beat of the outgoing packet.");
        describeCorePacketOutput(gtry::findOutputPin(dataOut.endofpacket)->getName().c_str(), true, dataOut.endofpacket.getWidth(), 
                "Indicates the last beat of a packet.");
        describeCorePacketOutput(gtry::findOutputPin(dataOut.empty)->getName().c_str(), false, dataOut.empty.getWidth(), 
                "Can be used to indicate how many words of the payload contain valid data. Simply forwarded from the input packet stream.");
    }

    if (dataOut.has_error)
    {
        describeCorePacketOutput(gtry::findOutputPin(dataOut.error)->getName().c_str(), true, dataOut.error.getWidth(),
            "Indicates errors.");
    }

    describeCorePacketOutput(gtry::findOutputPin(dataOut.data)->getName().c_str(), false, dataOut.data.getWidth(), 
            "Packet payload");

    if (dataOut.has_userdata)
    {
        describeCorePacketOutput(gtry::findOutputPin(dataOut.userdata)->getName().c_str(), false, dataOut.userdata.getWidth(),
            "User defined data that is passed through.");
    }
}


void ARCGenerator::describeCoreClock(const char *name, const char *description)
{
    auto &table = *m_docuHooks.tables["tbl:common_signals"];
    table.addRow();
    table(-1, 0) << name;
    table(-1, 1) << "1b";
    table(-1, 2) << "in";
    table(-1, 3) << description;

    m_waveformSignalsAll.push_back({
        .n = name, 
        .style = gtry::reporting::Waveform::defaultClock
    });
    m_waveformSignalsInitialization.push_back(m_waveformSignalsAll.back());
}

void ARCGenerator::describeCoreCommonSignal(const char *name, bool isInput, bool isBit, gtry::BitWidth width, const char *description)
{  
    auto &table = *m_docuHooks.tables["tbl:common_signals"];
    table.addRow();
    table(-1, 0) << name;
    table(-1, 1) << width;
    table(-1, 2) << (isInput?"in":"out");
    table(-1, 3) << description;

    if (isInput)
        m_waveformSignalsAll.push_back({
            .n = name, 
            .style = (isBit?gtry::reporting::Waveform::defaultInBin:gtry::reporting::Waveform::defaultInHex)
        });
    else
        m_waveformSignalsAll.push_back({
            .n = name, 
            .style = (isBit?gtry::reporting::Waveform::defaultBin:gtry::reporting::Waveform::defaultHex)
        });

    m_waveformSignalsInitialization.push_back(m_waveformSignalsAll.back());
}

void ARCGenerator::describeCoreCmdInput(const char *name, bool isBit, gtry::BitWidth width, const char *description)
{
    auto &table = *m_docuHooks.tables["tbl:input_cmd_stream"];
    table.addRow();
    table(-1, 0) << name;
    table(-1, 1) << width;
    table(-1, 2) << "in";
    table(-1, 3) << description;

    m_waveformSignalsAll.push_back({
        .n = name, 
        .style = (isBit?gtry::reporting::Waveform::defaultInBin:gtry::reporting::Waveform::defaultInHex)
    });
    m_waveformSignalsInitialization.push_back(m_waveformSignalsAll.back());

    m_cmdStreamDesc << "\n    " << name << " (" << width << ')';
}

void ARCGenerator::describeCoreResultOutput(const char *name, bool isBit, gtry::BitWidth width, const char *description)
{
    auto &table = *m_docuHooks.tables["tbl:output_out_stream"];
    table.addRow();
    table(-1, 0) << name;
    table(-1, 1) << width;
    table(-1, 2) << "out";
    table(-1, 3) << description;

    m_waveformSignalsAll.push_back({
        .n = name, 
        .style = (isBit?gtry::reporting::Waveform::defaultBin:gtry::reporting::Waveform::defaultHex)
    });

    m_resultStreamDesc << "\n    " << name << " (" << width << ')';
}


void ARCGenerator::describeCorePacketInput(const char *name, bool isBit, gtry::BitWidth width, const char *description)
{
    auto &table = *m_docuHooks.tables["tbl:input_packet_stream"];
    table.addRow();
    table(-1, 0) << name;
    table(-1, 1) << width;
    table(-1, 2) << "in";
    table(-1, 3) << description;

    m_waveformSignalsAll.push_back({
        .n = name, 
        .style = (isBit?gtry::reporting::Waveform::defaultInBin:gtry::reporting::Waveform::defaultInHex)
    });

    m_inputPacketStreamDesc << "\n    " << name << " (" << width << ')';
}

void ARCGenerator::describeCorePacketOutput(const char *name, bool isBit, gtry::BitWidth width, const char *description)
{
    auto &table = *m_docuHooks.tables["tbl:output_packet_stream"];
    table.addRow();
    table(-1, 0) << name;
    table(-1, 1) << width;
    table(-1, 2) << "out";
    table(-1, 3) << description;

    m_waveformSignalsAll.push_back({
        .n = name, 
        .style = (isBit?gtry::reporting::Waveform::defaultBin:gtry::reporting::Waveform::defaultHex)
    });

    m_outputPacketStreamDesc << "\n    " << name << " (" << width << ')';
}
