/*
	Synogate Anti Replay Core, an IP core implementing an algorithm
	designed to detect packet duplication based on packet sequence numbers.
	Copyright (C) 2021 Synogate UG (haftungsbeschraenkt), www.synogate.com 
	
	This program is commercial software: you must not copy, modify or
	distribute it without explicit written license by Synogate to do so. 
	
	This program is distributed WITHOUT ANY WARRANTY; without even the
	implied warranty of MERCHANTABILITY.
*/

#pragma once

#include "Generator.h"

#include <gatery/frontend.h>
#include <gatery/simulation/ReferenceSimulator.h>
#include <gatery/simulation/waveformFormats/VCDSink.h>
#include <gatery/simulation/waveformFormats/MemoryTraceRecorder.h>

#include <synogate/reporting/BlockDiagram.h>

#include "AntiReplay.h"
#include "FrameAdapter.h"

#include <list>

struct Settings {
    enum class Interface {
        esn_sa_only,
        packet
    } interfaceType = Interface::packet;

    void load(gtry::utils::ConfigTree configTree);
};


struct DocuInfo {
    size_t firstStageBRAMSize = 0;
    size_t secondStageMemSizeSize = 0;

    size_t firstStageDataWidth = 0;
    size_t secondStageDataWidth = 0;

};


class ARCGenerator : public Generator {
    public:
        ARCGenerator(gtry::utils::ConfigTree &configTree);

        inline Settings &getSettings() { return m_settings; }
    protected:
        Settings m_settings;

        gtry::sim::ReferenceSimulator m_fuzzingSimulation;
        gtry::sim::ReferenceSimulator m_demonstrationSimulation;

        std::optional<gtry::sim::VCDSink> m_vcdSinkFuzzing;
        std::optional<gtry::sim::VCDSink> m_vcdSinkDemonstration;
        std::optional<gtry::sim::MemoryTraceRecorder> m_memoryTraceRecorder;
        gtry::sim::MemoryTrace m_memoryTrace;

        std::vector<gtry::reporting::Waveform::TraceSignal> m_waveformSignalsInitialization;
        std::vector<gtry::reporting::Waveform::TraceSignal> m_waveformSignalsAll;

        std::stringstream m_cmdStreamDesc;
        std::stringstream m_resultStreamDesc;
        std::stringstream m_inputPacketStreamDesc;
        std::stringstream m_outputPacketStreamDesc;

        std::optional<size_t> m_coreLatency;
        std::optional<size_t> m_packetDataWidth;
        std::optional<size_t> m_packetEmptyWidth;

        std::optional<size_t> m_resetCycles;
        std::optional<bool> m_clockNeedsReset;
        gtry::hlim::Clock *m_clock = nullptr;

        virtual void prepDocu(gtry::reporting::InterfaceDocumentation &doc) override;
        virtual void buildCore() override;
        virtual void prepExport() override;
        virtual void runSimulations() override;
        virtual void finishDocu(gtry::reporting::InterfaceDocumentation &doc) override;

        void documentClocks();
        void documentMemoryIntefaces();

        void buildAntiReplayCore();
        void buildPacketIFCore();

        //void drawAntiReplayCore(gtry::reporting::BlockDiagram &diagram, gtry::reporting::Shape &target, gtry::reporting::Shape *&input, gtry::reporting::Shape *&output);
        void drawPacketIFCore(gtry::reporting::BlockDiagram &diagram, gtry::reporting::Shape &target, gtry::reporting::Shape &input, gtry::reporting::Shape &output);

        void generateCmdDocumentation(arc::CmdStage &cmd);
        void generateResultDocumentation(gtry::Bit &valid, gtry::BVec &code);        
        void generateInputPacketDocumentation(arc::FrameStream& dataIn);
        void generateOutputPacketDocumentation(arc::FrameStream& dataOut);

        void describeCoreClock(const char *name, const char *description);
        void describeCoreCommonSignal(const char *name, bool isInput, bool isBit, gtry::BitWidth width, const char *description);
        void describeCoreCmdInput(const char *name, bool isBit, gtry::BitWidth width, const char *description);
        void describeCoreResultOutput(const char *name, bool isBit, gtry::BitWidth width, const char *description);
        void describeCorePacketInput(const char *name, bool isBit, gtry::BitWidth width, const char *description);
        void describeCorePacketOutput(const char *name, bool isBit, gtry::BitWidth width, const char *description);
};
