/*
	Synogate Anti Replay Core, an IP core implementing an algorithm
	designed to detect packet duplication based on packet sequence numbers.
	Copyright (C) 2021 Synogate UG (haftungsbeschraenkt), www.synogate.com 
	
	This program is commercial software: you must not copy, modify or
	distribute it without explicit written license by Synogate to do so. 
	
	This program is distributed WITHOUT ANY WARRANTY; without even the
	implied warranty of MERCHANTABILITY.
*/

#include "pch.h"
#include "AntiReplay.h"

#include <gatery/hlim/Subnet.h>
#include <gatery/hlim/RegisterRetiming.h>

#include <gatery/hlim/SignalDelay.h>
#include <gatery/export/DotExport.h>

#include <gatery/scl/algorithm/GCD.h>


namespace arc
{



	void resolveReadWriteConflicts(scl::AvalonMM& read, const scl::AvalonMM& write)
	{
		GroupScope entity{ GroupScope::GroupType::ENTITY };
		entity.setName("ReadModifyWriteCoupler");

		HCL_DESIGNCHECK_HINT(!read.readDataValid, "dynamic latency read port is not supported for read write conflict resolution");
		HCL_DESIGNCHECK_HINT(read.readData, "read port needs read data");
		HCL_DESIGNCHECK_HINT(write.write && write.writeData, "write port needs write enable and write data");
		HCL_DESIGNCHECK_HINT(write.writeData->getWidth() == read.readData->getWidth(), "read and write port width mismatch");

		BVec readAddress = read.address;
		BVec overrideData = *write.writeData;
		Bit override = write.address == read.address;

		for (size_t i = 0; i < read.readLatency; ++i)
		{
			GroupScope entity{ GroupScope::GroupType::ENTITY };
			entity.setName("Stage");

			readAddress = reg(readAddress);
			overrideData = reg(overrideData);
			override = reg(override);

			IF(*write.write & readAddress == write.address)
			{
				overrideData = *write.writeData;
				override = '1';
			}

			HCL_NAMED(readAddress);
			HCL_NAMED(overrideData);
			HCL_NAMED(override);
		}

		IF(override)
			*read.readData = overrideData;
	}

	void AntiReplayResult::pinIn(std::string_view prefix)
	{
		valid = gtry::pinIn().setName(std::string(prefix) + "valid");


	}

	void AntiReplayResult::pinOut(std::string_view prefix)
	{
	}

	AntiReplay::AntiReplay() :
		m_area{ "AntiReplay", true }
	{
		m_area.leave();
	}

	AntiReplay::AntiReplay(const CmdStage& cmd) :
		m_area{"AntiReplay", true}
	{
		init(cmd);
		m_area.leave();
	}

	void AntiReplay::init(const CmdStage& cmd)
	{
		auto ent = m_area.enter();

		constructFrom(cmd, m_cmdInput);
		m_cmd = m_cmdInput;
		m_cmdInput = cmd;

		m_cmdInput.valid.setResetValue('0');
		HCL_NAMED(m_cmd);

		m_cfg.connectionAddrWidth = cmd.connectionIndex.getWidth();
		m_cfg.virtualWindowWidth = cmd.windowSize.getWidth();
		m_cfg.sequenceNumberWidth = cmd.sequenceNumber.getWidth();
		m_cfg.load();
		m_cfg.check();
	}

	AntiReplay& AntiReplay::windowSize(BitWidth windowAddrWidth, BitWidth memoryWordWidth)
	{
		m_cfg.m_windowAddrWidth = windowAddrWidth;
		m_cfg.m_secondWordWidth = memoryWordWidth;
		return *this;
	}

	void arc::AntiReplay::generateSingleStage()
	{
		auto ent = m_area.enter();

		m_connectionActive = active();
		std::tie(m_aboveWindow, m_lastSequenceNumber) = lastSequenceNumber();
		std::tie(m_inPhysicalWindow, m_inVirtualWindow) = windowBounds(m_lastSequenceNumber, !m_aboveWindow);

		m_firstValid = firstLookupStage(m_lastSequenceNumber,
			m_cmd.opcode == OpSetSeqNr(),
			m_cmd.opcode == OpSetSeqNr() | (m_cmd.opcode == OpUpdate() & m_connectionActive & (m_aboveWindow | m_inPhysicalWindow)),
			m_cmd.opcode == OpUpdate() & m_connectionActive & m_aboveWindow
		);

		m_isReplay = m_firstValid;
		HCL_NAMED(m_isReplay);

		doRetiming();
	}

	void arc::AntiReplay::generateFixedLatency()
	{
		auto ent = m_area.enter();

		m_connectionActive = active();
		std::tie(m_aboveWindow, m_lastSequenceNumber) = lastSequenceNumber();
		std::tie(m_inPhysicalWindow, m_inVirtualWindow) = windowBounds(m_lastSequenceNumber, !m_aboveWindow);

		Bit updateState = m_cmd.opcode == OpUpdate() & m_connectionActive & (m_aboveWindow | m_inPhysicalWindow);
		IF(m_cmd.opcode == OpSetSeqNr())
			updateState = '1';
		HCL_NAMED(updateState);

		m_firstValid = firstLookupStage(m_lastSequenceNumber,
			m_cmd.opcode == OpSetSeqNr(),
			updateState,
			m_cmd.opcode == OpUpdate() & m_connectionActive & m_aboveWindow
		);

		BVec bitmap = secondStageMemory(m_lastSequenceNumber, updateState, m_aboveWindow, m_firstValid);
		m_isReplay = isReplay(bitmap);
		HCL_NAMED(m_isReplay);

		doRetiming();
	}

	Bit arc::AntiReplay::active() const
	{
		auto ent = m_area.enter("active");

		Memory<Bit> sa_mem(m_cmd.connectionIndex.getWidth().count());
		sa_mem.initZero();
		//HCL_NAMED(sa_mem);
		sa_mem.setName("sa_mem");

		Bit connectionActive = sa_mem[m_cmd.connectionIndex];
		retime(connectionActive, sa_mem.readLatencyHint()); // for sa_mem
		HCL_NAMED(connectionActive);

		Bit set_activate = (m_cmd.opcode == OpActivate()) | (m_cmd.opcode == OpDeactivate());
		HCL_NAMED(set_activate);
		IF(m_cmd.valid & set_activate)
			sa_mem[m_cmd.connectionIndex] = m_cmd.opcode == OpActivate();

		return connectionActive;
	}

	std::tuple<Bit, BVec> arc::AntiReplay::lastSequenceNumber() const
	{
		auto ent = m_area.enter("lastSequenceNumber");

		return lastSequenceNumber(
			m_cmd.connectionIndex, 
			m_cmd.sequenceNumber, 
			m_cmd.valid & m_cmd.opcode == OpSetSeqNr(),
			m_cmd.valid & m_cmd.opcode == OpUpdate() & m_connectionActive,
			limitBitPerCompare()
		);
	}

	std::tuple<Bit, BVec> arc::AntiReplay::lastSequenceNumber(BVec connectionIdx, BVec sequenceNumber, Bit set, Bit updateGreater, size_t bitsPerStep) const
	{
		Bit equal = '1';
		Bit greater = '0';
		BVec lastSequenceNumber = zext(0u, sequenceNumber.getWidth().bits());

		SymbolSelect sym{ bitsPerStep };
		const size_t numStages = sequenceNumber.size() / bitsPerStep;
		(*GroupScope::get())["numStages"] = numStages;
		for (size_t i = numStages - 1; i < numStages; --i)
		{
			auto ent = Area{ "stage" }.enter();

			Memory<BVec> mem(connectionIdx.getWidth().count(), sym.symbolWidth);

			BVec lastSeq = mem[connectionIdx];
			HCL_NAMED(lastSeq);

			lastSequenceNumber(sym[i]) = lastSeq;

			greater |= equal & (sequenceNumber(sym[i]) > lastSeq);
			retime(greater, mem.readLatencyHint());  // for mem
			HCL_NAMED(greater);

			equal &= sequenceNumber(sym[i]) == lastSeq;
			HCL_NAMED(equal);

			IF(set | (updateGreater & greater))
				mem[connectionIdx] = sequenceNumber(sym[i]);
		}

		Bit aboveWindow = greater;
		if(m_cfg.pipelining != Settings::Pipelining::minimum)
			retime(aboveWindow); // critical timing path

		HCL_NAMED(aboveWindow);
		HCL_NAMED(lastSequenceNumber);

		return { aboveWindow, lastSequenceNumber };
	}

	std::tuple<Bit, Bit> arc::AntiReplay::windowBounds(BVec lastSeqNr, Bit belowLastSeqNr) const
	{
		auto ent = m_area.enter("windowBounds");

		BVec windowOffset = lastSeqNr - m_cmd.sequenceNumber;
		HCL_NAMED(windowOffset);
		
		Bit inPhysicalWindow = belowLastSeqNr & (windowOffset < m_cfg.windowSize().bits());
		HCL_NAMED(inPhysicalWindow);

		Bit inVirtualWindow = inPhysicalWindow;
		if (!m_cmd.windowSize.empty())
		{
			HCL_DESIGNCHECK_HINT(m_cmd.windowSize.getWidth().count() >= m_cfg.windowSize().bits(), "virtual window width must hold physical window size");

			BVec virtualWindowSize = m_cmd.windowSize.getWidth();
			IF(m_cmd.valid & (m_cmd.opcode == OpSetWindowSize()))
				virtualWindowSize = m_cmd.windowSize;

			virtualWindowSize = reg(virtualWindowSize, m_cfg.windowSize().bits());
			HCL_NAMED(virtualWindowSize);

			inVirtualWindow = belowLastSeqNr & (zext(windowOffset) < zext(virtualWindowSize));
		}
		HCL_NAMED(inVirtualWindow);

		return { inPhysicalWindow, inVirtualWindow };
	}

	BVec arc::AntiReplay::secondStageMemory(BVec lastSeqNr, Bit updateState, Bit aboveWin, Bit wordValid) const
	{
		auto ent = m_area.enter("SecondLookupStage");

		BVec address = pack(
			m_cmd.connectionIndex,
			m_cmd.sequenceNumber(
				m_cfg.secondWordAddrWidth().bits(), m_cfg.firstWordAddrWidth()
			)
		);
		HCL_NAMED(address);

		Memory<BVec> mem(address.getWidth().count(), m_cfg.secondWordWidth());

		BVec bitmap = ConstBVec(m_cfg.secondWordWidth());
		IF(m_cmd.valid)
			bitmap = mem[address]; // this read should issue in cycle 1 but is delayed by retiming
		retime(bitmap, mem.readLatencyHint()); // for mem
		HCL_NAMED(bitmap);

		HCL_NAMED(updateState);
		IF(m_cmd.valid & updateState)
			mem[address] = secondStageUpdate(bitmap & wordValid, lastSeqNr, aboveWin);

		return bitmap;
	}

	Bit arc::AntiReplay::isReplay(const BVec& bitmap) const
	{
		auto ent = m_area.enter("IsReplay");

		BVec bitAddr = m_cmd.sequenceNumber(0, m_cfg.secondWordAddrWidth());
		HCL_NAMED(bitAddr);

		Bit isReplay = mux_large(bitAddr, bitmap) & m_firstValid;
		HCL_NAMED(isReplay);
		return isReplay;
	}

	BVec arc::AntiReplay::secondStageUpdate(BVec bitmap, BVec lastSeqNr, Bit aboveWindow) const
	{
		auto entity = Area{ "secondStageUpdate" }.enter();
		HCL_NAMED(bitmap);

		Selection segmentAddr = Selection::Slice(
			(int)m_cfg.secondWordAddrWidth().bits(),
			(int)(m_cmd.sequenceNumber.getWidth() - m_cfg.secondWordAddrWidth()).bits()
		);

		IF(m_cmd.opcode == OpSetSeqNr())
			bitmap = 0;
		IF(aboveWindow & m_cmd.sequenceNumber(segmentAddr) != lastSeqNr(segmentAddr))
			bitmap = 0;

		bitmap |= demux(m_cmd.sequenceNumber(0, m_cfg.secondWordAddrWidth()));
		setName(bitmap, "bitmap_next");
		return bitmap;
	}

	BVec arc::AntiReplay::ringBufferMask(BVec put, BVec get) const
	{
		auto ent = Area("ringBufferMask").enter();
		HCL_NAMED(put);
		HCL_NAMED(get);

		std::vector<Bit> maskVec(put.getWidth().count(), '0');

		Bit overrun = get > put;
		HCL_NAMED(overrun);

		for (size_t i = 0; i < maskVec.size(); ++i)
		{
			Bit gl = get < i;
			Bit pg = put > i;

			IF(!overrun)
				maskVec[i] = gl & pg;
			ELSE
				maskVec[i] = gl | pg;
		}

		BVec mask = pack(maskVec);
		HCL_NAMED(mask);
		return mask;
	}

	BVec arc::AntiReplay::mux_large_internal(const BVec& selector, const BVec& data, size_t bitLimit) const
	{
		if (selector.size() > bitLimit && bitLimit)
		{
			BVec subword = mux_large_internal(selector(bitLimit, selector.size() - bitLimit), data, bitLimit);
			retime(subword);
			return muxWord(selector(0, bitLimit), subword);
		}
		else
		{
			return muxWord(selector, data);
		}
	}

	Bit arc::AntiReplay::firstLookupStage(BVec lastSeqNr, Bit clearWindow, Bit setWordValid, Bit updateLastSeqNr) const
	{
		auto ent = m_area.enter("FirstLookupStage");

		// seqNr is split into |-- high bits --|-- first stage address --|-- second stage address --|
		Selection seqNrSlice = Selection::Slice(
			m_cfg.secondWordAddrWidth().bits(), // offset
			m_cfg.firstWordAddrWidth().bits() // width
		);

		// first stage bitmap lookup
		Memory<BVec> mem(
			m_cmd.connectionIndex.getWidth().count(),
			BitWidth{ m_cfg.firstWordAddrWidth().count() }
		);

		BVec bitmap = mem[m_cmd.connectionIndex];
		retime(bitmap, mem.readLatencyHint()); // for mem

		BVec bitAddr = m_cmd.sequenceNumber(seqNrSlice);
		Bit secondStageValid = mux_large(bitAddr, bitmap);
		if (m_cfg.pipelining != Settings::Pipelining::minimum)
			retime(secondStageValid); // critical timing path

		HCL_NAMED(bitmap);
		HCL_NAMED(secondStageValid);
		HCL_NAMED(bitAddr);

		IF(clearWindow)
			bitmap = 0;

		IF(updateLastSeqNr)
		{
			bitmap &= ~ringBufferMask(
				m_cmd.sequenceNumber(seqNrSlice),
				lastSeqNr(seqNrSlice)
			);

			BVec diff = m_cmd.sequenceNumber - lastSeqNr;
			IF(diff >= m_cfg.windowSize().bits())
				bitmap = 0;
		}

		IF(setWordValid)
			bitmap |= demux(bitAddr);

		setName(bitmap, "bitmap_next");
		IF(m_cmd.valid)
			mem[m_cmd.connectionIndex] = bitmap;

		return secondStageValid;
	}

	Bit arc::AntiReplay::mux_large(BVec selector, BVec data) const
	{
		auto ent = Area{ "mux_large" }.enter();
		HCL_NAMED(selector);
		HCL_NAMED(data);

		size_t bitLimit = 0;
		switch (m_cfg.pipelining)
		{
		case Settings::Pipelining::moderate:
			bitLimit = 6;
			break;
		case Settings::Pipelining::aggressive:
			bitLimit = 2;
			break;
		default:;
		}

		bitLimit = ent.config("step_selector_bit_limit").as(bitLimit);

		BVec out = mux_large_internal(selector, data, bitLimit);
		HCL_NAMED(out);
		return out.lsb();
	}

	size_t arc::AntiReplay::limitBitPerCompare() const
	{
		size_t numberWidth = m_cmd.sequenceNumber.size();
		size_t defaultBitsPerStep;
		switch (m_cfg.pipelining)
		{
		case Settings::Pipelining::moderate:
			defaultBitsPerStep = scl::gcd(16, numberWidth);
			break;
		case Settings::Pipelining::aggressive:
			defaultBitsPerStep = scl::gcd(8, numberWidth);
			break;
		default:
			defaultBitsPerStep = 0;
		}

		size_t bitsPerStep = GroupScope::get()->config("step_compare_bit_limit").as(defaultBitsPerStep);
		if (bitsPerStep > numberWidth || bitsPerStep == 0)
			bitsPerStep = numberWidth;

		HCL_DESIGNCHECK_HINT(numberWidth % bitsPerStep == 0, "sequence number with must be divisable by step_compare_bit_limit");
		return bitsPerStep;
	}

	void arc::AntiReplay::retime(const ElementarySignal& signal, size_t numStages) const
	{
		SignalReadPort port = signal.getReadPort();
		
		for(size_t i = 0; i < numStages; ++i)
			m_retimeList.push_back(port);
	}

	void arc::AntiReplay::doRetiming()
	{
		auto& circuit = DesignScope::get()->getCircuit();
		/*
		auto findMemRegisters = [&] {
			auto subnet = hlim::Subnet::all(circuit);
			std::set<hlim::Node_Register*> anchoredRegisters;

			// Anchor all registers driven by memory
			for (auto& n : subnet) {
				if (auto* reg = dynamic_cast<hlim::Node_Register*>(n)) {
					bool drivenByMemory = false;
					std::set<hlim::BaseNode*> alreadyEncountered;
					for (auto nh : reg->exploreInput((unsigned)hlim::Node_Register::Input::DATA)) {
						if (alreadyEncountered.contains(nh.node())) {
							nh.backtrack();
							continue;
						}
						alreadyEncountered.insert(nh.node());
						if (nh.isNodeType<hlim::Node_MemPort>()) {
							drivenByMemory = true;
							break;
						}
						if (!nh.node()->isCombinatorial()) {
							nh.backtrack();
							continue;
						}
					}
					if (drivenByMemory)
						anchoredRegisters.insert(reg);
				}
			}
			return anchoredRegisters;
		};
		*/
		/*
		auto subnet = hlim::Subnet::all(circuit);
		for (auto n : subnet) {
			if (auto* memPort = dynamic_cast<hlim::Node_MemPort*>(n)) {
				if (memPort->isReadPort()) {
					auto subnet2 = hlim::Subnet::all(circuit);
					hlim::retimeForwardToOutput(circuit, subnet2, findMemRegisters(), { .node = memPort, .port = (unsigned)hlim::Node_MemPort::Outputs::rdData }, true);
				}
			}
		}
		*/

		/*
				circuit.optimizeSubnet(subnet);

				subnet = hlim::Subnet::all(circuit);
		*/
		auto ent = m_area.enter();
		auto subnet = hlim::Subnet::all(circuit);
		for (auto it = m_retimeList.rbegin(); it != m_retimeList.rend(); ++it) {

			subnet = hlim::Subnet::fromNodeGroup(m_area.getNodeGroup());
			bool ok = hlim::retimeForwardToOutput(circuit, subnet, *it, { .ignoreRefs = true, .failureIsError = false, .downstreamDisableForwardRT = true });
			if (!ok)
			{
				m_cmdInput = reg(m_cmdInput, { .allowRetimingForward = true });
				m_numPipelineStages++;
				subnet = hlim::Subnet::fromNodeGroup(m_area.getNodeGroup());
				hlim::retimeForwardToOutput(circuit, subnet, *it, { .ignoreRefs = true, .failureIsError = true, .downstreamDisableForwardRT = true });
			}
		}

		// add regular input register
		m_cmdInput = reg(m_cmdInput);
		m_numPipelineStages++;

		m_area["latency"] = m_numPipelineStages;
		m_area["window_size"] = windowSize();
	}


	void arc::AntiReplay::result(AntiReplayResult& out)
	{
		auto ent = m_area.enter("result");

		out.valid = m_cmd.valid & (m_cmd.opcode == OpUpdate());
		out.valid.setResetValue('0');
		out.code = PassCheckDisabled();

		IF(m_connectionActive)
		{
			IF(m_aboveWindow)
			{
				out.code = PassAboveWindow();
			}
			ELSE IF(m_inPhysicalWindow & m_inVirtualWindow)
			{
				IF(m_isReplay)
					out.code = FailReplay();
				ELSE
					out.code = PassInPhysicalWindow();
			}
			ELSE IF(m_inVirtualWindow)
			{
				out.code = PassInVirtualWindow();
			}
			ELSE
			{
				out.code = FailBelowWindow();
			}
		}
		HCL_NAMED(out);
	}

	void CmdStage::pinIn(std::string_view prefix)
	{
		valid = gtry::pinIn().setName(std::string{ prefix } + "valid");
		opcode = gtry::pinIn(opcode.getWidth()).setName(std::string{ prefix } + "opcode");
		connectionIndex = gtry::pinIn(connectionIndex.getWidth()).setName(std::string{ prefix } + "connectionIndex");
		sequenceNumber = gtry::pinIn(sequenceNumber.getWidth()).setName(std::string{ prefix } + "sequenceNumber");
		windowSize = gtry::pinIn(windowSize.getWidth()).setName(std::string{ prefix } + "windowSize");
	}

	void arc::CmdStage::loadConfig()
	{
		GroupScope& scope = *GroupScope::get();

		connectionIndex = scope.config("connection_addr_width").as(1_b);
		windowSize = scope.config("virtual_window_width").as(24_b);
		sequenceNumber = scope.config("sequence_number_width").as(64_b);
	}

	BitWidth Settings::windowSize() const
	{
		return BitWidth{ windowAddrWidth().count() - secondWordWidth().bits() };
	}

	BitWidth Settings::secondWordAddrWidth() const
	{
		return BitWidth::count(secondWordWidth().bits());
	}

	BitWidth Settings::firstWordWidth() const
	{
		return BitWidth{windowAddrWidth().count() / secondWordWidth()};
	}

	BitWidth Settings::firstWordAddrWidth() const
	{
		return BitWidth::count(firstWordWidth().bits());
	}

	void Settings::load()
	{
		GroupScope& scope = *GroupScope::get();

		m_windowAddrWidth = scope.config("window_addr_width").as(m_windowAddrWidth);
		m_secondWordWidth = scope.config("memory_word_width").as(m_secondWordWidth);
		pipelining = scope.config("pipelining").as(pipelining);
	}
	
	void arc::Settings::check()
	{
		size_t windowStorageBits = windowAddrWidth().count();
		size_t windowStorageWordWidth = secondWordWidth().bits();

		HCL_ASSERT_HINT(windowStorageWordWidth > 1, "Second stage memory must have at least two bits. Please contact us for single stage lookup support.");
		HCL_ASSERT_HINT(windowStorageBits > 1, "Window size must hold at least two items. Please contact us for single item lookup support.");
		HCL_ASSERT_HINT(windowStorageBits > windowStorageWordWidth, "Window storage must be at least two second stage memory words in size. Increase window_addr_width or reduce memory_word_width.");
		HCL_ASSERT_HINT(windowStorageBits % windowStorageWordWidth == 0, "Window storage must be a mutltiple of second stage memory word width. Change window_addr_width or memory_word_width to meet this requirement.");
	}
}
