/*
	Synogate Anti Replay Core, an IP core implementing an algorithm
	designed to detect packet duplication based on packet sequence numbers.
	Copyright (C) 2021 Synogate UG (haftungsbeschraenkt), www.synogate.com 
	
	This program is commercial software: you must not copy, modify or
	distribute it without explicit written license by Synogate to do so. 
	
	This program is distributed WITHOUT ANY WARRANTY; without even the
	implied warranty of MERCHANTABILITY.
*/

#pragma once
#include <gatery/frontend.h>
#include <gatery/scl/Avalon.h>
#include <gatery/utils/ConfigTree.h>

#include <boost/property_tree/ptree.hpp>

namespace arc
{
	struct Settings
	{
		BitWidth windowAddrWidth() const { return m_windowAddrWidth; }
		BitWidth secondWordWidth() const { return m_secondWordWidth; }

		BitWidth windowSize() const;
		BitWidth secondWordAddrWidth() const;
		BitWidth firstWordWidth() const;
		BitWidth firstWordAddrWidth() const;

		BitWidth m_windowAddrWidth = 12_b;
		BitWidth m_secondWordWidth = 128_b;

		BitWidth connectionAddrWidth = 9_b;
		BitWidth virtualWindowWidth = 17_b;
		BitWidth sequenceNumberWidth = 64_b;

		enum class Pipelining {
			minimum,
			moderate,
			aggressive
		} pipelining = Pipelining::moderate;

		void load();
		void check();
	};

	GTRY_CONST_BVEC_DESC(PassAboveWindow,		"b001", 	"The packet is deemed new because its sequence number was larger than the largest so far seen sequence number for that connection.");
	GTRY_CONST_BVEC_DESC(PassCheckDisabled,		"b011", 	"The packet is deemed new because actually checking was disabled for that connection.");
	GTRY_CONST_BVEC_DESC(PassInVirtualWindow,	"b101", 	"The packet is deemed new by definition because its sequence number is older than the physical window size but still within the virtual window size. This never occurs if the virtual window size is set to be equal or smaller than the physical one.");
	GTRY_CONST_BVEC_DESC(PassInPhysicalWindow,	"b111",		"The packet is deemed new because it falls within the virtual and physical windows and had not been marked as having been seen before.");
	GTRY_CONST_BVEC_DESC(FailBelowWindow,		"b000", 	"The packet is deemed a replay because it is older than the virtual window size.");
	GTRY_CONST_BVEC_DESC(FailReplay,			"b010",		"The packer is deemed a replay because it fell within the virtual and physical windows and had been marked as having been seen before.");

	GTRY_CONST_BVEC_DESC(OpUpdate,			"b000", 	"Initiates an actual replay check for the given sequence number and connection index, triggering a reply on the result stream. Also updates the internal state as having received that packet.");
	GTRY_CONST_BVEC_DESC(OpActivate,		"b101", 	"Activates replay checks for a given connection index. OpUpdate querries on that connection index will no longer return PassCheckDisabled.");
	GTRY_CONST_BVEC_DESC(OpDeactivate,		"b001", 	"Deactivates replay checks for a given connection index. OpUpdate querries on that connection index will return PassCheckDisabled.");
	GTRY_CONST_BVEC_DESC(OpSetSeqNr,		"b010", 	"For a given connection index, clears and initializes the list of all seen packets to the given sequence number, so that all other sequence numbers are deemed unseen. Connections are disabled and uninitialized on reset. Use OpSetSeqNr and a sequence number of 0 to initialize a connection.");
	GTRY_CONST_BVEC_DESC(OpSetWindowSize,	"b011", 	"Sets the virtual window size for all connections. This can also be larger than the actual physical size. The reset value equals the physical window size to effectively disable the virtual window feature.");

	struct CmdStage
	{
		CmdStage() = default;

		Bit valid;
		BVec opcode = 3_b;
		BVec connectionIndex;
		BVec sequenceNumber;
		BVec windowSize;

		void pinIn(std::string_view prefix);
		void loadConfig();
	};

	struct AntiReplayResult
	{
		Bit valid;
		BVec code;

		void pinIn(std::string_view prefix);
		void pinOut(std::string_view prefix);
	};

	void resolveReadWriteConflicts(scl::AvalonMM& read, const scl::AvalonMM& write);

	class AntiReplay
	{
	public:

	public:
		AntiReplay(const CmdStage& cmd);

		AntiReplay& windowSize(BitWidth windowAddrWidth, BitWidth memoryWordWidth);
		BitWidth windowSize() const { return m_cfg.windowSize(); }

		virtual void generateSingleStage();
		virtual void generateFixedLatency();
		
		void result(AntiReplayResult& out);
		size_t latency() const { return m_numPipelineStages; }

		// methods public for unit tests
		virtual Bit active() const;
		virtual std::tuple<Bit, BVec> lastSequenceNumber() const;
		virtual std::tuple<Bit, BVec> lastSequenceNumber(BVec connectionIdx, BVec sequenceNumber, Bit set, Bit updateGreater, size_t bitsPerStep) const;

	protected:
		AntiReplay();
		void init(const CmdStage& cmd);

		virtual std::tuple<Bit,Bit> windowBounds(BVec lastSeqNr, Bit belowLastSeqNr) const;
		virtual BVec ringBufferMask(BVec put, BVec get) const;
		virtual Bit mux_large(BVec selector, BVec data) const;
		virtual BVec mux_large_internal(const BVec& selector, const BVec& data, size_t bitLimit) const;
		virtual Bit firstLookupStage(BVec lastSeqNr, Bit clearWindow, Bit setWordValid, Bit updateLastSeqNr) const;
		virtual BVec secondStageMemory(BVec lastSeqNr, Bit updateState, Bit aboveWin, Bit wordValid) const;
		virtual BVec secondStageUpdate(BVec bitmap, BVec lastSeqNr, Bit aboveWindow) const;
		virtual Bit isReplay(const BVec& bitmap) const;

		size_t limitBitPerCompare() const;

		void retime(const ElementarySignal& signal, size_t numStages = 1) const;
		void doRetiming();

	protected:
		Area m_area;
		Settings m_cfg;
		CmdStage m_cmd;
		CmdStage m_cmdInput;
		size_t m_numPipelineStages = 0;

		mutable std::vector<hlim::NodePort> m_retimeList;

		// genActive
		Bit m_connectionActive;

		// genLastSequenceNumber
		BVec m_lastSequenceNumber;
		Bit m_aboveWindow;

		// genWindowBounds
		Bit m_inPhysicalWindow;
		Bit m_inVirtualWindow;

		// first lookup stage
		Bit m_firstValid;

		// result
		Bit m_isReplay;
	};
}

BOOST_HANA_ADAPT_STRUCT(arc::Settings, windowSize, secondWordWidth);
BOOST_HANA_ADAPT_STRUCT(arc::CmdStage, valid, opcode, connectionIndex, sequenceNumber, windowSize);
BOOST_HANA_ADAPT_STRUCT(arc::AntiReplayResult, valid, code);
