/*
	Synogate Anti Replay Core, an IP core implementing an algorithm
	designed to detect packet duplication based on packet sequence numbers.
	Copyright (C) 2021 Synogate UG (haftungsbeschraenkt), www.synogate.com 
	
	This program is commercial software: you must not copy, modify or
	distribute it without explicit written license by Synogate to do so. 
	
	This program is distributed WITHOUT ANY WARRANTY; without even the
	implied warranty of MERCHANTABILITY.
*/

#include "pch.h"
#include "AntiReplayDocu.h"


#include <gatery/utils/Range.h>
#include <synogate/reporting/BlockDiagram.h>


#include <external/magic_enum.hpp>
#include <boost/format.hpp>

namespace arc
{


void fillInMemoryInfo(gtry::hlim::NodeGroup *memory, const std::string &name, gtry::reporting::Block &target)
{
    target.setTitle(name);

    std::stringstream tooltip;
    tooltip 
        << name 
        << "\n   Path: " << memory->instancePath()
        << "\n   Geometry: " << memory->properties()["num_words"] << 'x' << memory->properties()["word_width"] 
                    << " (" << BitWidth(memory->properties()["num_words"].as<size_t>() * memory->properties()["word_width"].as<size_t>()) << ')'
        << "\n   Latency: " << memory->properties()["readLatency"]
        << "\n   Type: " << memory->properties()["type"]
        << "\n   Primitive: " << memory->properties()["primitive"].as<std::string>("");

    target.setToolTip(tooltip.str());

}

void drawAntiReplayCore(gtry::hlim::NodeGroup *arc, gtry::reporting::BlockDiagram &diagram, gtry::reporting::Shape &target, gtry::reporting::Shape *&input, gtry::reporting::Shape *&output)
{
    using namespace gtry::reporting;

    auto *secondStageGroup = arc->findChild("SecondLookupStage0");
    auto *secondStageMemoryGroup = secondStageGroup->findChild("scl_memory0");

    auto &secondStage = target.add<Block>()
            .setTitle("Second Stage")
            .setToolTip("Second stage of the physical window implementation. This stage actually holds a bitmap with one bit per window entry.")
            .setMargin({20.0f, 40.0f, 20.0f, 20.0f});

    if (secondStageMemoryGroup->properties()["type"].as<std::string>() == "EXTERNAL") {
        auto &secondStageMemory = target.add<Block>()
                .setStyle("bd_entity2")
                .aboveOf(&secondStage, 20.0f)
                .hCenterOn(&secondStage);    

        fillInMemoryInfo(secondStageMemoryGroup, "External Memory", secondStageMemory);

        std::vector<Pin*> allPins;
        
        auto &props = secondStageMemoryGroup->findChild("physical_memory0")->properties();

        if (props["numPorts"].as<size_t>() != 2) {
            // Something is seriously wrong, the external memory shouldhave 2 ports, but we don't want to fail on docu, so ignore this part
        } else {

            if (props["port_0_has_readEnable"].as<bool>())
                allPins.push_back(&(secondStageMemory.add<Pin>()
                    .setToolTip(props["port_0_pinName_readEnable"].as<std::string>())
                ));

            allPins.push_back(&(secondStageMemory.add<Pin>()
                .setToolTip(props["port_0_pinName_addr"].as<std::string>())
            ));

            allPins.push_back(&(secondStageMemory.add<Pin>()
                .setToolTip(props["port_0_pinName_readData"].as<std::string>())
            ));

            if (props["port_1_has_writeEnable"].as<bool>())
                allPins.push_back(&(secondStageMemory.add<Pin>()
                    .setToolTip(props["port_1_pinName_writeEnable"].as<std::string>())
                ));

            allPins.push_back(&(secondStageMemory.add<Pin>()
                .setToolTip(props["port_1_pinName_addr"].as<std::string>())
            ));

            allPins.push_back(&(secondStageMemory.add<Pin>()
                .setToolTip(props["port_1_pinName_writeData"].as<std::string>())
            ));

            for (auto i : utils::Range<size_t>(1, allPins.size()))
                allPins[i]->rightOf(allPins[i-1], 4.0f);

            auto &routingPoint = secondStageMemory.add<ConnectionRouting>()
                .belowOf(allPins.front(), 20.0f)
                .hCenterOn(&secondStageMemory);

            for (auto i : utils::Range<size_t>(0, allPins.size()/2))
                diagram.canvas().connect(&routingPoint, Shape::WEST, allPins[i], Shape::SOUTH)
                    .setStyle("bd_connection_no_head");

            for (auto i : utils::Range(allPins.size()/2, allPins.size()))
                diagram.canvas().connect(&routingPoint, Shape::EAST, allPins[i], Shape::SOUTH)
                    .setStyle("bd_connection_no_head");

            diagram.canvas().connect(&secondStage, Shape::NORTH, &routingPoint, Shape::SOUTH)
                .setStyle("bd_connection_no_head")
                .setWidth(secondStageMemoryGroup->properties()["word_width"].as<std::string>() + " bit word width");
        }

    } else {
        auto &secondStageMemory = target.add<Block>()
                .setStyle("bd_bram")
                .incMinWidthHeight(80.0f, 20.0f)
                .aboveOf(&secondStage, 20.0f)
                .hCenterOn(&secondStage);
        fillInMemoryInfo(secondStageMemoryGroup, "Memory", secondStageMemory);

        diagram.canvas().connect(&secondStage, Shape::NORTH, &secondStageMemory, Shape::SOUTH)
            .setStyle("bd_connection_no_head")
            .setWidth(secondStageMemoryGroup->properties()["word_width"].as<std::string>() + " bit word width");

    }
    


    const size_t numPipelineStages = arc->properties()["latency"].as<size_t>();

    auto &inputRouting = target.add<ConnectionRouting>()
        .setHorizontal(5.0f)
        .vCenterOn(&secondStage);
    input = &inputRouting;

    Shape *lastPipeline = &inputRouting;

    for (auto i = 0; i < numPipelineStages; i++) {
        auto &reg = target.add<gtry::reporting::Register>()
            .incMinWidthHeight(10.0f, 5.0f)
            .rightOf(lastPipeline, numPipelineStages > 10?10.0f:16.0f)
            .vCenterOn(&secondStage)
            .setToolTip("Pipeline register, actually distributed throughout the core.\n    Total latency: " + std::to_string(numPipelineStages));

        diagram.canvas().connect(lastPipeline, Shape::EAST, &reg, Shape::WEST);

        lastPipeline = &reg;
    }




    auto *lastSequenceNumberGroup = arc->findChild("lastSequenceNumber0");
    auto *lastSequenceNumberMemoryGroup = lastSequenceNumberGroup->findChild("stage0")->findChild("scl_memory0");
    auto numLastSeqNumberStages = lastSequenceNumberGroup->properties()["numStages"].as<size_t>();

    auto &controlLogic = target.add<Block>()
            .setTitle("Control Logic")
            .setToolTip((boost::format("Control logic w. memories for holding connection status (activation, highest sequence number).\nHighest sequence number memory and comparisons are split into %d stages, but only the memory for one is shown here.")
                        % numLastSeqNumberStages).str())
            .rightOf(lastPipeline, 50.0f)
            .vCenterOn(&secondStage)
            .setMargin({20.0f, 40.0f, 20.0f, 20.0f});

    auto &controlLogicMemGroup = target.add<Group>()
        .hCenterOn(&controlLogic)
        .aboveOf(&controlLogic, 5.0f);

    auto &activeMemory = controlLogicMemGroup.add<Block>()
            .setStyle("bd_bram")
            .incMinWidthHeight(80.0f, 20.0f);

    auto *activeMemoryGroup = arc->findChild("active0")->findChild("scl_memory0");
    fillInMemoryInfo(activeMemoryGroup, "Memory", activeMemory);            

    diagram.canvas().connect(&controlLogic, Shape::NORTH, &activeMemory, Shape::SOUTH)
        .setStyle("bd_connection_no_head")
        .setSrcLoc({2,3});

    auto &lastSeqNumberMem = controlLogicMemGroup.add<Block>()
            .setStyle("bd_bram")
            .incMinWidthHeight(80.0f, 20.0f)
            .rightOf(&activeMemory, 5.0f);

    fillInMemoryInfo(lastSequenceNumberMemoryGroup, 
                        (boost::format("Memory x%d") % numLastSeqNumberStages).str(), 
                        lastSeqNumberMem);            

    diagram.canvas().connect(&controlLogic, Shape::NORTH, &lastSeqNumberMem, Shape::SOUTH)
        .setStyle("bd_connection_no_head")
        .setSrcLoc({1,3});



    auto &firstStage = target.add<Block>()
            .setTitle("First Stage")
            .setToolTip("First stage of the physical window implementation.")
            .setMargin({20.0f, 40.0f, 20.0f, 20.0f})
            .rightOf(&controlLogic, 50.0f)
            .vCenterOn(&secondStage);

    auto &firstStageBRam = target.add<Block>()
            .setStyle("bd_bram")
            .incMinWidthHeight(80.0f, 20.0f)
            .hCenterOn(&firstStage)
            .aboveOf(&firstStage, 5.0f);

    auto *firstStageMemoryGroup = arc->findChild("FirstLookupStage0")->findChild("scl_memory0");
    fillInMemoryInfo(firstStageMemoryGroup, "Memory", firstStageBRam);            

    diagram.canvas().connect(&firstStage, Shape::NORTH, &firstStageBRam, Shape::SOUTH)
        .setStyle("bd_connection_no_head");




    diagram.canvas().connect(lastPipeline, Shape::EAST, &controlLogic, Shape::WEST);

    lastPipeline = &controlLogic;

    lastPipeline->leftOf(&firstStage, 20.0f);

    diagram.canvas().connect(lastPipeline, Shape::EAST, &firstStage, Shape::WEST);

    lastPipeline = &firstStage;

    lastPipeline->leftOf(&secondStage, 20.0f);

    diagram.canvas().connect(lastPipeline, Shape::EAST, &secondStage, Shape::WEST);

    lastPipeline = &secondStage;

    auto &outputRouting = target.add<ConnectionRouting>()
        .rightOf(lastPipeline, 50.0f)
        .vCenterOn(&secondStage);
    output = &outputRouting;

    diagram.canvas().connect(lastPipeline, Shape::EAST, &outputRouting, Shape::WEST)
        .setStyle("bd_connection_no_head");



}

}