/*
	Synogate Anti Replay Core, an IP core implementing an algorithm
	designed to detect packet duplication based on packet sequence numbers.
	Copyright (C) 2021 Synogate UG (haftungsbeschraenkt), www.synogate.com 
	
	This program is commercial software: you must not copy, modify or
	distribute it without explicit written license by Synogate to do so. 
	
	This program is distributed WITHOUT ANY WARRANTY; without even the
	implied warranty of MERCHANTABILITY.
*/

#include "pch.h"
#include "FrameAdapter.h"

namespace arc
{
	AntiReplay& FrameAdapter::setupAntiReplay(const CmdStage& cmd)
	{
		HCL_ASSERT(!m_ar);
		auto ent = m_area.enter();
		
		m_cmd = cmd;
		m_cmd.valid &= m_inReady;

		m_ar.emplace(m_cmd);
		return *m_ar;
	}

	FrameStream& FrameAdapter::dataIn(BitWidth width)
	{
		auto ent = m_area.enter();

		m_frameIn.setup(width);
		return m_frameIn;
	}

	Stream<std::tuple<FrameStream, Valid<BVec>>>& FrameAdapter::generate()
	{
		auto ent = m_area.enter();

		Stream<std::tuple<FrameStream, Bit>> frameIn{ .data = std::make_tuple(
			m_frameIn,
			m_cmd.valid & (m_cmd.opcode == OpUpdate())
		)};

		// valid if data or update command present
		frameIn.valid = m_frameIn.valid | frameIn.get<Bit>();
		frameIn.valid &= m_inReady;
		setName(m_inReady, "dbg_dataIn2");
		setName(m_frameIn.valid, "dbg_dataIn3");
		setName(frameIn.get<Bit>(), "dbg_dataIn4");
		setName(m_cmd, "m_cmd");

		Stream<std::tuple<FrameStream, Bit>> frameFifo;
		{
			auto ent = Area{ "FrameFifo" }.enter();
			frameIn.fifo(m_ar->latency() * 3, 0, frameFifo);
		}

		AntiReplayResult arResult;
		m_ar->result(arResult);

		Stream<BVec> arResultStream{
			.valid = arResult.valid,
			.data = arResult.code
		};
		Stream<BVec> arResultFifo;
		arResultStream.fifo(m_ar->latency() * 3, m_ar->latency(), arResultFifo);

		m_out.get<FrameStream>() = constructFrom(m_frameIn);
		merge(frameFifo, arResultFifo, m_out);

		// mask individual valids by stream valid
		m_out.get<FrameStream>().valid &= m_out.valid;
		m_out.get<Valid<BVec>>().valid &= m_out.valid;

		m_inReady = frameIn.ready & arResultStream.ready;
		HCL_NAMED(m_inReady);

		// TODO: we need the enable scope :(
		{
			std::tuple<FrameStream, Valid<BVec>> data_reg = constructFrom(m_out.data);
			Bit valid_reg;

			get<0>(data_reg).valid.setResetValue('0');
			get<1>(data_reg).valid.setResetValue('0');

			data_reg = reg(data_reg);
			valid_reg = reg(valid_reg, '0');
			setName(data_reg, "output_reg_data");
			setName(valid_reg, "output_reg_valid");

			std::tuple<FrameStream, Valid<BVec>> data_tmp = m_out.data;
			Bit valid_tmp = m_out.valid;
			m_out.data = data_reg;
			m_out.valid = valid_reg;

			IF(m_out.ready)
			{
				data_reg = data_tmp;
				valid_reg = valid_tmp;
			}
		}

		return m_out;
	}

	void FrameAdapter::assignPins()
	{
		pinOut(m_inReady).setName("in_ready");
		m_frameIn.pinIn("in_frame_");

		m_out.ready = pinIn().setName("out_ready");
		m_out.get<FrameStream>().pinOut("out_frame_");
		pinOut(m_out.get<Valid<BVec>>().valid).setName("out_result_valid");
		pinOut(m_out.get<Valid<BVec>>().data).setName("out_result_data");
	}

}

