/*
	Synogate Anti Replay Core, an IP core implementing an algorithm
	designed to detect packet duplication based on packet sequence numbers.
	Copyright (C) 2021 Synogate UG (haftungsbeschraenkt), www.synogate.com 
	
	This program is commercial software: you must not copy, modify or
	distribute it without explicit written license by Synogate to do so. 
	
	This program is distributed WITHOUT ANY WARRANTY; without even the
	implied warranty of MERCHANTABILITY.
*/

#pragma once
#include "AntiReplay.h"
#include "FrameStream.h"
#include <gatery/scl/Fifo.h>

namespace arc
{
	template<typename T>
	struct Valid
	{
		BOOST_HANA_DEFINE_STRUCT(Valid,
			(Bit, valid),
			(T, data)
		);
	};

	template<typename T>
	struct Stream
	{
		Bit ready;
		Bit valid;
		T data;

		template<typename T2>
		T2& get() { return std::get<T2>(data); }
		template<typename T2>
		const T2& get() const { return std::get<T2>(data); }

		//void operator <<= (Ready& master);

		void fifo(size_t capacity, size_t readyAllowance, Stream<T>& out);
		Valid<T> asValid() const { return { valid, data }; }

		void pinIn(std::string_view prefix)
		{
			gtry::pinOut(ready).setName(std::string{ prefix } + "ready");
			valid = gtry::pinIn().setName(std::string{ prefix } + "valid");
		}

		void pinOut(std::string_view prefix)
		{
			ready = gtry::pinIn().setName(std::string{ prefix } + "ready");
			gtry::pinOut(valid).setName(std::string{ prefix } + "valid");
		}
	};

	template<typename T>
	inline void Stream<T>::fifo(size_t capacity, size_t readyAllowance, Stream<T>& ret)
	{
		HCL_DESIGNCHECK(capacity >= readyAllowance);

		scl::Fifo<T> fifo(capacity, data);
		if (!readyAllowance)
		{
			ready = !fifo.full();
			fifo.push(data, ready & valid);
		}
		else
		{
			ready = !fifo.almostFull(readyAllowance);
			fifo.push(data, !fifo.full() & valid);
		}

		ret.valid = !fifo.empty();
		fifo.pop(ret.data, ret.ready & ret.valid);
	}

	template<typename Tresult>
	void merge(
		Stream<std::tuple<FrameStream, Bit>>& data, 
		Stream<Tresult>& result, 
		Stream<std::tuple<FrameStream, Valid<Tresult>>>& out
	)
	{
		auto ent = Area{ "merge" }.enter();

		data.ready = out.ready;
		out.valid = data.valid;
		out.template get<FrameStream>() = data.template get<FrameStream>();

		Valid<Tresult>& outResult = out.template get<Valid<Tresult>>();
		result.ready = out.ready;
		outResult.valid = result.valid;
		outResult.data = result.data;
		
		IF(data.valid)
		{
			IF(!data.get<Bit>())
			{
				// stall result while not requested
				result.ready = '0';
				outResult.valid = '0';
			}
			IF(data.get<Bit>() & !result.valid)
			{
				// stall data while result requested and not present
				data.ready = '0';
				out.valid = '0';
			}
		}
	}

	class FrameAdapter
	{
	public:
		// 1. setup AntiReplay core and dataIn stream
		AntiReplay& setupAntiReplay(const CmdStage& cmd);
		FrameStream& dataIn() { return m_frameIn; }
		FrameStream& dataIn(BitWidth width);
		const Bit& readyIn() const { return m_inReady; }

//		CmdStage& cmd() { return m_cmd; }

		// 2. build fifo's
		Stream<std::tuple<FrameStream, Valid<BVec>>>& generate();

		// 3. assign pins
		void assignPins();

	private:
		Area m_area{ "FrameAdapter" };

		FrameStream m_frameIn;
		CmdStage m_cmd;
		std::optional<AntiReplay> m_ar;

		Bit m_inReady;

		Stream<std::tuple<FrameStream, Valid<BVec>>> m_out;

	};

}

