/*
	Synogate Anti Replay Core, an IP core implementing an algorithm
	designed to detect packet duplication based on packet sequence numbers.
	Copyright (C) 2021 Synogate UG (haftungsbeschraenkt), www.synogate.com 
	
	This program is commercial software: you must not copy, modify or
	distribute it without explicit written license by Synogate to do so. 
	
	This program is distributed WITHOUT ANY WARRANTY; without even the
	implied warranty of MERCHANTABILITY.
*/

#include "pch.h"
#include "FrameStream.h"

#include <random>

namespace arc
{
	Bit FrameStream::in_frame() const
	{
		Bit is_frame;

		IF(valid & endofpacket)
			is_frame = '0';

		is_frame = reg(is_frame, '0');

		IF(valid & startofpacket)
			is_frame = '1';

		return is_frame;
	}


	void FrameStream::setup(BitWidth width)
	{
		data = width;
		empty = BitWidth::count(width.bits() / 8);

		channel = 0_b;
		error = 0_b;
		userdata = 0_b;
	}

	void FrameStream::setup(gtry::utils::ConfigTree config)
	{
		symbolWidth = config["dataBitsPerSymbol"].as(symbolWidth);
		emptyWithinPacket = config["emptyWithinPacket"].as(emptyWithinPacket);
		errorDescriptor = config["errorDescriptor"].as(errorDescriptor);
		firstSymbolInHighOrderBits = config["firstSymbolInHighOrderBits"].as(firstSymbolInHighOrderBits);
		maxChannel = config["maxChannel"].as(maxChannel);
		readyLatency = config["readyLatency"].as(readyLatency);
		readyAllowance = config["readyAllowance"].as(readyLatency); // default for readyAllowance is readyLatency

		if (readyAllowance < readyLatency)
			gtry::trace() << "FrameStream readyAllowance must be at least readyLatency.";
		readyAllowance = std::max(readyAllowance, readyLatency);

		std::optional<BitWidth> channel_width;
		if (config["channel_width"])
			channel_width = config["channel_width"].as<BitWidth>();

		BitWidth data_width = config["data_width"].as(16_b);

		std::optional<BitWidth> error_width;
		if (config["error_width"])
			error_width = config["error_width"].as<BitWidth>();

		std::optional<BitWidth> userdata_width;
		if (config["userdata_width"])
			userdata_width = config["userdata_width"].as<BitWidth>();

		// channel
		if (!channel_width && maxChannel)
			channel_width = BitWidth::count(maxChannel);
		if (channel_width)
		{
			HCL_ASSERT(*channel_width >= BitWidth::count(maxChannel));
			channel = *channel_width;
			has_channel = true;
		}
		else
		{
			channel = 0_b;
		}

		// data
		data = data_width;

		// error
		if (!error_width && !errorDescriptor.empty())
			error_width = BitWidth{ errorDescriptor.size() };
		if (error_width)
		{
			error = *error_width;
			has_error = true;
		}
		else
		{
			error = 0_b;
		}

		// empty
		HCL_DESIGNCHECK(data_width.bits() % symbolWidth.bits() == 0);
		HCL_DESIGNCHECK(data_width.bits() >= symbolWidth.bits());
		empty = BitWidth::count(data_width / symbolWidth);

		// userdata
		if (userdata_width)
		{
			userdata = *userdata_width;
			has_userdata = true;
		}
		else
		{
			userdata = 0_b;
		}
	}

	void FrameStream::pinIn(std::string_view prefix)
	{
		valid = gtry::pinIn().setName(std::string{ prefix } + "valid");
		data = gtry::pinIn(data.getWidth()).setName(std::string{ prefix } + "data");

		startofpacket = gtry::pinIn().setName(std::string{ prefix } + "startofpacket");
		endofpacket = gtry::pinIn().setName(std::string{ prefix } + "endofpacket");
		empty = gtry::pinIn(empty.getWidth()).setName(std::string{ prefix } + "empty");

		if (has_error)
		{
			error = gtry::pinIn(error.getWidth()).setName(std::string{ prefix } + "error");
		}

		if (has_channel)
		{
			channel = gtry::pinIn(channel.getWidth()).setName(std::string{ prefix } + "channel");
		}

		if (has_userdata)
		{
			userdata = gtry::pinIn(userdata.getWidth()).setName(std::string{ prefix } + "userdata");
		}
	}

	void FrameStream::pinOut(std::string_view prefix)
	{
		gtry::pinOut(valid).setName(std::string{ prefix } + "valid");
		gtry::pinOut(data).setName(std::string{ prefix } + "data");

		gtry::pinOut(startofpacket).setName(std::string{ prefix } + "startofpacket");
		gtry::pinOut(endofpacket).setName(std::string{ prefix } + "endofpacket");
		gtry::pinOut(empty).setName(std::string{ prefix } + "empty");

		if (has_error)
		{
			gtry::pinOut(error).setName(std::string{ prefix } + "error");
		}

		if (has_channel)
		{
			gtry::pinOut(channel).setName(std::string{ prefix } + "channel");
		}

		if (has_userdata)
		{
			gtry::pinOut(userdata).setName(std::string{ prefix } + "userdata");
		}
	}

	FrameStreamTestDriver::FrameStreamTestDriver(const Clock& clock, const FrameStream& stream, const Bit& ready) :
		m_clock(clock),
		m_stream(stream),
		m_ready(ready)
	{
	}

	void FrameStreamTestDriver::push(std::vector<uint8_t> data)
	{
		HCL_ASSERT(data.size() % m_stream.symbolWidth.bytes() == 0);
		m_tasks.push(std::move(data));
	}

	void FrameStreamTestDriver::pushRandom(size_t numFrames, uint32_t seed)
	{
		std::mt19937 rng{ seed };
		for (size_t i = 0; i < numFrames; ++i)
			pushRandom(rng);
	}

	const std::vector<uint8_t>& FrameStreamTestDriver::pushBeats(std::span<const uint64_t> beatData)
	{
		size_t bpb = m_stream.data.size() / 8;
		size_t bpbData = std::min(bpb, sizeof(uint64_t));

		std::vector<uint8_t> data;
		for (uint64_t beat : beatData)
		{
			size_t i;
			for (i = 0; i < bpbData; ++i)
				data.push_back(uint8_t(beat >> i * 8));
			for (; i < bpb; ++i)
				data.push_back(0);
		}

		push(move(data));
		return m_tasks.back();
	}

	const std::vector<uint8_t>& FrameStreamTestDriver::pushRandom()
	{
		std::mt19937 rng{ std::random_device{}() };
		return pushRandom(rng);
	}

	const std::vector<uint8_t>& FrameStreamTestDriver::pushRandom(std::mt19937& rng, size_t byteLimit)
	{
		if(byteLimit == 0)
			byteLimit = m_stream.data.size() * 3 / 8;

		size_t alignment = m_stream.symbolWidth.bytes();
		size_t numBytes = rng() % byteLimit + 1;
		numBytes = (numBytes + alignment - 1) / alignment * alignment;

		std::vector<uint8_t> frame(numBytes);
		for (uint8_t& byte : frame)
			byte = (uint8_t)rng();

		push(move(frame));
		return m_tasks.back();
	}

	SimProcess FrameStreamTestDriver::operator()()
	{
		while (true)
		{
			simu(m_stream.valid) = 0;
			simu(m_stream.startofpacket).invalidate();
			simu(m_stream.endofpacket).invalidate();
			simu(m_stream.data).invalidate();
			simu(m_stream.empty).invalidate();
			
			while (m_tasks.empty())
				co_await WaitClk(m_clock);

			const std::vector<uint8_t>& rawData = m_tasks.front();
			sim::DefaultBitVectorState data = sim::createDefaultBitVectorState(rawData.size(), rawData.data());
			m_tasks.pop();

			size_t bpb = m_stream.data.size();
			size_t numBeats = (data.size() + bpb - 1) / bpb;
			for (size_t i = 0; i < numBeats; ++i)
			{
				size_t bitsLeft = std::min(bpb + bpb * i, data.size()) - bpb * i;

				simu(m_stream.valid) = 1;
				simu(m_stream.startofpacket) = i == 0 ? 1 : 0;
				simu(m_stream.endofpacket) = i == numBeats - 1 ? 1 : 0;

				sim::DefaultBitVectorState dataState = data.extract(bpb * i, bitsLeft);
				dataState.resize(m_stream.data.size());

				simu(m_stream.data) = dataState;
				simu(m_stream.empty) = bpb / m_stream.symbolWidth.bits() - bitsLeft / m_stream.symbolWidth.bits();

				while(simu(m_ready) == 0)
					co_await WaitClk(m_clock);

				co_await WaitClk(m_clock);
			}
		}

	}

	FrameStreamTestVerifier::FrameStreamTestVerifier(const Clock& clock, const FrameStream& stream, const Bit& ready) :
		m_clock(clock),
		m_stream(stream),
		m_ready(ready)
	{
	}

	void FrameStreamTestVerifier::pushExpectedFrame(std::vector<uint8_t> data)
	{
		m_expectedFrames.emplace(std::move(data));
	}

	SimProcess FrameStreamTestVerifier::operator()()
	{
		while (true)
		{
			while(simu(m_stream.valid) == 0 || simu(m_ready) == 0)
				co_await WaitClk(m_clock);

			if (simu(m_stream.startofpacket) != 1)
				std::cerr << "failed: simu(m_stream.startofpacket) == 1\n";

			sim::DefaultBitVectorState data;
			while(true)
			{
				while (simu(m_stream.valid) == 0 || simu(m_ready) == 0)
					co_await WaitClk(m_clock);

				sim::DefaultBitVectorState beat = simu(m_stream.data);
				uint64_t empty = simu(m_stream.empty);
				beat.resize(m_stream.data.size() - empty * m_stream.symbolWidth.bits());

				size_t offset = data.size();
				data.resize(offset + beat.size());
				data.insert(beat, offset, beat.size());
				if (simu(m_stream.endofpacket) != 0)
					break;

				co_await WaitClk(m_clock);
			}
			co_await WaitClk(m_clock);

			HCL_ASSERT(!m_expectedFrames.empty());

			const std::vector<uint8_t>& expectedVec = m_expectedFrames.front();
			sim::DefaultBitVectorState expected = sim::createDefaultBitVectorState(expectedVec.size(), expectedVec.data());
			if (expected != data)
				std::cerr << "failed: expected == data\n";

			m_expectedFrames.pop();
		}
	}
}
