/*
	Synogate Anti Replay Core, an IP core implementing an algorithm
	designed to detect packet duplication based on packet sequence numbers.
	Copyright (C) 2021 Synogate UG (haftungsbeschraenkt), www.synogate.com 
	
	This program is commercial software: you must not copy, modify or
	distribute it without explicit written license by Synogate to do so. 
	
	This program is distributed WITHOUT ANY WARRANTY; without even the
	implied warranty of MERCHANTABILITY.
*/

#pragma once
#include <synogate/reporting.h>
#include <gatery/utils/ConfigTree.h>

#include <random>
#include <span>

namespace arc
{
	struct FrameStream
	{
		FrameStream() = default;
		FrameStream(BitWidth width) { setup(width); }

		Bit valid;
		BVec channel;
		BVec data;
		BVec error;

		// Packet Transfer Signals
		BVec empty;
		Bit startofpacket;
		Bit endofpacket;

		BVec userdata;

		// AVST Properties
		BitWidth					symbolWidth = 8_b;
		bool						emptyWithinPacket = false;
		std::vector<std::string>	errorDescriptor;
		bool						firstSymbolInHighOrderBits = true;
		size_t						maxChannel = 0;
		size_t						readyLatency = 0;
		size_t						readyAllowance = 0;

		bool has_channel = false;
		bool has_error = false;
		bool has_userdata = false;

		Bit in_frame() const;

		void setup(BitWidth dataWidth);
		void setup(gtry::utils::ConfigTree config);

		void pinIn(std::string_view prefix);
		void pinOut(std::string_view prefix);
	};

	class FrameStreamTestDriver
	{
	public:
		FrameStreamTestDriver(const Clock& clock, const FrameStream& stream, const Bit& ready);
		
		void push(std::vector<uint8_t> data);
		void pushRandom(size_t numFrames, uint32_t seed);

		const std::vector<uint8_t>& pushBeats(std::span<const uint64_t> beatData);
		const std::vector<uint8_t>& pushRandom();
		const std::vector<uint8_t>& pushRandom(std::mt19937& rng, size_t byteLimit = 0);
		
		bool busy() const { return !m_tasks.empty(); }
		SimProcess operator () ();

	private:
		Clock m_clock;
		FrameStream m_stream;
		Bit m_ready;
		std::queue<std::vector<uint8_t>> m_tasks;
	};

	class FrameStreamTestVerifier
	{
	public:
		FrameStreamTestVerifier(const Clock& clock, const FrameStream& stream, const Bit& ready);

		void pushExpectedFrame(std::vector<uint8_t> data);

		bool busy() const { return !m_expectedFrames.empty(); }
		SimProcess operator () ();

	private:
		Clock m_clock;
		FrameStream m_stream;
		Bit m_ready;
		std::queue<std::vector<uint8_t>> m_expectedFrames;
	};
}

BOOST_HANA_ADAPT_STRUCT(arc::FrameStream,
	valid, channel, data, error, empty, startofpacket, endofpacket, userdata,
	symbolWidth, emptyWithinPacket, errorDescriptor, firstSymbolInHighOrderBits, maxChannel, readyLatency, readyAllowance,
	has_channel, has_error, has_userdata);
