/*
	Synogate Anti Replay Core, an IP core implementing an algorithm
	designed to detect packet duplication based on packet sequence numbers.
	Copyright (C) 2021 Synogate UG (haftungsbeschraenkt), www.synogate.com 
	
	This program is commercial software: you must not copy, modify or
	distribute it without explicit written license by Synogate to do so. 
	
	This program is distributed WITHOUT ANY WARRANTY; without even the
	implied warranty of MERCHANTABILITY.
*/

#include "pch.h"
#include "Generator.h"

#include <synogate/export/vhdl/SynogateCodeFormatting.h>

#include <gatery/export/vhdl/VHDLExport.h>

#include <gatery/scl/synthesisTools/GHDL.h>
#include <gatery/scl/synthesisTools/IntelQuartus.h>
#include <gatery/scl/synthesisTools/XilinxVivado.h>

#include <gatery/scl/arch/FromConfig.h>
#include <gatery/scl/arch/general/FPGADevice.h>

#ifdef USE_TESTS
# define BOOST_TEST_MODULE "Unit tests for ..."
# define BOOST_TEST_NO_MAIN
# define BOOST_TEST_ALTERNATIVE_INIT_API
# include <boost/test/unit_test.hpp>
#endif

#include <filesystem>
#include <boost/program_options.hpp>
#include <boost/algorithm/string.hpp>

using namespace boost::program_options;

void FuzzingTest::load(gtry::utils::ConfigTree tree)
{
    numBatches = tree["num_batches"].as(numBatches);
    numCyclesPerBatch = tree["num_cycles_per_batch"].as(numCyclesPerBatch);
    reportProgress = tree["report_progress"].as(reportProgress);
}

void BasicSettings::load(gtry::utils::ConfigTree &tree)
{
    {
        auto vhdl = tree["vhdl"];
        writeVHDL = !vhdl["disable"].as(!writeVHDL);
        vhdlPath = vhdl["path"].as(vhdlPath.string());

        std::filesystem::path vhdlDir = vhdlPath;
        if (vhdlDir.has_extension())
            vhdlDir = vhdlDir.parent_path();

        testbenchPath = vhdl["testbench_path"].as(vhdlDir.string());
        library = vhdl["library"].as(library);
        instantiationTemplateVHDL = vhdl["instantiation_template_vhdl"].as((vhdlDir / (library + "_inst.vhd")).string());

        if (boost::iequals(instantiationTemplateVHDL->string(), "No"))
            instantiationTemplateVHDL.reset();
    }
    synthesisTool = tree["synthesis_tool"].as(synthesisTool);

    {
        auto doc = tree["doc"];
        writeDoc = !doc["disable"].as(!writeDoc);
        docPath = doc["path"].as(docPath.string());
    }

    {
        auto waveform = tree["waveforms"];
        writeWaveforms = !waveform["disable"].as(!writeWaveforms);
        waveformPath = waveform["path"].as(waveformPath.string());
    }

    fuzzingTest.load(tree["fuzzing_test"]);

    auto clockConfig = tree["clocks"];
    for (auto it = clockConfig.mapBegin(); it != clockConfig.mapEnd(); ++it)
    {
        gtry::ClockConfig& cc = clocks.emplace_back();
        cc.setName(it.key());
        cc.loadConfig(*it);

        gtry::trace() << "found " << cc;
    }
}


Generator::Generator(gtry::utils::ConfigTree &configTree, std::string coreName, std::string defaultLibTopName) : m_configTree(configTree), m_coreName(std::move(coreName))
{
    m_basicSettings.library = std::move(defaultLibTopName);
    m_basicSettings.load(configTree);

    setupTargetTechnology(configTree["target_technology"]);

    gtry::hlim::NodeGroup::configTree(configTree["instance"]);
    setupClock();
}

void Generator::build()
{
    gtry::reporting::InterfaceDocumentation doc(m_basicSettings.docPath);
    prepDocu(doc);

    buildCore();
    
    if (m_basicSettings.writeVHDL) {
        m_vhdlExport.emplace(m_basicSettings.vhdlPath, m_basicSettings.testbenchPath);
        m_vhdlExport->setLibrary(m_basicSettings.library);
        prepExport();
        (*m_vhdlExport)(m_design.getCircuit());
    }

    runSimulations();
    finishDocu(doc);
    if (m_basicSettings.writeDoc)
        doc.generate();
}

void Generator::setupTargetTechnology(const gtry::utils::ConfigTree &configTree)
{
    if (configTree)
        m_design.setTargetTechnology(gtry::scl::arch::createTargetTechnologyFromConfig(configTree));
}

void Generator::prepDocu(gtry::reporting::InterfaceDocumentation &doc)
{
    auto &configSec = doc.generateDefaultMainPage(m_coreName);
    m_docuHooks.sections["config"] = &configSec;


}


void Generator::buildCore()
{
    gtry::GroupScope& ent = gtry::DesignScope::get()->getRootGroup();
    ent.setName(m_basicSettings.library);
}

void Generator::prepExport()
{
    if (m_basicSettings.synthesisTool == "none") {
        // default
        m_vhdlExport->writeProjectFile("files.txt");
    } else if (m_basicSettings.synthesisTool == "ghdl") {
        m_vhdlExport->targetSynthesisTool(new gtry::GHDL());
        m_vhdlExport->writeStandAloneProjectFile("ghdl.sh");
        m_vhdlExport->writeClocksFile("clocks.xdc");
    } else if (m_basicSettings.synthesisTool == "intel_quartus") {
        m_vhdlExport->targetSynthesisTool(new gtry::IntelQuartus());
        m_vhdlExport->writeProjectFile(m_basicSettings.library + ".tcl");
        m_vhdlExport->writeStandAloneProjectFile(m_basicSettings.library + ".qsf");
        m_vhdlExport->writeClocksFile("clocks.sdc");
        m_vhdlExport->writeConstraintsFile("constraints.sdc");
    } else if (m_basicSettings.synthesisTool == "xilinx_vivado") {
        m_vhdlExport->targetSynthesisTool(new gtry::XilinxVivado());    
        m_vhdlExport->writeProjectFile("vivado.tcl");
        m_vhdlExport->writeClocksFile("clocks.xdc");
        m_vhdlExport->writeConstraintsFile("constraints.xdc");
    } else 
        throw std::runtime_error((std::string("Invalid synthesis tool: ") + m_basicSettings.synthesisTool).c_str());

    m_vhdlExport->setFormatting(new synogate::SynogateCodeFormatting(m_coreName));

    if(m_basicSettings.instantiationTemplateVHDL)
        m_vhdlExport->writeInstantiationTemplateVHDL(*m_basicSettings.instantiationTemplateVHDL);
}

void Generator::runSimulations()
{
}


void Generator::finishDocu(gtry::reporting::InterfaceDocumentation &doc)
{
    gtry::reporting::Section& section = *m_docuHooks.sections["config"];

    section << "This section lists all configuration options that may have an influence on the generated core.";

    {
        auto &table = section.addTable("Configuration Options Summary", "tbl:configOpts");
        table.addHeader(2);
        table(-1, 0) << "Parameter Name";
        table(-1, 1) << "Parameter Value";
        table << "Selection of the most important configuration options and inferred parameters at a glance.";

        m_docuHooks.tables["tbl:configOpts"] = &table;
    }

    {
        auto& tab = section.addTable("Synthesis Tool Parameters", "tbl:params_synthTool");
        tab.addHeader(2);
        tab(-1, 0) << "Parameter Name";
        tab(-1, 1) << "Parameter Value";

        tab.addRow();
        tab(-1, 0) << "Synthesis tool";
        tab(-1, 1) << m_basicSettings.synthesisTool;
                
        tab << "Chosen synthesis tool configuration";
    }    

    if (auto *fpgaDevice = m_design.getTargetTechnology<gtry::scl::arch::FPGADevice>()) {

        auto& tab = section.addTable("Target Technology Parameters", "tbl:params_target_tech");
        tab.addHeader(2);
        tab(-1, 0) << "Parameter Name";
        tab(-1, 1) << "Parameter Value";

        tab.addRow();
        tab(-1, 0) << "Vendor";
        tab(-1, 1) << fpgaDevice->getVendor();

        tab.addRow();
        tab(-1, 0) << "Family";
        tab(-1, 1) << fpgaDevice->getFamily();

        tab.addRow();
        tab(-1, 0) << "Device";
        tab(-1, 1) << fpgaDevice->getDevice();

        for (const auto &mem : fpgaDevice->getEmbeddedMemories().getList()) {
            tab.addRow();
            tab(-1, 0) << "Has " << mem->getDesc().memoryName;
            tab(-1, 1) << "yes";
        }

        tab << "Chosen target technology configuration";
    }
    // todo: do for each root clock

    {
        auto& tab = section.addTable("Clock parameters", "tbl:params_clk");
        tab.addHeader(2);
        tab(-1, 0) << "Parameter Name";
        tab(-1, 1) << "Parameter Value";

#if 0
        auto print = [&](std::string_view label, const auto& value)
        {
            if (value)
            {
                tab.addRow();
                tab(-1, 0) << label;
                if constexpr (std::is_enum_v<std::remove_reference_t<decltype(*value)>>)
                    tab(-1, 1) << magic_enum::enum_name(*value);
                else
                    tab(-1, 1) << *value;
            }
        };

        auto &clk = m_basicSettings.clocks.front();
        print("clock name", clk.name());

        if (clk.absoluteFrequency())
        {
            double period_mhz = double(clk.absoluteFrequency()->numerator()) / clk.absoluteFrequency()->denominator();
            period_mhz /= 1'000'000;
            std::string text = std::to_string(size_t(std::round(period_mhz))) + " MHz";
            print("period", std::make_optional(text));
        }

        print("clock edge", clk.triggerEvent());
        print("reset name", clk.resetName());
        print("reset type", clk.resetType());
        print("memory reset type", clk.memoryResetType());
        print("reset active", clk.resetActive());
        print("initialize registers", std::make_optional(clk.initializeRegs()?"yes":"no"));
        print("initialize memory", std::make_optional(clk.initializeMemory()?"yes":"no"));

        auto it = clk.attributes().userDefinedVendorAttributes.find("all");
        if(it != clk.attributes().userDefinedVendorAttributes.end())
            for (const auto& attr : it->second)
                print(attr.first, std::make_optional(attr.second.value));
#else

        auto print = [&](std::string_view label, const auto& value)
        {
            tab.addRow();
            tab(-1, 0) << label;
            if constexpr (std::is_enum_v<std::remove_reference_t<decltype(value)>>)
                tab(-1, 1) << magic_enum::enum_name(value);
            else
                tab(-1, 1) << value;
        };

        auto &clk = *m_design.getCircuit().getClocks()[0];
        
        print("clock name", clk.getName());
        print("clock reset", clk.getResetName());

        {
            size_t period_mhz = clk.getAbsoluteFrequency().numerator() / clk.getAbsoluteFrequency().denominator() / 1'000'000;
            std::string text = std::to_string(period_mhz) + " MHz";
            print("frequency", text);
        }

        print("trigger event", clk.getTriggerEvent());
        print("reset type", clk.getRegAttribs().resetType);
        print("memory reset type", clk.getRegAttribs().memoryResetType);
        print("reset active", clk.getRegAttribs().resetActive);
        print("initialize registers", clk.getRegAttribs().initializeRegs?"yes":"no");
        print("initialize memory", clk.getRegAttribs().initializeMemory?"yes":"no");

#endif
                
        tab << "Chosen clock configuration";        
    }    

    std::function<void(gtry::hlim::NodeGroup&)> docGroup = [&](gtry::hlim::NodeGroup& scope) {

        gtry::utils::PropertyTree prop = scope.properties();
        if (prop.size() > 1) // ignore always present name property
        {
            section << "<h4>" << scope.instancePath() << "</h4>";

            auto& tab = section.addTable(scope.getName() + " instance parameter", "tbl:" + scope.instancePath());
            tab.addHeader(2);
            tab(-1, 0) << "Parameter Name";
            tab(-1, 1) << "Parameter Value";

            for (auto it = prop.begin(); it != prop.end(); ++it)
            {
                std::string key = it->first.as<std::string>();
                if (key != "name")
                {
                    tab.addRow();
                    tab(-1, 0) << key;
                    tab(-1, 1) << "<pre>" << it->second << "</pre>";
                }
            }
            tab << scope.instancePath() + " instance parameter";
        }
        
        for (auto&& child : scope.getChildren())
            docGroup(*child);
    };
//    docGroup(*gtry::DesignScope::get()->getRootGroup().nodeGroup());

    gtry::utils::PropertyTree properties = gtry::DesignScope::get()->instanceProperties(true);
    section << "The following lists the configuration options of the core itself in yaml form. Missing values are filled in with the used defaults.";
    section << "<div style=\"overflow:auto;border:1px solid gray;background: rgb(240, 240, 240);margin:10px;padding:10px;max-height: 250px;\"><pre><code class=\"yaml\">" << properties << "</code></pre></div>";
}

void Generator::setupClock()
{
    if (m_basicSettings.clocks.empty())
    {
        gtry::trace() << "no clock found, use default clock settings.";
        gtry::ClockConfig& cfg = m_basicSettings.clocks.emplace_back();
        cfg.setName("clock");
    }

    gtry::ClockConfig& cc = m_basicSettings.clocks.front();
    if (!cc.absoluteFrequency())
    {
        gtry::trace() << "no clock period specified, use 100 MHz for simulation and export.";
        cc.setAbsoluteFrequency(gtry::ClockConfig::ClockRational{ 100'000'000, 1 });
    }
}

static void setEnvVar(const char* name, const std::string& value)
{
#ifdef WIN32
    if (_putenv_s(name, value.c_str()))
        throw std::runtime_error("_putenv_s failed");
#else
    if (setenv(name, value.c_str(), true))
        throw std::runtime_error("setenv failed");
#endif
}

static void setupEnvironmentVars(int argc, char* argv[], std::string configFile)
{
    using std::filesystem::path;

    if (argc > 0)
    {
        path binPath = weakly_canonical(path{ argv[0] }).parent_path();
        setEnvVar("BinaryPath", binPath.string());
    }

    {
        path configPath = weakly_canonical(path{ configFile }).parent_path();
        setEnvVar("ConfigPath", configPath.string());
    }

    {
        path workPath = std::filesystem::current_path();
        setEnvVar("CurrentPath", workPath.string());
    }
}

int generatorMain(int argc, char* argv[], std::function<int(gtry::utils::ConfigTree&)> generate)
{
    using namespace boost::program_options;

#ifdef USE_TESTS
    bool run_boost_test = false;
    for (int i = 1; i < argc; ++i)
    {
        run_boost_test |= !strncmp(argv[i], "--run_test", 10);
        run_boost_test |= !strncmp(argv[i], "-t", 2);
        run_boost_test |= !strncmp(argv[i], "--list_content", 14);
    }
    if (run_boost_test)
    {
        return boost::unit_test::unit_test_main(init_unit_test, argc, argv);
    }
#endif

    options_description desc("Anti-Replay Core generator");
    desc.add_options()
        ("help", "help text")
        ("config", value<std::string>()->required(), "Read configuration from file.")
        ("trace", value<std::string>(), "Report to file. Use - for stderr.")
        ;

    positional_options_description pdesc;
    pdesc.add("config", 1);

    variables_map vm;
    try {
        store(command_line_parser(argc, argv).options(desc).positional(pdesc).run(), vm);
        notify(vm);

        if (vm.count("help")) {
            std::cout << desc << '\n';
            return 0;
        }

        gtry::utils::ConfigTree configTree;
        if (vm.count("config"))
        {
            std::string configFile = vm["config"].as<std::string>();
            setupEnvironmentVars(argc, argv, configFile);
            configTree.loadFromFile(configFile);
        }

        std::ofstream traceFile;
        if (vm.count("trace"))
        {
            std::string traceFileName = vm["trace"].as<std::string>();
            if (traceFileName == "-")
                gtry::trace::outputStream(&std::cerr);
            else
            {
                traceFile.open(traceFileName, std::ofstream::binary);
                if (!traceFile)
                    std::cerr << "failed to open trace file " << traceFileName << '\n';
                else
                    gtry::trace::outputStream(&traceFile);
            }
        }

        return generate(configTree);
    }
    catch (const boost::program_options::error& e)
    {
        std::cerr << "command line error: " << e.what() << "\n\n";
        std::cerr << desc << '\n';
        return 1;
    }
    catch (const std::exception& e)
    {
        std::cerr << "error: " << e.what() << '\n';
        return 1;
    }
    return 0;
}
