/*
	Synogate Anti Replay Core, an IP core implementing an algorithm
	designed to detect packet duplication based on packet sequence numbers.
	Copyright (C) 2021 Synogate UG (haftungsbeschraenkt), www.synogate.com 
	
	This program is commercial software: you must not copy, modify or
	distribute it without explicit written license by Synogate to do so. 
	
	This program is distributed WITHOUT ANY WARRANTY; without even the
	implied warranty of MERCHANTABILITY.
*/

#pragma once

#include <synogate/reporting.h>
#include <gatery/frontend/DesignScope.h>

#include <gatery/simulation/waveformFormats/MemoryTrace.h>
#include <gatery/export/vhdl/VHDLExport.h>
#include <gatery/utils/ConfigTree.h>


#include <filesystem>
#include <map>
#include <memory>
#include <boost/lexical_cast.hpp>

namespace gtry::vhdl {
    class VHDLExport;
}

struct FuzzingTest {
    size_t numBatches = 2;
    size_t numCyclesPerBatch = 128;
    bool reportProgress = false;

    void load(gtry::utils::ConfigTree tree);
};

struct BasicSettings {
    bool writeVHDL = true;
    std::filesystem::path vhdlPath = "vhdl/";
    std::filesystem::path testbenchPath = "vhdl/";

    bool writeDoc = true;
    std::filesystem::path docPath = "doc/";

    bool writeWaveforms = true;
    std::filesystem::path waveformPath = "waveforms/";

    FuzzingTest fuzzingTest;

    std::string library = "gatery";
    std::optional<std::filesystem::path> instantiationTemplateVHDL;

    std::string synthesisTool = "none";

    std::vector<gtry::ClockConfig> clocks;

    void load(gtry::utils::ConfigTree &tree);
};

struct DocumentationHooks {
    std::map<std::string, gtry::reporting::Section*> sections;
    std::map<std::string, gtry::reporting::Table*> tables;
    std::map<std::string, gtry::reporting::Figure*> figures;
    std::map<std::string, gtry::reporting::Waveform*> waveforms;
    std::map<std::string, std::unique_ptr<gtry::sim::MemoryTrace>> traces;
};

class Generator {
    public:
        Generator(gtry::utils::ConfigTree &configTree, std::string coreName, std::string defaultLibTopName);

        void build();

        inline BasicSettings &getBasicSettings() { return m_basicSettings; }
        inline DocumentationHooks &getDocuHooks() { return m_docuHooks; }

    protected:
        gtry::utils::ConfigTree &m_configTree;
        std::string m_coreName;

        gtry::DesignScope m_design;

        BasicSettings m_basicSettings;
        DocumentationHooks m_docuHooks;

        std::optional<gtry::vhdl::VHDLExport> m_vhdlExport;

        void setupTargetTechnology(const gtry::utils::ConfigTree &configTree);

        virtual void prepDocu(gtry::reporting::InterfaceDocumentation &doc);
        virtual void buildCore();
        virtual void prepExport();
        virtual void runSimulations();
        virtual void finishDocu(gtry::reporting::InterfaceDocumentation &doc);

        void setupClock();

};

int generatorMain(int argc, char* argv[], std::function<int(gtry::utils::ConfigTree&)> generate);

template<typename T>
int generatorMain(int argc, char* argv[])
{
    return generatorMain(argc, argv, [](gtry::utils::ConfigTree& config) {
        T generator(config);
        generator.build();
        return 0;
    });
}
