/*
	Synogate Anti Replay Core, an IP core implementing an algorithm
	designed to detect packet duplication based on packet sequence numbers.
	Copyright (C) 2021 Synogate UG (haftungsbeschraenkt), www.synogate.com 
	
	This program is commercial software: you must not copy, modify or
	distribute it without explicit written license by Synogate to do so. 
	
	This program is distributed WITHOUT ANY WARRANTY; without even the
	implied warranty of MERCHANTABILITY.
*/

#include "pch.h"
#include "AntiReplayCoreModel.h"

namespace arc
{
    AntiReplayCoreModel::AntiReplayCoreModel(uint64_t physicalSize, uint64_t virtualSize):
		m_physicalWindowSize(physicalSize),
		m_virtualWindowSize(virtualSize)
	{
	}

	AntiReplayCoreModel& AntiReplayCoreModel::windowSize(uint64_t physicalSize, uint64_t virtualSize)
	{
		m_physicalWindowSize = physicalSize;
		m_virtualWindowSize = virtualSize;
		return *this;
    }

	AntiReplayCoreModel::Result AntiReplayCoreModel::execute(const Cmd& cmd)
	{
		Result result = { .valid = false };
		switch (cmd.opcode) 
		{
		case 0: // OpUpdate
			result.code = update(m_state[cmd.connectionIndex], cmd.sequenceNumber);
			result.valid = true;
			break;
		case 1: // OpDeactivate
			m_state[cmd.connectionIndex].active = false;
			break;
		case 5: // OpActivate
			m_state[cmd.connectionIndex].active = true;
			break;
		case 2: // OpSetSeqNr
			m_state[cmd.connectionIndex].seen.clear();
			m_state[cmd.connectionIndex].seen.insert(cmd.sequenceNumber);
			break;
		case 3: // OpSetWindowSize
			m_virtualWindowSize = cmd.windowSize;
			break;
		default:
			assert(!"unknown opcode");
		}
		return result;
	}

	uint64_t AntiReplayCoreModel::update(SaState& sa, uint64_t sequenceNumber)
	{
		if (!sa.active)
		{
			return 3; // PassCheckDisabled
		}
		else if (sa.seen.empty())
		{
			assert(!"sa not initialized");
			return -1; // undefined
		}
		else if (uint64_t tow = *sa.seen.rbegin(); tow < sequenceNumber) 
		{
			sa.seen.insert(sequenceNumber);
			
			// Garbage collect unneeded
			size_t windowToKeep = std::max(m_virtualWindowSize, m_physicalWindowSize);
			if (sequenceNumber > windowToKeep) {
				size_t threshold = sequenceNumber - windowToKeep;
				while (!sa.seen.empty() && *sa.seen.begin() < threshold)
					sa.seen.erase(sa.seen.begin());
			}
			return 1; // PassAboveWindow
		}
		else if (uint64_t offset = tow - sequenceNumber; offset < std::min(m_physicalWindowSize, m_virtualWindowSize)) 
		{
			if (sa.seen.contains(sequenceNumber))
			{
				return 2; // FailReplay
			}
			else 
			{
				sa.seen.insert(sequenceNumber);
				return 7; // PassInPhysicalWindow
			}
		}
		else if (offset < m_virtualWindowSize)
		{
			return 5; // PassInVirtualWindow
		}
		else
		{
			return 0; // FailBelowWindow
		}
	}
}
