/*
	Synogate Anti Replay Core, an IP core implementing an algorithm
	designed to detect packet duplication based on packet sequence numbers.
	Copyright (C) 2021 Synogate UG (haftungsbeschraenkt), www.synogate.com 
	
	This program is commercial software: you must not copy, modify or
	distribute it without explicit written license by Synogate to do so. 
	
	This program is distributed WITHOUT ANY WARRANTY; without even the
	implied warranty of MERCHANTABILITY.
*/

#include "pch.h"
#include "AntiReplayFuzzer.h"

namespace arc
{
	AntiReplayVerifier::AntiReplayVerifier(Clock clock, CmdStage cmd, AntiReplayResult out, BitWidth windowSize, bool outputExpectedResult) :
		m_clock(clock),
		m_cmd(cmd),
		m_out(out),
		m_model(windowSize.bits(), windowSize.bits())
	{
		if (outputExpectedResult) {
			m_expectedCode = pinIn(3_b).setName("verify_code");
			m_noError = pinIn().setName("verify_ok");
		}

		m_returnCodeHistorgram = {};
	}

	AntiReplayVerifier::AntiReplayVerifier(Clock clock, CmdStage cmd, Valid<BVec> out, BitWidth windowSize, bool outputExpectedResult) :
		AntiReplayVerifier(clock, cmd, AntiReplayResult{ .valid = out.valid, .code = out.data }, windowSize, outputExpectedResult)
	{}

	AntiReplayVerifier& AntiReplayVerifier::windowSize(uint64_t physicalSize, uint64_t virtualSize)
	{
		m_model.windowSize(physicalSize, virtualSize);
		return *this;
	}

	SimProcess AntiReplayVerifier::operator () ()
	{
		while (true)
		{
			if (simu(m_cmd.valid) && (!m_cmdReady || simu(*m_cmdReady)))
			{
				Cmd& cmd = m_tasks.emplace();
				cmd.opcode = simu(m_cmd.opcode);
				cmd.connectionIndex = simu(m_cmd.connectionIndex);
				cmd.sequenceNumber = simu(m_cmd.sequenceNumber);
				cmd.windowSize = simu(m_cmd.windowSize);
			}

			if (simu(m_out.valid) && (!m_outReady || simu(*m_outReady)))
			{
				bool processed = false;

				while (!m_tasks.empty() && !processed)
				{
					const Cmd cmd = m_tasks.front();
					m_tasks.pop();
					processed = execute(cmd);
				}
				assert(processed);
			}
			co_await WaitClk(m_clock);
		}
	}

	bool AntiReplayVerifier::execute(const Cmd& cmd)
	{
		auto result = m_model.execute(cmd);
		if (result.valid) {
			expectedResult(result.code);
			return true;
		}
		return false;
	}

	void AntiReplayVerifier::expectedResult(uint64_t value)
	{
		m_returnCodeHistorgram[value]++;

		uint64_t outCode = simu(m_out.code);

		if(simu(m_out.code).defined() != 7)
			std::cerr << "failed: simu(m_out.code).defined() == 7\n";

		if (value != outCode)
			std::cerr << "failed: value == outCode\n";
		//assert(value == outCode);

		if (m_expectedCode)
			simu(*m_expectedCode) = value;
		if (m_noError)
			simu(*m_noError) = value == outCode ? 1 : 0;
	}


	void AntiReplayVerifier::checkAllReturnCodesCovered(size_t minOccurances)
	{
		std::array<size_t, 5> validOpcodes = {0b000, 0b101, 0b001, 0b010, 0b011};
		for (auto code : validOpcodes)
			if (m_returnCodeHistorgram[code] < minOccurances) {
				std::cerr << "The opcode " << code << " only occured " << m_returnCodeHistorgram[code] << " times but we would like to see it at least " << minOccurances << " to be confident about the test." << std::endl;
				assert(m_returnCodeHistorgram[code] >= minOccurances);
			}
	}	

	std::tuple<uint64_t, uint64_t> AntiReplayFuzzer::next()
	{
		uint64_t sa = m_rng() & m_saMask;
		int32_t dist = int32_t(m_rng()) >> (32 - m_distanceBits);

		uint64_t tow = m_topOfWindow[sa];
		uint64_t sai = tow + dist;
		if (dist >= 0 && sai < tow)
			sai = std::numeric_limits<uint64_t>::max();
		else if (dist < 0 && sai > tow)
			sai = 0;

		if (sai > tow)
			m_topOfWindow[sa] = sai;

		return { sa, sai };
	}

	void arc::AntiReplayFuzzer::activate(uint64_t connectionIndex, bool activate)
	{
		m_tasks.emplace(Cmd{
			.opcode = activate ? 5u : 1u,
			.connectionIndex = connectionIndex
		});
	}

	void arc::AntiReplayFuzzer::setseq(uint64_t connectionIndex, uint64_t sequenceNumber, bool updateTow)
	{
		m_tasks.emplace(Cmd{
			.opcode = 2,
			.connectionIndex = connectionIndex,
			.sequenceNumber = sequenceNumber
		});

		if (updateTow)
		{
			m_topOfWindow[connectionIndex] = sequenceNumber;
		}
	}

	void arc::AntiReplayFuzzer::update(uint64_t connectionIndex, uint64_t sequenceNumber)
	{
		m_tasks.emplace(Cmd{
			.opcode = 0,
			.connectionIndex = connectionIndex,
			.sequenceNumber = sequenceNumber
		});
	}

	void arc::AntiReplayFuzzer::virtualWindowSize(uint64_t size)
	{
		m_tasks.emplace(Cmd{
			.opcode = 3,
			.windowSize = size
		});
	}

	void arc::AntiReplayFuzzer::updateNext()
	{
		auto [connectionIndex, sequenceNumber] = next();
		update(connectionIndex, sequenceNumber);
	}

	SimProcess arc::AntiReplayFuzzer::operator()()
	{
		HCL_ASSERT(m_clock);

		while (true)
		{
			simu(m_cmd.valid) = 0;
			simu(m_cmd.opcode).invalidate();
			simu(m_cmd.connectionIndex).invalidate();
			simu(m_cmd.sequenceNumber).invalidate();
			simu(m_cmd.windowSize).invalidate();

			while (m_tasks.empty())
				co_await WaitClk(*m_clock);

			const Cmd& cmd = m_tasks.front();
			simu(m_cmd.valid) = 1;
			simu(m_cmd.opcode) = cmd.opcode;

			if (cmd.opcode != 3)
			{
				simu(m_cmd.connectionIndex) = cmd.connectionIndex;

				if(cmd.opcode != 1 && cmd.opcode != 5)
					simu(m_cmd.sequenceNumber) = cmd.sequenceNumber;
			}
			else
			{
				simu(m_cmd.windowSize) = cmd.windowSize;
			}
			m_tasks.pop();

			while(m_cmdReady && simu(*m_cmdReady) == 0)
				co_await WaitClk(*m_clock);

			co_await WaitClk(*m_clock);
		}

	}


}


