/*
	Synogate Anti Replay Core, an IP core implementing an algorithm
	designed to detect packet duplication based on packet sequence numbers.
	Copyright (C) 2021 Synogate UG (haftungsbeschraenkt), www.synogate.com 
	
	This program is commercial software: you must not copy, modify or
	distribute it without explicit written license by Synogate to do so. 
	
	This program is distributed WITHOUT ANY WARRANTY; without even the
	implied warranty of MERCHANTABILITY.
*/

#pragma once

#include <gatery/frontend.h>

#include "../AntiReplay.h"
#include "../FrameAdapter.h"

#include "AntiReplayCoreModel.h"

#include <random>
#include <stdint.h>
#include <map>
#include <set>

namespace arc
{
	class AntiReplayFuzzer
	{
		struct Cmd
		{
			uint64_t opcode = 0;
			uint64_t connectionIndex = 0;
			uint64_t sequenceNumber = 0;
			uint64_t windowSize = 0;
		};
	public:
		AntiReplayFuzzer(uint32_t seed) : m_rng(seed) {}

		AntiReplayFuzzer& saMask(uint64_t mask) { m_saMask = mask; return *this; }
		AntiReplayFuzzer& distanceBits(uint64_t bits) { m_distanceBits = bits; return *this; }
		AntiReplayFuzzer& cmd(const Clock& clock, const CmdStage& cmd) { m_clock = clock; m_cmd = cmd; return *this; }
		AntiReplayFuzzer& cmdReady(const Bit& ready) { m_cmdReady = ready; return *this; }

		std::tuple<uint64_t, uint64_t> next();

		void activate(uint64_t connectionIndex, bool activate = true);
		void setseq(uint64_t connectionIndex, uint64_t sequenceNumber, bool updateTow);
		void update(uint64_t connectionIndex, uint64_t sequenceNumber);
		void virtualWindowSize(uint64_t size);
		void updateNext();
		bool busy() const { return !m_tasks.empty(); }

		SimProcess operator () ();

	private:
		std::mt19937 m_rng;
		uint64_t m_saMask = 0;
		uint64_t m_distanceBits = 1;

		std::map<uint64_t, uint64_t> m_topOfWindow;

		std::queue<Cmd> m_tasks;

		std::optional<Clock> m_clock;
		CmdStage m_cmd;
		std::optional<Bit> m_cmdReady;
	};

	class AntiReplayVerifier
	{
		using Cmd = AntiReplayCoreModel::Cmd;

	public:
		AntiReplayVerifier(Clock clock, CmdStage cmd, AntiReplayResult out, BitWidth windowSize, bool outputExpectedResult = true);
		AntiReplayVerifier(Clock clock, CmdStage cmd, Valid<BVec> out, BitWidth windowSize, bool outputExpectedResult = true);

		AntiReplayVerifier& cmdReady(const Bit& ready) { m_cmdReady = ready; return *this; }
		AntiReplayVerifier& outReady(const Bit& ready) { m_outReady = ready; return *this; }
		AntiReplayVerifier& windowSize(uint64_t physicalSize, uint64_t virtualSize);

		SimProcess operator () ();

		void checkAllReturnCodesCovered(size_t minOccurances);

	protected:
		bool execute(const Cmd& cmd);

		void expectedResult(uint64_t value);

	private:
		Clock m_clock;
		CmdStage m_cmd;
		AntiReplayResult m_out;
		std::optional<Bit> m_cmdReady;
		std::optional<Bit> m_outReady;

		std::optional<BVec> m_expectedCode;
		std::optional<Bit> m_noError;

		std::queue<Cmd> m_tasks;
		AntiReplayCoreModel m_model;

		std::array<size_t, 8> m_returnCodeHistorgram;
	};
}
