/*
	Synogate Anti Replay Core, an IP core implementing an algorithm
	designed to detect packet duplication based on packet sequence numbers.
	Copyright (C) 2021 Synogate UG (haftungsbeschraenkt), www.synogate.com 
	
	This program is commercial software: you must not copy, modify or
	distribute it without explicit written license by Synogate to do so. 
	
	This program is distributed WITHOUT ANY WARRANTY; without even the
	implied warranty of MERCHANTABILITY.
*/

#include "pch.h"
#ifdef USE_TESTS
#include <boost/test/unit_test.hpp>
#include <boost/test/data/dataset.hpp>
#include <boost/test/data/test_case.hpp>
#include <boost/test/data/monomorphic.hpp>

#include <gatery/simulation/waveformFormats/VCDSink.h>
#include <gatery/export/vhdl/VHDLExport.h>
#include <gatery/export/DotExport.h>

#include <gatery/utils/Range.h>
#include <gatery/hlim/Subnet.h>
#include <gatery/hlim/SignalDelay.h>
#include <gatery/hlim/RegisterRetiming.h>

#include <gatery/export/DotExport.h>
#include <gatery/hlim/Subnet.h>
#include <gatery/scl/synthesisTools/IntelQuartus.h>
#include <gatery/scl/synthesisTools/XilinxVivado.h>


#include <gatery/hlim/supportNodes/Node_Attributes.h>

#include "../AntiReplay.h"

#include "AntiReplayFuzzer.h"

#include "../FrameAdapter.h"

using namespace boost::unit_test;
using namespace gtry;

class AntiReplayTest : public ClockedTest, protected arc::AntiReplay
{
public:
	AntiReplayTest()
	{
		cmd.connectionIndex = 1_b;
		cmd.windowSize = 16_b;
		cmd.sequenceNumber = 32_b;
		cmd.pinIn("cmd_");

		init(cmd);
	}

protected:
	arc::CmdStage cmd;
};

BOOST_FIXTURE_TEST_CASE(AntiReplay_comb_retiming_test, BoostUnitTestSimulationFixture)
{
	Clock clock(ClockConfig{}.setAbsoluteFrequency(100'000'000).setName("clock"));
	ClockScope clkScp(clock);

	arc::CmdStage cmd;
	cmd.connectionIndex = 4_b;
	cmd.windowSize = 12_b;
	cmd.sequenceNumber = 64_b;

	arc::AntiReplayResult out;
	cmd.pinIn("cmd_");

	arc::AntiReplay ar{ cmd };
	ar.windowSize(6_b, 1_b);
	ar.generateSingleStage();
	ar.result(out);

	pinOut(out.valid).setName("out_valid");
	pinOut(out.code).setName("out_code");
/*
    {
		auto subnet = hlim::Subnet::all(design.getCircuit());

		design.getCircuit().optimizeSubnet(subnet);
		
	    hlim::retimeForward(design.getCircuit(), subnet);
		design.getCircuit().optimizeSubnet(subnet);
		

        hlim::SignalDelay delays;
        delays.compute(hlim::Subnet::all(design.getCircuit()));

        {
            DotExport exp("final_signalDelays.dot");
            exp(design.getCircuit(), hlim::ConstSubnet::all(design.getCircuit()), delays);
            exp.runGraphViz("final_signalDelays.svg");
        
		}

		{
            hlim::NodePort criticalOutput;
            unsigned criticalBit = ~0u;
            float criticalTime = 0.0f;
            for (auto &n : design.getCircuit().getNodes())
                for (auto i : gtry::utils::Range(n->getNumOutputPorts())) {
                    hlim::NodePort np = {.node = n.get(), .port = i};
                    auto d = delays.getDelay(np);
                    for (auto i : gtry::utils::Range(d.size())) {   
                        if (d[i] > criticalTime) {
                            criticalTime = d[i];
                            criticalOutput = np;
                            criticalBit = i;
                        }
                    }
                }

			hlim::ConstSubnet criticalPathSubnet;
			{
				hlim::NodePort np = criticalOutput;
				unsigned bit = criticalBit;
				while (np.node != nullptr) {
					criticalPathSubnet.add(np.node);
					unsigned criticalInputPort, criticalInputBit;
					np.node->estimateSignalDelayCriticalInput(delays, np.port, bit, criticalInputPort, criticalInputBit);
					if (criticalInputPort == ~0u)
						np.node = nullptr;
					else {
						np = np.node->getDriver(criticalInputPort);
						bit = criticalInputBit;
					}
				}
			}

			criticalPathSubnet.dilate(true, true);


            DotExport exp("final_criticalPath.dot");
            exp(design.getCircuit(), criticalPathSubnet, delays);
            exp.runGraphViz("final_criticalPath.svg");
		}
	}
*/
	addSimulationProcess([&]()->SimProcess {

		simu(cmd.valid) = 0;
		for (size_t i = 0; i < 4; ++i)
			co_await WaitClk(clock);

		simu(cmd.valid) = 1;
		simu(cmd.opcode) = 2;
		simu(cmd.connectionIndex) = 3;
		simu(cmd.sequenceNumber) = 64;
		simu(cmd.windowSize) = 0;
		co_await WaitClk(clock);

		simu(cmd.opcode) = 5;
		co_await WaitClk(clock);

		simu(cmd.opcode) = 0;
		co_await WaitClk(clock);

		simu(cmd.valid) = 0;
		co_await WaitClk(clock);

		simu(cmd.sequenceNumber) = 2;
		co_await WaitClk(clock);

		simu(cmd.sequenceNumber) = 128;
		co_await WaitClk(clock);

		for (size_t i = 60; i < 70; ++i)
		{
			simu(cmd.valid) = 1;
			simu(cmd.sequenceNumber) = i;
			co_await WaitClk(clock);
		}

		for (size_t i = 80; i < 100; i += 2)
		{
			simu(cmd.valid) = 1;
			simu(cmd.sequenceNumber) = i;
			co_await WaitClk(clock);
		}

		for (size_t i = 81; i < 100; i += 2)
		{
			simu(cmd.valid) = 1;
			simu(cmd.sequenceNumber) = i;
			co_await WaitClk(clock);
		}

		simu(cmd.valid) = 1;
		simu(cmd.sequenceNumber) = 164;
		co_await WaitClk(clock);
	});

	addSimulationProcess(arc::AntiReplayVerifier{ clock, cmd, out, ar.windowSize() });

	//sim::VCDSink vcd{ design.getCircuit(), getSimulator(), "AntiReplay_comb_retiming_test.vcd" };
	//vcd.addAllPins();
	//vcd.addAllNamedSignals();

	design.getCircuit().postprocess(gtry::DefaultPostprocessing{});
	//design.visualize("AntiReplay_comb_retiming_test");

	runTicks(clock.getClk(), 64);
}

BOOST_FIXTURE_TEST_CASE(AntiReplay_comb_fuzz_test, BoostUnitTestSimulationFixture)
{
	Clock clock(ClockConfig{}.setAbsoluteFrequency(100'000'000).setName("clock"));
	ClockScope clkScp(clock);

	arc::CmdStage cmd;
	cmd.connectionIndex = 1_b;
	cmd.windowSize = 10_b;
	cmd.sequenceNumber = 64_b;

	arc::AntiReplayResult out;
	cmd.pinIn("cmd_");

	arc::AntiReplay ar{ cmd };
	ar.windowSize(6_b, 1_b);
	ar.generateSingleStage();
	ar.result(out);

	pinOut(out.valid).setName("out_valid");
	pinOut(out.code).setName("out_code");

	arc::AntiReplayFuzzer fuzzer{ 18055 };
	fuzzer.distanceBits(9);
	fuzzer.saMask(1);

	addSimulationProcess([&]() -> SimProcess {

		for (size_t i = 0; i < 2; ++i)
		{
			simu(cmd.valid) = 1;
			simu(cmd.opcode) = 2;
			simu(cmd.connectionIndex) = i;
			simu(cmd.sequenceNumber) = 0;
			simu(cmd.windowSize) = 0;
			co_await WaitClk(clock);

			simu(cmd.opcode) = 5;
			co_await WaitClk(clock);
		}
		simu(cmd.opcode) = 3;
		simu(cmd.windowSize) = 128;
		co_await WaitClk(clock);

		simu(cmd.valid) = 0;
		simu(cmd.opcode) = 0;
		simu(cmd.windowSize) = 0;

		for (size_t i = 0; i < 4; ++i)
			co_await WaitClk(clock);

		simu(cmd.valid) = 1;
		while (true)
		{
			auto [saId, saIdx] = fuzzer.next();
			simu(cmd.connectionIndex) = saId;
			simu(cmd.sequenceNumber) = saIdx;
			co_await WaitClk(clock);
		}

	});

	addSimulationProcess(arc::AntiReplayVerifier{ clock, cmd, out, ar.windowSize() });

	//sim::VCDSink vcd{ design.getCircuit(), getSimulator(), "AntiReplay_comb_fuzz_test.vcd" };
	//vcd.addAllPins();
	//vcd.addAllNamedSignals();

	design.getCircuit().postprocess(gtry::DefaultPostprocessing{});
	//design.visualize("AntiReplay_comb_fuzz_test");

	runTicks(clock.getClk(), 1024);
}

BOOST_FIXTURE_TEST_CASE(AntiReplay_qdr_fuzz_test, BoostUnitTestSimulationFixture)
{
	Clock clock(ClockConfig{}.setAbsoluteFrequency(100'000'000).setName("clock"));
	ClockScope clkScp(clock);

	arc::CmdStage cmd;
	cmd.connectionIndex = 1_b;
	cmd.windowSize = 12_b;
	cmd.sequenceNumber = 64_b;

	arc::AntiReplayResult out;
	cmd.pinIn("cmd_");

	arc::AntiReplay ar{ cmd };
	ar.windowSize(8_b, 16_b);
	ar.generateFixedLatency();
	ar.result(out);

	pinOut(out.valid).setName("out_valid");
	pinOut(out.code).setName("out_code");

	arc::AntiReplayFuzzer fuzzer{ 18055 };
	fuzzer.distanceBits(10);
	fuzzer.saMask(1);

	addSimulationProcess([&]() -> SimProcess {

		for (size_t i = 0; i < 2; ++i)
		{
			simu(cmd.valid) = 1;
			simu(cmd.opcode) = 2;
			simu(cmd.connectionIndex) = i;
			simu(cmd.sequenceNumber) = 0;
			simu(cmd.windowSize) = 0;
			co_await WaitClk(clock);

			simu(cmd.opcode) = 5;
			co_await WaitClk(clock);
		}
		simu(cmd.opcode) = 3;
		simu(cmd.windowSize) = 300;
		co_await WaitClk(clock);

		simu(cmd.valid) = 0;
		simu(cmd.opcode) = 0;
		simu(cmd.windowSize) = 0;

		for (size_t i = 0; i < 4; ++i)
			co_await WaitClk(clock);

		simu(cmd.valid) = 1;
		while (true)
		{
			auto [saId, saIdx] = fuzzer.next();
			simu(cmd.connectionIndex) = saId;
			simu(cmd.sequenceNumber) = saIdx;
			co_await WaitClk(clock);
		}

		});

	addSimulationProcess(arc::AntiReplayVerifier{ clock, cmd, out, ar.windowSize() });

	sim::VCDSink vcd{ design.getCircuit(), getSimulator(), "AntiReplay_qdr_fuzz_test.vcd" };
	vcd.addAllPins();
//	vcd.addAllNamedSignals();
/*
	{
		auto subnet = hlim::Subnet::all(design.getCircuit());
		design.getCircuit().optimizeSubnet(subnet);
		design.visualize("AntiReplay_qdr_fuzz_test_before");
	}
*/
	//design.visualize("AntiReplay_qdr_fuzz_test_before");
	design.getCircuit().postprocess(gtry::DefaultPostprocessing{});
	//design.visualize("AntiReplay_qdr_fuzz_test");

	//{
	//	auto subnet = hlim::ConstSubnet::all(design.getCircuit());
	//
	//	{
	//		auto* group = design.getCircuit().getRootNodeGroup()->findChild("AntiReplay")->findChild("ringBufferMask");
	//		for (auto* n : group->getNodes())
	//			subnet.remove(n);
	//	}
	//	{
	//		auto* group = design.getCircuit().getRootNodeGroup()->findChild("AntiReplay")->findChild("result");
	//		for (auto* n : group->getNodes())
	//			subnet.remove(n);
	//	}
	//	{
	//		auto* group = design.getCircuit().getRootNodeGroup()->findChild("AntiReplay")->findChild("FirstLookupStage")->findChild("flat_mux");
	//		for (auto* n : group->getNodes())
	//			subnet.remove(n);
	//	}
	//
	//
	//	DotExport exp("AntiReplay_qdr_fuzz_test_small.dot");
	//	exp(design.getCircuit(), subnet);
	//	exp.runGraphViz("AntiReplay_qdr_fuzz_test_small.svg");
	//}

	runTicks(clock.getClk(), 1024);
}

BOOST_FIXTURE_TEST_CASE(merge_test, BoostUnitTestSimulationFixture)
{
	Clock clock(ClockConfig{}.setAbsoluteFrequency(200'000'000).setName("clock"));
	ClockScope clkScp(clock);

	arc::Stream<std::tuple<arc::FrameStream, Bit>> data = { .data = std::make_tuple(32_b, Bit{}) };
	arc::Stream<BVec> result{ .data = 3_b };
	arc::Stream<std::tuple<arc::FrameStream, arc::Valid<BVec>>> out;
	arc::merge(data, result, out);

	data.pinIn("data_");
	data.get<arc::FrameStream>().pinIn("data_frame_");
	data.get<Bit>() = pinIn().setName("data_merge");

	result.pinIn("result_");
	result.data = pinIn(result.data.getWidth()).setName("result_data");

	out.pinOut("out_");
	out.get<arc::FrameStream>().pinOut("out_frame_");
	pinOut(out.get<arc::Valid<BVec>>().valid).setName("out_result_valid");
	pinOut(out.get<arc::Valid<BVec>>().data).setName("out_result_data");

	size_t testPhaseCounter = 0;
	BVec testPhase = pinIn(8_b).setName("phase");

	addSimulationProcess([&]() -> SimProcess {
		auto& dataIn = data.get<arc::FrameStream>();
		auto& dataOut = out.get<arc::FrameStream>();
		auto& mergeIn = data.get<Bit>();
		auto& resultOut = out.get<arc::Valid<BVec>>();

		simu(data.valid) = 0;
		simu(result.valid) = 0;
		simu(out.ready) = 0;
		simu(testPhase) = testPhaseCounter++;
		co_await WaitClk(clock);

		// result always present cases
		simu(result.valid) = 1;
		simu(result.data) = 7;
		co_await WaitClk(clock);
		BOOST_TEST(simu(out.valid) == 0);

		// push unrelated data frame
		simu(mergeIn) = 0;
		simu(dataIn.data) = 0;
		simu(dataIn.empty) = 0;
		simu(dataIn.startofpacket) = 1;
		simu(dataIn.endofpacket) = 1;
		simu(dataIn.valid) = 1;
		simu(data.valid) = 1;
		simu(testPhase) = testPhaseCounter++;
		co_await WaitClk(clock);
		BOOST_TEST(simu(out.valid) == 1);
		BOOST_TEST(simu(dataOut.valid) == 1);
		BOOST_TEST(simu(data.ready) == 0);
		BOOST_TEST(simu(result.ready) == 0);

		simu(out.ready) = 1;
		simu(testPhase) = testPhaseCounter++;
		co_await WaitClk(clock);
		BOOST_TEST(simu(out.valid) == 1);
		BOOST_TEST(simu(dataOut.valid) == 1);
		BOOST_TEST(simu(data.ready) == 1);
		BOOST_TEST(simu(result.ready) == 0);

		simu(out.ready) = 0;
		simu(mergeIn) = 1;
		simu(testPhase) = testPhaseCounter++;
		co_await WaitClk(clock);
		BOOST_TEST(simu(out.valid) == 1);
		BOOST_TEST(simu(resultOut.valid) == 1);
		BOOST_TEST(simu(dataOut.valid) == 1);
		BOOST_TEST(simu(data.ready) == 0);
		BOOST_TEST(simu(result.ready) == 0);

		simu(out.ready) = 1;
		simu(testPhase) = testPhaseCounter++;
		co_await WaitClk(clock);
		BOOST_TEST(simu(out.valid) == 1);
		BOOST_TEST(simu(resultOut.valid) == 1);
		BOOST_TEST(simu(dataOut.valid) == 1);
		BOOST_TEST(simu(data.ready) == 1);
		BOOST_TEST(simu(result.ready) == 1);

		// merge without data
		simu(out.ready) = 0;
		simu(dataIn.valid) = 0;
		simu(testPhase) = testPhaseCounter++;
		co_await WaitClk(clock);
		BOOST_TEST(simu(out.valid) == 1);
		BOOST_TEST(simu(resultOut.valid) == 1);
		BOOST_TEST(simu(dataOut.valid) == 0);
		BOOST_TEST(simu(result.ready) == 0);

		simu(out.ready) = 1;
		simu(dataIn.valid) = 0;
		simu(testPhase) = testPhaseCounter++;
		co_await WaitClk(clock);
		BOOST_TEST(simu(out.valid) == 1);
		BOOST_TEST(simu(resultOut.valid) == 1);
		BOOST_TEST(simu(dataOut.valid) == 0);
		BOOST_TEST(simu(result.ready) == 1);

		// stall on result without data
		simu(out.ready) = 0;
		simu(result.valid) = 0;
		simu(testPhase) = testPhaseCounter++;
		co_await WaitClk(clock);
		BOOST_TEST(simu(data.ready) == 0);

		simu(out.ready) = 1;
		simu(testPhase) = testPhaseCounter++;
		co_await WaitClk(clock);
		BOOST_TEST(simu(data.ready) == 0);

		simu(result.valid) = 1;
		simu(testPhase) = testPhaseCounter++;
		co_await WaitClk(clock);
		BOOST_TEST(simu(data.ready) == 1);
		BOOST_TEST(simu(result.ready) == 1);

		// stall on result with data
		simu(out.ready) = 0;
		simu(result.valid) = 0;
		simu(dataIn.valid) = 1;
		simu(testPhase) = testPhaseCounter++;
		co_await WaitClk(clock);
		BOOST_TEST(simu(data.ready) == 0);

		simu(out.ready) = 1;
		simu(testPhase) = testPhaseCounter++;
		co_await WaitClk(clock);
		BOOST_TEST(simu(data.ready) == 0);

		simu(result.valid) = 1;
		simu(testPhase) = testPhaseCounter++;
		co_await WaitClk(clock);
		BOOST_TEST(simu(data.ready) == 1);
		BOOST_TEST(simu(result.ready) == 1);

		simu(data.valid) = 0;
		simu(result.valid) = 0;
		simu(testPhase) = testPhaseCounter++;
		co_await WaitClk(clock);

		// multi cycle data with merge stall
		simu(result.valid) = 0;
		simu(data.valid) = 1;
		simu(dataIn.valid) = 1;
		simu(dataIn.endofpacket) = 0;
		simu(dataIn.data) = 1;
		simu(mergeIn) = 0;
		simu(testPhase) = testPhaseCounter++;
		co_await WaitClk(clock);
		BOOST_TEST(simu(data.ready) == 1);

		simu(mergeIn) = 0;
		simu(dataIn.startofpacket) = 0;
		simu(dataIn.data) = 2;
		co_await WaitClk(clock);
		BOOST_TEST(simu(data.ready) == 1);

		simu(dataIn.endofpacket) = 1;
		simu(dataIn.data) = 3;
		simu(mergeIn) = 1;
		co_await WaitClk(clock);
		BOOST_TEST(simu(data.ready) == 0);

		simu(result.valid) = 1;
		co_await WaitClk(clock);
		BOOST_TEST(simu(data.ready) == 1);
		BOOST_TEST(simu(result.ready) == 1);
		BOOST_TEST(simu(out.valid) == 1);

		// stop test
		simu(data.valid) = 0;
		simu(result.valid) = 0;
		simu(out.ready) = 0;
		simu(mergeIn) = 0;
		co_await WaitClk(clock);
	});

	//sim::VCDSink vcd{ design.getCircuit(), getSimulator(), "merge_test.vcd" };
	//vcd.addAllPins();
	//vcd.addAllNamedSignals();

	design.getCircuit().postprocess(gtry::DefaultPostprocessing{});

	runTicks(clock.getClk(), 32);
}

BOOST_FIXTURE_TEST_CASE(FrameAdapter_test, BoostUnitTestSimulationFixture)
{
	Clock clock(ClockConfig{}.setAbsoluteFrequency(200'000'000).setName("clock"));
	ClockScope clkScp(clock);

	arc::CmdStage cmd;
	cmd.connectionIndex = 2_b;
	cmd.windowSize = 16_b;
	cmd.sequenceNumber = 64_b;
	cmd.pinIn("in_cmd_");

	arc::FrameAdapter adapter;
	adapter.setupAntiReplay(cmd).windowSize(4_b, 16_b).generateFixedLatency();
	arc::FrameStream& dataIn = adapter.dataIn(32_b);
	setName(dataIn, "dbg_dataIn");
	auto& out = adapter.generate();
	adapter.assignPins();

	addSimulationProcess([&]() -> SimProcess {

		simu(dataIn.valid) = 0;
		simu(cmd.valid) = 0;
		simu(out.ready) = 0;
		co_await WaitClk(clock);

		// push frame
		for (size_t i = 0; i < 3; ++i)
		{
			simu(dataIn.valid) = 1;
			simu(dataIn.data) = i;
			simu(dataIn.empty) = 0;
			simu(dataIn.startofpacket) = i == 0 ? 1 : 0;
			simu(dataIn.endofpacket) = i == 2 ? 1 : 0;
			co_await WaitClk(clock);
		}
		simu(dataIn.valid) = 0;
		co_await WaitClk(clock);

		simu(out.ready) = 1;
		co_await WaitClk(clock);
		co_await WaitClk(clock);
		co_await WaitClk(clock);

		for (size_t i = 0; i < 3; ++i)
		{
			simu(dataIn.valid) = 1;
			simu(dataIn.data) = i;
			simu(dataIn.empty) = 0;
			simu(dataIn.startofpacket) = i == 0 ? 1 : 0;
			simu(dataIn.endofpacket) = i == 2 ? 1 : 0;
			co_await WaitClk(clock);
		}

		co_return;
	});

	//sim::VCDSink vcd{ design.getCircuit(), getSimulator(), "FrameAdapter_test.vcd" };
	//vcd.addAllPins();
	//vcd.addAllNamedSignals();

	design.getCircuit().postprocess(gtry::DefaultPostprocessing{});

	//{
	//	hlim::Subnet subnet;// = hlim::Subnet::fromNodeGroup(design.getCircuit().getRootNodeGroup()->findChild("scl_fifo
	//	for (auto &n : design.getCircuit().getNodes())
	//		if (n->getGroup()->getName() == "scl_fifo") {
	//			subnet.add(n.get());
	//		}
	//
	//	subnet.dilate(true, true);
	//	subnet.dilate(true, true);
	//	//subnet.dilate(true, true);
	//
	//	DotExport exp("fifo.dot");
	//	exp(design.getCircuit(), subnet.asConst());
	//	exp.runGraphViz("fifo.svg");
	//}


	runTicks(clock.getClk(), 32);
}
BOOST_FIXTURE_TEST_CASE(AntiReplay_qdr_export_test, BoostUnitTestSimulationFixture)
{
	Clock clock(ClockConfig{}.setAbsoluteFrequency(200'000'000).setName("clock").setResetType(ClockConfig::ResetType::NONE));
	ClockScope clkScp(clock);

	arc::CmdStage cmd;
	cmd.connectionIndex = 7_b;
	cmd.windowSize = 16_b;
	cmd.sequenceNumber = 64_b;

	arc::AntiReplayResult out;
	cmd.pinIn("cmd_");

	arc::AntiReplay ar{ cmd };
	ar.windowSize(12_b, 128_b);
	ar.generateFixedLatency();
	ar.result(out);

	//std::cerr << "pipeline stages: " << ar.latency() << '\n';

	pinOut(out.valid).setName("out_valid");
	pinOut(out.code).setName("out_code");

	arc::AntiReplayFuzzer fuzzer{ 18055 };
	fuzzer.distanceBits(10);
	fuzzer.saMask(1);

	addSimulationProcess([&]() -> SimProcess {

		for (size_t i = 0; i < 2; ++i)
		{
			simu(cmd.valid) = 1;
			simu(cmd.opcode) = 2;
			simu(cmd.connectionIndex) = i;
			simu(cmd.sequenceNumber) = 0;
			simu(cmd.windowSize) = 0;
			co_await WaitClk(clock);

			simu(cmd.opcode) = 5;
			co_await WaitClk(clock);
		}
		simu(cmd.opcode) = 3;
		simu(cmd.windowSize) = 300;
		co_await WaitClk(clock);

		simu(cmd.valid) = 0;
		simu(cmd.opcode) = 0;
		simu(cmd.windowSize) = 0;

		for (size_t i = 0; i < 4; ++i)
			co_await WaitClk(clock);

		simu(cmd.valid) = 1;
		while (true)
		{
			auto [saId, saIdx] = fuzzer.next();
			simu(cmd.connectionIndex) = saId;
			simu(cmd.sequenceNumber) = saIdx;
			co_await WaitClk(clock);
		}

		});

	addSimulationProcess(arc::AntiReplayVerifier{ clock, cmd, out, ar.windowSize() });

	//sim::VCDSink vcd{ design.getCircuit(), getSimulator(), "AntiReplay_qdr_export_test.vcd" };
	//vcd.addAllPins();
	//vcd.addAllNamedSignals();

	design.getCircuit().postprocess(gtry::DefaultPostprocessing{});

	//gtry::vhdl::VHDLExport vhdl{ "export/AntiReplay_qdr_export_test/AntiReplay_qdr_export_test.vhdl" };
	//vhdl.targetSynthesisTool(new XilinxVivado());
	//vhdl.writeProjectFile("AntiReplay_qdr_export_test.tcl");
	//vhdl.writeClocksFile("AntiReplay_qdr_export_test.xdc");
	//vhdl.addTestbenchRecorder(getSimulator(), "AntiReplay_qdr_export_test_TB");
	//vhdl(design.getCircuit());


	runTicks(clock.getClk(), 1024*4);
}

enum class TestOp
{
	udpate = 0,
	activate = 5,
	deactivate = 1,
	setseqnr = 2,
	setwindowsize = 3
};

BOOST_FIXTURE_TEST_CASE(AntiReplay_active_test, BoostUnitTestSimulationFixture)
{
	Clock clock(ClockConfig{}.setAbsoluteFrequency(200'000'000).setName("clock").setResetType(ClockConfig::ResetType::NONE));
	ClockScope clkScp(clock);

	arc::CmdStage cmd;
	cmd.connectionIndex = 1_b;
	cmd.windowSize = 16_b;
	cmd.sequenceNumber = 64_b;
	cmd.pinIn("cmd_");

	arc::AntiReplay ar{ cmd };
	Bit connectionActive = ar.active();
	pinOut(connectionActive).setName("connectionActive");

	addSimulationProcess([&]() -> SimProcess {

		std::array<uint32_t, 2> connState{ {} };

		// check initial state (connection disabled)
		simu(cmd.valid) = 1;
		simu(cmd.opcode) = (uint32_t)TestOp::udpate;
		simu(cmd.sequenceNumber) = 0;
		simu(cmd.windowSize) = 0;

		for (size_t i = 0; i < connState.size(); ++i)
		{
			simu(cmd.connectionIndex) = i;
			co_await WaitClk(clock);
			BOOST_TEST(simu(connectionActive) == connState[i]);
		}

		for (size_t i = 0; i < connState.size(); ++i)
		{
			simu(cmd.connectionIndex) = i;

			// activate connection
			simu(cmd.opcode) = (uint32_t)TestOp::activate;
			co_await WaitClk(clock);

			// check state
			simu(cmd.opcode) = (uint32_t)TestOp::udpate;
			co_await WaitClk(clock);
			BOOST_TEST(simu(connectionActive) == 1);

			// deactivate connection
			simu(cmd.opcode) = (uint32_t)TestOp::deactivate;
			co_await WaitClk(clock);

			// check state
			simu(cmd.opcode) = (uint32_t)TestOp::udpate;
			co_await WaitClk(clock);
			BOOST_TEST(simu(connectionActive) == 0);
		}

		// activate second connection to have both states
		simu(cmd.opcode) = (uint32_t)TestOp::activate;
		simu(cmd.connectionIndex) = 1;
		connState[1] = true;
		co_await WaitClk(clock);

		// check that other command have no effect on active state
		for (size_t i = 0; i < connState.size(); ++i)
		{
			simu(cmd.connectionIndex) = i;

			simu(cmd.opcode) = (uint32_t)TestOp::setseqnr;
			co_await WaitClk(clock);
			BOOST_TEST(simu(connectionActive) == connState[i]);

			simu(cmd.opcode) = (uint32_t)TestOp::setwindowsize;
			co_await WaitClk(clock);
			BOOST_TEST(simu(connectionActive) == connState[i]);

			simu(cmd.opcode) = (uint32_t)TestOp::udpate;
			co_await WaitClk(clock);
			BOOST_TEST(simu(connectionActive) == connState[i]);
		}

	});

	//sim::VCDSink vcd{ design.getCircuit(), getSimulator(), "AntiReplay_active_test.vcd" };
	//vcd.addAllPins();
	//vcd.addAllNamedSignals();

	design.getCircuit().postprocess(gtry::DefaultPostprocessing{});
	runTicks(clock.getClk(), 32);
}


BOOST_DATA_TEST_CASE_F(BoostUnitTestSimulationFixture, AntiReplay_lastSeqNr_test, data::make({ 16, 8, 1 }), bitPerStep)
{
	Clock clock(ClockConfig{}.setAbsoluteFrequency(200'000'000).setName("clock").setResetType(ClockConfig::ResetType::NONE));
	ClockScope clkScp(clock);

	arc::CmdStage cmd;
	cmd.connectionIndex = 1_b;
	cmd.windowSize = 16_b;
	cmd.sequenceNumber = 64_b;
	//cmd.pinIn("cmd_");

	arc::AntiReplay ar{ cmd };

	BVec connIdx = gtry::pinIn(1_b).setName("connIdx");
	BVec seqNr = gtry::pinIn(16_b).setName("seqNr");
	Bit set = gtry::pinIn().setName("set");
	Bit update = gtry::pinIn().setName("update");
	auto [aboveWindow, lastSeqNr] = ar.lastSequenceNumber(connIdx, seqNr, set, update, bitPerStep);
	pinOut(aboveWindow).setName("aboveWindow");
	pinOut(lastSeqNr).setName("lastSeqNr");

	addSimulationProcess([&]() -> SimProcess {

		for (size_t c = 0; c < 2; ++c)
		{
			simu(connIdx) = c;
			simu(seqNr) = 0;
			simu(set) = 1;
			simu(update) = 0;
			co_await WaitClk(clock);

			simu(set) = 0;
			BOOST_TEST(simu(aboveWindow) == 0);
			BOOST_TEST(simu(lastSeqNr) == 0);

			simu(update) = 1;
			for (size_t s = 1; s < 5; ++s)
			{
				size_t target = s * 127;
				simu(seqNr) = target;

				BOOST_TEST(simu(aboveWindow) == 1);
				BOOST_TEST(simu(lastSeqNr) == target - 127);

				co_await WaitClk(clock);

				BOOST_TEST(simu(aboveWindow) == 0);
				BOOST_TEST(simu(lastSeqNr) == target);

				for (size_t i = 0; i < 3; ++i)
				{
					simu(seqNr) = target - i * 3;
					BOOST_TEST(simu(aboveWindow) == 0);
					BOOST_TEST(simu(lastSeqNr) == target);
					co_await WaitClk(clock);
				}
			}

			const size_t highEsn = 4 * 127;

			simu(update) = 0;
			simu(seqNr) = 1;
			BOOST_TEST(simu(aboveWindow) == 0);
			BOOST_TEST(simu(lastSeqNr) == highEsn);
			co_await WaitClk(clock);

			simu(seqNr) = highEsn;
			BOOST_TEST(simu(aboveWindow) == 0);
			BOOST_TEST(simu(lastSeqNr) == highEsn);
			co_await WaitClk(clock);

			simu(seqNr) = highEsn + 1;
			BOOST_TEST(simu(aboveWindow) == 1);
			BOOST_TEST(simu(lastSeqNr) == highEsn);
			co_await WaitClk(clock);

			simu(seqNr) = ~0ull;
			BOOST_TEST(simu(aboveWindow) == 1);
			BOOST_TEST(simu(lastSeqNr) == highEsn);
			co_await WaitClk(clock);

		}
	});

	//sim::VCDSink vcd{ design.getCircuit(), getSimulator(), "AntiReplay_lastSeqNr_test.vcd" };
	//vcd.addAllPins();
	//vcd.addAllNamedSignals();

	design.getCircuit().postprocess(gtry::DefaultPostprocessing{});
	runTicks(clock.getClk(), 64);
}

BOOST_FIXTURE_TEST_CASE(AntiReplay_windowBounds_test, AntiReplayTest)
{
	windowSize(4_b, 1_b);
	const size_t winSize = windowSize().bits();
	BOOST_TEST(winSize == 15);

	BVec lastSeqNr = pinIn(32_b).setName("lastSeqNr");
	Bit aboveWindow = pinIn().setName("aboveWindow");
	auto [inPhysWin, inVirtWin] = windowBounds(lastSeqNr, !aboveWindow);
	pinOut(inPhysWin).setName("inPhysWin");
	pinOut(inVirtWin).setName("inVirtWin");

	addSimulationProcess([=]() -> SimProcess {

		// set defaults
		simu(cmd.valid) = 1;
		simu(cmd.opcode) = (uint32_t)TestOp::udpate;
		simu(cmd.windowSize) = 16;

		simu(lastSeqNr) = 32;
		simu(aboveWindow) = 0;

		for (size_t i = 0; i < 33; ++i)
		{
			simu(cmd.sequenceNumber) = i;
			size_t inWin = (i > 32 - winSize) ? 1 : 0;
			BOOST_TEST(simu(inPhysWin) == inWin);
			BOOST_TEST(simu(inVirtWin) == inWin);
			co_await WaitClk(clock());
		}

		// check bigger virtual window
		simu(cmd.opcode) = (uint32_t)TestOp::setwindowsize;
		simu(cmd.windowSize) = 16;

		for (size_t i = 0; i < 33; ++i)
		{
			simu(cmd.sequenceNumber) = i;
			size_t isPhysWin = (i > 32 - winSize) ? 1 : 0;
			size_t isVirtWin = (i > 32 - 16) ? 1 : 0;
			BOOST_TEST(simu(inPhysWin) == isPhysWin);
			BOOST_TEST(simu(inVirtWin) == isVirtWin);
			co_await WaitClk(clock());
		}

		// check smaller virtual window
		simu(cmd.opcode) = (uint32_t)TestOp::setwindowsize;
		simu(cmd.windowSize) = 14;

		for (size_t i = 0; i < 33; ++i)
		{
			simu(cmd.sequenceNumber) = i;
			size_t isPhysWin = (i > 32 - winSize) ? 1 : 0;
			size_t isVirtWin = (i > 32 - 14) ? 1 : 0;
			BOOST_TEST(simu(inPhysWin) == isPhysWin);
			BOOST_TEST(simu(inVirtWin) == isVirtWin);
			co_await WaitClk(clock());
		}

		// no virtual window
		simu(cmd.opcode) = (uint32_t)TestOp::setwindowsize;
		simu(cmd.windowSize) = 0;

		for (size_t i = 0; i < 33; ++i)
		{
			simu(cmd.sequenceNumber) = i;
			size_t isPhysWin = (i > 32 - winSize) ? 1 : 0;
			BOOST_TEST(simu(inPhysWin) == isPhysWin);
			BOOST_TEST(simu(inVirtWin) == 0);
			co_await WaitClk(clock());
		}

		stopTest();
	});
}

BOOST_FIXTURE_TEST_CASE(AntiReplay_ringBufferMask_get_before_put_test, AntiReplayTest)
{
	BVec put = pinIn(4_b).setName("put");
	BVec get = pinIn(4_b).setName("get");
	BVec mask = ringBufferMask(put, get);
	pinOut(mask).setName("mask");

	addSimulationProcess([=]() -> SimProcess {

		// mask all bits between get and put if get <= put
		for(size_t p = 0; p < 16; ++p)
			for (size_t g = 0; g <= p; ++g)
			{
				simu(put) = p;
				simu(get) = g;

				const size_t result = simu(mask);
				for (size_t i = 0; i < 16; ++i)
				{
					if (i <= g || i >= p)
						BOOST_TEST((result & (1ull << i)) == 0);
					else
						BOOST_TEST((result & (1ull << i)) != 0);
				}
				co_await WaitClk(clock());
			}

		stopTest();
	});
}

BOOST_FIXTURE_TEST_CASE(AntiReplay_ringBufferMask_put_before_get_test, AntiReplayTest)
{
	BVec put = pinIn(4_b).setName("put");
	BVec get = pinIn(4_b).setName("get");
	BVec mask = ringBufferMask(put, get);
	pinOut(mask).setName("mask");

	addSimulationProcess([=]() -> SimProcess {

		// mask all bits not between get and put if put < get
		for (size_t g = 0; g < 16; ++g)
			for (size_t p = 0; p < g; ++p)
			{
				simu(put) = p;
				simu(get) = g;

				const size_t result = simu(mask);
				for (size_t i = 0; i < 16; ++i)
				{
					if (i > g || i < p)
						BOOST_TEST((result & (1ull << i)) != 0);
					else
						BOOST_TEST((result & (1ull << i)) == 0);
				}
				co_await WaitClk(clock());
			}

		stopTest();
	});
}

BOOST_FIXTURE_TEST_CASE(AntiReplay_mux_large_test, AntiReplayTest)
{
	BVec sel = pinIn(5_b).setName("sel");
	BVec in = pinIn(BitWidth{ sel.getWidth().count() }).setName("in");
	Bit out = mux_large_internal(sel, in, 2).front();
	pinOut(out).setName("out");

	addSimulationProcess([=]() -> SimProcess {

		for (size_t i = 0; i < sel.getWidth().count(); ++i)
		{
			simu(sel) = i;
			
			simu(in) = ~(1ull << i);
			BOOST_TEST(simu(out) == 0);
			co_await WaitClk(clock());

			simu(in) = 1ull << i;
			BOOST_TEST(simu(out) == 1);
			co_await WaitClk(clock());
		}
		
		stopTest();
	});
}

class AntiReplayFirstStageTest : public AntiReplayTest
{
public:
	void setup() override
	{
		AntiReplayTest::setup();

		lastSeqNr = pinIn(32_b).setName("lastSeqNr");
		clearWindow = pinIn().setName("clearWindow");
		setWordValid = pinIn().setName("setWordValid");
		updateLastSeqNr = pinIn().setName("updateLastSeqNr");
		secondValid = firstLookupStage(lastSeqNr, clearWindow, setWordValid, updateLastSeqNr);
		pinOut(secondValid).setName("secondValid");

		segmentSize = m_cfg.secondWordWidth().bits();
		numBits = m_cfg.firstWordWidth().bits();

		addSimulationProcess([=]() -> SimProcess {

			simu(cmd.valid) = 1;
			simu(cmd.opcode) = 0;
			simu(cmd.connectionIndex) = 0;
			simu(cmd.sequenceNumber) = 0;

			simu(lastSeqNr) = 0;
			simu(setWordValid) = 0;
			simu(clearWindow) = 0;
			simu(updateLastSeqNr) = 0;

			co_return;
		});
	}

protected:
	BVec lastSeqNr;
	Bit clearWindow;
	Bit setWordValid;
	Bit updateLastSeqNr;
	Bit secondValid;

	size_t segmentSize = 0;
	size_t numBits = 0;

};
BOOST_FIXTURE_TEST_CASE(AntiReplay_firstStage_clear_test, AntiReplayFirstStageTest)
{
	addSimulationProcess([=]() -> SimProcess {

		// set all bits
		simu(setWordValid) = 1;
		for (size_t i = 0; i < numBits; ++i)
		{
			simu(cmd.sequenceNumber) = i * segmentSize;
			co_await WaitClk(clock());
			BOOST_TEST(simu(secondValid) == 1);
		}
		simu(setWordValid) = 0;

		// clear
		simu(clearWindow) = 1;
		co_await WaitClk(clock());
		simu(clearWindow) = 0;

		// check bit state
		for (size_t i = 0; i < numBits; ++i)
		{
			simu(cmd.sequenceNumber) = i * segmentSize;
			BOOST_TEST(simu(secondValid) == 0);
			co_await WaitClk(clock());
		}

		stopTest();
	});
}

BOOST_FIXTURE_TEST_CASE(AntiReplay_firstStage_update_test, AntiReplayFirstStageTest)
{
	addSimulationProcess([=]() -> SimProcess {

		// clear and set bitmap
		simu(cmd.sequenceNumber) = 0;
		simu(clearWindow) = 1;
		simu(setWordValid) = 1;
		co_await WaitClk(clock());
		simu(clearWindow) = 0;
		BOOST_TEST(simu(secondValid) == 1);

		// set bit to clear later
		simu(cmd.sequenceNumber) = 1 * segmentSize;
		co_await WaitClk(clock());
		BOOST_TEST(simu(secondValid) == 1);

		// clear range of bits
		simu(cmd.sequenceNumber) = 2 * segmentSize;
		simu(setWordValid) = 1;
		simu(updateLastSeqNr) = 1;
		co_await WaitClk(clock());

		simu(setWordValid) = 0;
		simu(updateLastSeqNr) = 0;

		// check bit state
		for (size_t i = 0; i < numBits; ++i)
		{
			simu(cmd.sequenceNumber) = i * segmentSize;
			size_t expected = i == 0 || i == 2 ? 1 : 0;
			BOOST_TEST(simu(secondValid) == expected);
			co_await WaitClk(clock());
		}

		stopTest();
	});
}

BOOST_FIXTURE_TEST_CASE(AntiReplay_firstStage_update_overrun_test, AntiReplayFirstStageTest)
{
	addSimulationProcess([=]() -> SimProcess {

		// set all bits
		simu(setWordValid) = 1;
		for (size_t i = 0; i < numBits; ++i)
		{
			simu(cmd.sequenceNumber) = i * segmentSize;
			co_await WaitClk(clock());
			BOOST_TEST(simu(secondValid) == 1);
		}

		// clear range put < get
		simu(lastSeqNr) = 2 * segmentSize;
		simu(cmd.sequenceNumber) = numBits * segmentSize;
		simu(updateLastSeqNr) = 1;
		co_await WaitClk(clock());
		simu(updateLastSeqNr) = 0;

		// check all bits
		for (size_t i = 0; i < numBits; ++i)
		{
			simu(cmd.sequenceNumber) = i * segmentSize;
			size_t expected = i < 3 ? 1 : 0;
			BOOST_TEST(simu(secondValid) == expected);
			co_await WaitClk(clock());
		}

		stopTest();
	});
}

BOOST_FIXTURE_TEST_CASE(AntiReplay_firstStage_update_over_window_test, AntiReplayFirstStageTest)
{
	addSimulationProcess([=]() -> SimProcess {

		for (size_t r = 0; r < 4; ++r)
		{
			// set all bits
			simu(setWordValid) = 1;
			for (size_t i = 0; i < numBits; ++i)
			{
				simu(cmd.sequenceNumber) = i * segmentSize;
				co_await WaitClk(clock());
				BOOST_TEST(simu(secondValid) == 1);
			}
			simu(setWordValid) = 0;

			// clear all
			simu(cmd.sequenceNumber) = (numBits + r) * segmentSize;
			simu(updateLastSeqNr) = 1;
			co_await WaitClk(clock());
			simu(updateLastSeqNr) = 0;

			// check all bits
			for (size_t i = 0; i < numBits; ++i)
			{
				simu(cmd.sequenceNumber) = i * segmentSize;
				BOOST_TEST(simu(secondValid) == 0);
				co_await WaitClk(clock());
			}
		}

		stopTest();
	});
}

BOOST_FIXTURE_TEST_CASE(AntiReplay_secondStageUpdate_seqseqnr_test, AntiReplayTest)
{
	windowSize(8_b, 32_b);

	BVec lastSeqNr = pinIn(32_b).setName("lastSeqNr");
	Bit aboveWindow = pinIn().setName("aboveWindow");
	BVec bitmap = pinIn(32_b).setName("bitmap");
	BVec bitmap_next = secondStageUpdate(bitmap, lastSeqNr, aboveWindow);
	pinOut(bitmap_next).setName("bitmap_next");

	addSimulationProcess([=]() -> SimProcess {

		simu(cmd.valid) = 0;
		simu(cmd.opcode) = 0;
		simu(cmd.connectionIndex) = 0;
		simu(cmd.sequenceNumber) = 0;

		simu(lastSeqNr) = 0;
		simu(aboveWindow) = 0;
		simu(bitmap) = 0;

		co_await WaitClk(clock());

		// clear on new connection
		simu(cmd.valid) = 1;
		simu(cmd.opcode) = 2; // OpSetSeqNr
		simu(bitmap) = 0xFFFFFFFF;

		for (size_t i = 0; i < 32; ++i)
		{
			simu(cmd.sequenceNumber) = 0xFFFFFF00 | i;
			BOOST_TEST(simu(bitmap_next) == 1u << i);
			co_await WaitClk(clock());
		}

		stopTest();
	});
}

BOOST_FIXTURE_TEST_CASE(AntiReplay_secondStageUpdate_new_window_word_test, AntiReplayTest)
{
	windowSize(8_b, 32_b);

	BVec lastSeqNr = pinIn(32_b).setName("lastSeqNr");
	Bit aboveWindow = pinIn().setName("aboveWindow");
	BVec bitmap = pinIn(32_b).setName("bitmap");
	BVec bitmap_next = secondStageUpdate(bitmap, lastSeqNr, aboveWindow);
	pinOut(bitmap_next).setName("bitmap_next");

	addSimulationProcess([=]() -> SimProcess {

		simu(cmd.valid) = 0;
		simu(cmd.opcode) = 0;
		simu(cmd.connectionIndex) = 0;
		simu(cmd.sequenceNumber) = 0;

		simu(lastSeqNr) = 0;
		simu(aboveWindow) = 0;
		simu(bitmap) = 0;

		co_await WaitClk(clock());

		// clear on new connection
		simu(cmd.valid) = 1;
		simu(cmd.opcode) = 0; // OpSetSeqNr
		simu(bitmap) = 0xFFFFFFFF;
		simu(aboveWindow) = 1;

		simu(cmd.sequenceNumber) = 33;
		BOOST_TEST(simu(bitmap_next) == 2);
		co_await WaitClk(clock());

		stopTest();
	});
}


#endif
